/*-------------------------------------------------------------------------
 *
 * nodeMergejoin.c
 *	  routines supporting merge joins
 *
 * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *	  $PostgreSQL: pgsql/src/backend/executor/nodeMergejoin.c,v 1.82.2.1 2007/02/02 00:07:28 tgl Exp $
 *
 *-------------------------------------------------------------------------
 */
/*
 * INTERFACE ROUTINES
 *		ExecMergeJoin			mergejoin outer and inner relations.
 *		ExecInitMergeJoin		creates and initializes run time states
 *		ExecEndMergeJoin		cleans up the node.
 *
 * NOTES
 *
 *		Merge-join is done by joining the inner and outer tuples satisfying
 *		join clauses of the form ((= outerKey innerKey) ...).
 *		The join clause list is provided by the query planner and may contain
 *		more than one (= outerKey innerKey) clause (for composite sort key).
 *
 *		However, the query executor needs to know whether an outer
 *		tuple is "greater/smaller" than an inner tuple so that it can
 *		"synchronize" the two relations. For example, consider the following
 *		relations:
 *
 *				outer: (0 ^1 1 2 5 5 5 6 6 7)	current tuple: 1
 *				inner: (1 ^3 5 5 5 5 6)			current tuple: 3
 *
 *		To continue the merge-join, the executor needs to scan both inner
 *		and outer relations till the matching tuples 5. It needs to know
 *		that currently inner tuple 3 is "greater" than outer tuple 1 and
 *		therefore it should scan the outer relation first to find a
 *		matching tuple and so on.
 *
 *		Therefore, when initializing the merge-join node, we look up the
 *		associated sort operators.	We assume the planner has seen to it
 *		that the inputs are correctly sorted by these operators.  Rather
 *		than directly executing the merge join clauses, we evaluate the
 *		left and right key expressions separately and then compare the
 *		columns one at a time (see MJCompare).
 *
 *
 *		Consider the above relations and suppose that the executor has
 *		just joined the first outer "5" with the last inner "5". The
 *		next step is of course to join the second outer "5" with all
 *		the inner "5's". This requires repositioning the inner "cursor"
 *		to point at the first inner "5". This is done by "marking" the
 *		first inner 5 so we can restore the "cursor" to it before joining
 *		with the second outer 5. The access method interface provides
 *		routines to mark and restore to a tuple.
 *
 *
 *		Essential operation of the merge join algorithm is as follows:
 *
 *		Join {
 *			get initial outer and inner tuples				INITIALIZE
 *			do forever {
 *				while (outer != inner) {					SKIP_TEST
 *					if (outer < inner)
 *						advance outer						SKIPOUTER_ADVANCE
 *					else
 *						advance inner						SKIPINNER_ADVANCE
 *				}
 *				mark inner position							SKIP_TEST
 *				do forever {
 *					while (outer == inner) {
 *						join tuples							JOINTUPLES
 *						advance inner position				NEXTINNER
 *					}
 *					advance outer position					NEXTOUTER
 *					if (outer == mark)						TESTOUTER
 *						restore inner position to mark		TESTOUTER
 *					else
 *						break	// return to top of outer loop
 *				}
 *			}
 *		}
 *
 *		The merge join operation is coded in the fashion
 *		of a state machine.  At each state, we do something and then
 *		proceed to another state.  This state is stored in the node's
 *		execution state information and is preserved across calls to
 *		ExecMergeJoin. -cim 10/31/89
 */
#include <float.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "postgres.h"

#include "access/nbtree.h"
#include "access/htup.h"
#include "access/heapam.h"
#include "catalog/pg_amop.h"
#include "catalog/pg_type.h"
#include "executor/execdebug.h"
#include "executor/execdefs.h"
#include "executor/nodeMergejoin.h"
#include "executor/tuptable.h"
#include "miscadmin.h"
#include "storage/buffile.h"
#include "utils/acl.h"
#include "utils/builtins.h"
#include "utils/lsyscache.h"
#include "utils/memutils.h"
#include "utils/syscache.h"
#include "utils/tuplestore.h"

#include "catalog/pg_namespace.h" /*Added  by Yasin SJ*/


/*
 * Comparison strategies supported by MJCompare
 *
 * XXX eventually should extend these to support descending-order sorts.
 * There are some tricky issues however about being sure we are on the same
 * page as the underlying sort or index as to which end NULLs sort to.
 */
typedef enum
{
	MERGEFUNC_LT,				/* raw "<" operator */ /*YASIN: This is used in SJ*/
	MERGEFUNC_CMP,				/* -1 / 0 / 1 three-way comparator */
} MergeFunctionKind;

/* Runtime data for each mergejoin clause */
typedef struct MergeJoinClauseData
{
	/* Executable expression trees */
	ExprState  *lexpr;			/* left-hand (outer) input expression */
	ExprState  *rexpr;			/* right-hand (inner) input expression */

	/*
	 * If we have a current left or right input tuple, the values of the
	 * expressions are loaded into these fields:
	 */
	Datum		ldatum;			/* current left-hand value */
	Datum		rdatum;			/* current right-hand value */
	bool		lisnull;		/* and their isnull flags */
	bool		risnull;

	/*
	 * Remember whether mergejoin operator is strict (usually it will be).
	 * NOTE: if it's not strict, we still assume it cannot return true for one
	 * null and one non-null input.
	 */
	bool		mergestrict;

	/*
	 * The comparison strategy in use, and the lookup info to let us call the
	 * needed comparison routines.	eqfinfo is the "=" operator itself.
	 * cmpfinfo is either the btree comparator or the "<" operator.
	 */
	MergeFunctionKind cmpstrategy;
	FmgrInfo	eqfinfo;
	FmgrInfo	cmpfinfo;  /* In case of SJ it always refers to the "<" Operator*/
	FmgrInfo	minusfinfo; /*Changed by Yasin SJ*/
	/*Changed by S Pearson*/
	FmgrInfo	productfinfo;

	/*Changed by Yasin SJ*/
	Datum		rnextdatum;			/* next right-hand value */
	bool		rnextisnull;


} MergeJoinClauseData;


#define MarkInnerTuple(innerTupleSlot, mergestate) \
		ExecCopySlot((mergestate)->mj_MarkedTupleSlot, (innerTupleSlot))

/*
 * Added by S Pearson
 */
//-------------------------------------------------------
//  Keeps track of memory usage
//-------------------------------------------------------
#define USEMEM(state, amt) ((state)->availMem -= (amt))
#define FREEMEM(state, amt) ((state)->availMem += (amt))

#define OUTER_PLAN INT_MAX
#define INNER_PLAN INT_MIN
#define INTERNAL_HIBERNATE 1024

/******************************************
 * Prototypes for functions for DBSimJoin *
 ******************************************/
int levenshteinDistance(MergeJoinState* node, char* c1, int c1Len, char* c2, int c2Len);
int getWindowOffset(int numPartitions, int partition1, int partition2);
bool isPartitionInMemory(MergeJoinState *node, Tuplestorestate *partition);
bool areWindowsInMemory(MergeJoinState *node, Tuplestorestate *win1, Tuplestorestate *win2);
bool isPartitionJoinReady(MergeJoinState *node, Tuplestorestate *partition);
bool areWindowsJoinReady(MergeJoinState *node, Tuplestorestate *win1, Tuplestorestate *win2);
void hibernateNewPartitions(MergeJoinState *node);
void hibernateNewWindows(MergeJoinState *node);
void hibernateNewWPartitions(MergeJoinState *node);
void hibernateNewWWindows(MergeJoinState *node);
void getPivotsFromPartition(MergeJoinState *node, Tuplestorestate *partition, int numPivots);
void getPivotsFromWindows(MergeJoinState *node, Tuplestorestate *window0, Tuplestorestate *window1, int numPivots);
void partition(MergeJoinState *node, Tuplestorestate *partition,
		int numPartitions, int partitionOffset, int windowOffset,
		Tuplestorestate **partitionArray, Tuplestorestate **windowArray);
void partitionWindows(MergeJoinState *node, Tuplestorestate *window0, Tuplestorestate *window1,
		int numPartitions, int wPartitionOffset, int wWindowOffset,
		Tuplestorestate **wPartitionArray0, Tuplestorestate **wPartitionArray1,
		Tuplestorestate **wWindowArray0, Tuplestorestate **wWindowArray1);
void resetPivots(MergeJoinState *state);
TupleDesc initNewTupleDescTableID(TupleDesc origDesc);
void initTupleFromPlan(TupleTableSlot *origSlot, TupleTableSlot *newSlot, bool isOuter);
TupleTableSlot *getTuple(MergeJoinState *mergestate, Tuplestorestate *partition);
TupleTableSlot *getNextTuple(MergeJoinState *mergestate, Tuplestorestate *partition);
bool pivotIsOuter(MergeJoinState *mergestate, HeapTuple pivot);
bool isOuter(TupleTableSlot *tuple);
bool fromSameTable(TupleTableSlot *slot1, TupleTableSlot *slot2);
void addTuple(MergeJoinState *mergestate, TupleTableSlot *tuple, int numPivots,
		Tuplestorestate **partitionArray, int partitionOffset,
		Tuplestorestate **windowArray, int windowOffset);
void resetTuple(MergeJoinState *mergestate, TupleTableSlot *slot1);

/*
 * MJExamineQuals
 *
 * This deconstructs the list of mergejoinable expressions, which is given
 * to us by the planner in the form of a list of "leftexpr = rightexpr"
 * expression trees in the order matching the sort columns of the inputs.
 * We build an array of MergeJoinClause structs containing the information
 * we will need at runtime.  Each struct essentially tells us how to compare
 * the two expressions from the original clause.
 *
 * The best, most efficient way to compare two expressions is to use a btree
 * comparison support routine, since that requires only one function call
 * per comparison.	Hence we try to find a btree opclass that matches the
 * mergejoinable operator.	If we cannot find one, we'll have to call both
 * the "=" and (often) the "<" operator for each comparison.
 */
static MergeJoinClause
MJExamineQuals(List *qualList, PlanState *parent)
{
	MergeJoinClause clauses;
	int			nClauses = list_length(qualList);
	int			iClause;
	ListCell   *l;

	clauses = (MergeJoinClause) palloc0(nClauses * sizeof(MergeJoinClauseData));

	iClause = 0;
	foreach(l, qualList)
	{
		OpExpr	   *qual = (OpExpr *) lfirst(l);
		MergeJoinClause clause = &clauses[iClause];
		Oid			ltop;
		Oid			gtop;
		RegProcedure ltproc;
		RegProcedure gtproc;
		AclResult	aclresult;
		CatCList   *catlist;
		int			i;

		if (!IsA(qual, OpExpr))
			elog(ERROR, "mergejoin clause is not an OpExpr");

		/*
		 * Prepare the input expressions for execution.
		 */
		clause->lexpr = ExecInitExpr((Expr *) linitial(qual->args), parent);
		clause->rexpr = ExecInitExpr((Expr *) lsecond(qual->args), parent);

		/*
		 * Check permission to call the mergejoinable operator. For
		 * predictability, we check this even if we end up not using it.
		 */
		aclresult = pg_proc_aclcheck(qual->opfuncid, GetUserId(), ACL_EXECUTE);
		if (aclresult != ACLCHECK_OK)
			aclcheck_error(aclresult, ACL_KIND_PROC,
					get_func_name(qual->opfuncid));

		/* Set up the fmgr lookup information */
		fmgr_info(qual->opfuncid, &(clause->eqfinfo));

		/* And remember strictness */
		clause->mergestrict = clause->eqfinfo.fn_strict;

		/*
		 * Lookup the comparison operators that go with the mergejoinable
		 * top-level operator.	(This will elog if the operator isn't
		 * mergejoinable, which would be the planner's mistake.)
		 */
		op_mergejoin_crossops(qual->opno,
				&ltop,
				&gtop,
				&ltproc,
				&gtproc);

		clause->cmpstrategy = MERGEFUNC_LT;

		/*
		 * Look for a btree opclass including all three operators. This is
		 * much like SelectSortFunction except we insist on matching all the
		 * operators provided, and it can be a cross-type opclass.
		 *
		 * XXX for now, insist on forward sort so that NULLs can be counted on
		 * to be high.
		 */
		catlist = SearchSysCacheList(AMOPOPID, 1,
				ObjectIdGetDatum(qual->opno),
				0, 0, 0);

		for (i = 0; i < catlist->n_members; i++)
		{
			HeapTuple	tuple = &catlist->members[i]->tuple;
			Form_pg_amop aform = (Form_pg_amop) GETSTRUCT(tuple);
			Oid			opcid = aform->amopclaid;

			if (aform->amopstrategy != BTEqualStrategyNumber)
				continue;
			if (!opclass_is_btree(opcid))
				continue;
			if (get_op_opclass_strategy(ltop, opcid) == BTLessStrategyNumber &&
					get_op_opclass_strategy(gtop, opcid) == BTGreaterStrategyNumber)
			{
				clause->cmpstrategy = MERGEFUNC_CMP;
				ltproc = get_opclass_proc(opcid, aform->amopsubtype,
						BTORDER_PROC);
				Assert(RegProcedureIsValid(ltproc));
				break;			/* done looking */
			}
		}

		ReleaseSysCacheList(catlist);

		/* Check permission to call "<" operator or cmp function */
		aclresult = pg_proc_aclcheck(ltproc, GetUserId(), ACL_EXECUTE);
		if (aclresult != ACLCHECK_OK)
			aclcheck_error(aclresult, ACL_KIND_PROC,
					get_func_name(ltproc));

		/* Set up the fmgr lookup information */
		fmgr_info(ltproc, &(clause->cmpfinfo));

		iClause++;
	}

	return clauses;
}

/*
 * Added by Yasin SJ
 * Similar to MJExamineQuals
 * MJExamineQuals_AddsMinus Loads also minus operators (Currently used in epsilon-join
 *
 * It always loads =, <, and -(minus) since all of them are needed for certain kinds of SJ
 * Does not load the btree comparison operator
 */
static MergeJoinClause
MJExamineQuals_AddsMinus(List *qualList, PlanState *parent)
{
	MergeJoinClause clauses;
	int			nClauses = list_length(qualList);
	int			iClause;
	ListCell   *l;

	clauses = (MergeJoinClause) palloc0(nClauses * sizeof(MergeJoinClauseData));

	iClause = 0;
	foreach(l, qualList)
	{
		OpExpr	   *qual = (OpExpr *) lfirst(l);
		MergeJoinClause clause = &clauses[iClause];
		Oid			ltop;
		Oid			gtop;
		RegProcedure ltproc;
		RegProcedure gtproc;
		AclResult	aclresult;
		CatCList   *catlist;
		int			i;

		Oid loptype; /*Added by Yasin SJ. oid of left operator's type*/
		Oid roptype; /*Added by Yasin SJ. oid of right operator's type*/

		if (!IsA(qual, OpExpr))
			elog(ERROR, "mergejoin clause is not an OpExpr");

		/*
		 * Prepare the input expressions for execution.
		 */
		clause->lexpr = ExecInitExpr((Expr *) linitial(qual->args), parent);
		clause->rexpr = ExecInitExpr((Expr *) lsecond(qual->args), parent);

		/*
		 * Check permission to call the mergejoinable operator. For
		 * predictability, we check this even if we end up not using it.
		 */
		aclresult = pg_proc_aclcheck(qual->opfuncid, GetUserId(), ACL_EXECUTE);
		if (aclresult != ACLCHECK_OK)
			aclcheck_error(aclresult, ACL_KIND_PROC,
					get_func_name(qual->opfuncid));

		/* Set up the fmgr lookup information */
		fmgr_info(qual->opfuncid, &(clause->eqfinfo));

		/* And remember strictness */
		clause->mergestrict = clause->eqfinfo.fn_strict;

		op_mergejoin_crossops_SJ(qual->opno,
				&ltop,
				&gtop,
				&ltproc, /*SJ also uses this*/
				&gtproc,
				&loptype,  /*Added by Yasin SJ*/
				&roptype); /*Added by Yasin SJ*/
		clause->cmpstrategy = MERGEFUNC_LT; /* SJ uses this value*/

		/* Gets the "-" (Minus) operator */
		HeapTuple	opertup;
		char		opname[NAMEDATALEN];
		StrNCpy(opname, "-", NAMEDATALEN);
		/*We search/load the tuple in OPERNAMENSP cache that corresponds to the "-" operator for the given type*/
		opertup = SearchSysCache(OPERNAMENSP, /*name of cache*/
				CStringGetDatum(opname),/*op name*/
				ObjectIdGetDatum(loptype),/*letf op type*/
				ObjectIdGetDatum(roptype),/*right op type*/
				ObjectIdGetDatum(PG_CATALOG_NAMESPACE)); /*id of pg_catalog namespace: PGNSP*/
		if (HeapTupleIsValid(opertup))
		{
			Oid minus_opr = HeapTupleGetOid(opertup);
			Oid minusproc = get_opcode(minus_opr);
			/* Check permission to call "-" operator */
			aclresult = pg_proc_aclcheck(minusproc, GetUserId(), ACL_EXECUTE);
			if (aclresult != ACLCHECK_OK)
				aclcheck_error(aclresult, ACL_KIND_PROC, get_func_name(minusproc));
			/*Set up the fmgr lookup information */
			fmgr_info(minusproc, &(clause->minusfinfo));
			ReleaseSysCache(opertup);
		}
		else /*If it is not valid*/
			elog(ERROR, "Failed to find - (minus) operator  for Similarity Join");

		/* Check permission to call "<" operator or cmp function */
		aclresult = pg_proc_aclcheck(ltproc, GetUserId(), ACL_EXECUTE);
		if (aclresult != ACLCHECK_OK)
			aclcheck_error(aclresult, ACL_KIND_PROC,get_func_name(ltproc));

		/* Set up the fmgr lookup information */
		fmgr_info(ltproc, &(clause->cmpfinfo));

		iClause++;
	}

	return clauses;
}

/*
 * Added by S Pearson
 * Similar to MJExamineQuals, used for multi-dimensional similarity join
 * MJExamineQuals_AddsMinus also loads product and addition operators
 *
 * It always loads =, <, -, and * since all of them are needed for
 * checking the Euclidean distance for multi-dimensional Similarity Join
 * Does not load the btree comparison operator
 */
static MergeJoinClause
MJExamineQuals_AddsProductMinus(List *qualList, PlanState *parent)
{
	MergeJoinClause clauses;
	int			nClauses = list_length(qualList);
	int			iClause;
	ListCell   *l;

	clauses = (MergeJoinClause) palloc0(nClauses * sizeof(MergeJoinClauseData));

	iClause = 0;
	foreach(l, qualList)
	{
		OpExpr	   *qual = (OpExpr *) lfirst(l);
		MergeJoinClause clause = &clauses[iClause];
		Oid			ltop;
		Oid			gtop;
		RegProcedure ltproc;
		RegProcedure gtproc;
		AclResult	aclresult;

		Oid loptype; /*oid of left operator's type*/
		Oid roptype; /*oid of right operator's type*/

		if (!IsA(qual, OpExpr))
			elog(ERROR, "mergejoin clause is not an OpExpr");

		/*
		 * Prepare the input expressions for execution.
		 */
		clause->lexpr = ExecInitExpr((Expr *) linitial(qual->args), parent);
		clause->rexpr = ExecInitExpr((Expr *) lsecond(qual->args), parent);

		/*
		 * Check permission to call the mergejoinable operator. For
		 * predictability, we check this even if we end up not using it.
		 */
		aclresult = pg_proc_aclcheck(qual->opfuncid, GetUserId(), ACL_EXECUTE);
		if (aclresult != ACLCHECK_OK)
			aclcheck_error(aclresult, ACL_KIND_PROC,
					get_func_name(qual->opfuncid));

		/* Set up the fmgr lookup information */
		fmgr_info(qual->opfuncid, &(clause->eqfinfo));

		/* And remember strictness */
		clause->mergestrict = clause->eqfinfo.fn_strict;

		op_mergejoin_crossops_SJ(qual->opno,
				&ltop,
				&gtop,
				&ltproc, /*SJ also uses this*/
				&gtproc,
				&loptype,  /*Added by Yasin SJ*/
				&roptype); /*Added by Yasin SJ*/
		clause->cmpstrategy = MERGEFUNC_LT; /* SJ uses this value*/

		/* Gets the "*" (Product) operator */
		HeapTuple	prodopertup;
		char		prodopname[NAMEDATALEN];
		StrNCpy(prodopname, "*", NAMEDATALEN);
		/*We search/load the tuple in OPERNAMENSP cache that corresponds to the "*" operator for the given type*/
		prodopertup = SearchSysCache(OPERNAMENSP, /*name of cache*/
				CStringGetDatum(prodopname),/*op name*/
				ObjectIdGetDatum(loptype),/*letf op type*/
				ObjectIdGetDatum(roptype),/*right op type*/
				ObjectIdGetDatum(PG_CATALOG_NAMESPACE)); /*id of pg_catalog namespace: PGNSP*/
		if (HeapTupleIsValid(prodopertup))
		{
			Oid product_opr = HeapTupleGetOid(prodopertup);
			Oid productproc = get_opcode(product_opr);
			/* Check permission to call "*" operator */
			aclresult = pg_proc_aclcheck(productproc, GetUserId(), ACL_EXECUTE);
			if (aclresult != ACLCHECK_OK)
				aclcheck_error(aclresult, ACL_KIND_PROC, get_func_name(productproc));
			/*Set up the fmgr lookup information */
			fmgr_info(productproc, &(clause->productfinfo));
			ReleaseSysCache(prodopertup);
		}
		else /*If it is not valid*/
			elog(ERROR, "Failed to find * (product) operator  for Similarity Join");

		/* Gets the "-" (Minus) operator */
		HeapTuple	opertup;
		char		opname[NAMEDATALEN];
		StrNCpy(opname, "-", NAMEDATALEN);
		/*We search/load the tuple in OPERNAMENSP cache that corresponds to the "-" operator for the given type*/
		opertup = SearchSysCache(OPERNAMENSP, /*name of cache*/
				CStringGetDatum(opname),/*op name*/
				ObjectIdGetDatum(loptype),/*letf op type*/
				ObjectIdGetDatum(roptype),/*right op type*/
				ObjectIdGetDatum(PG_CATALOG_NAMESPACE)); /*id of pg_catalog namespace: PGNSP*/
		if (HeapTupleIsValid(opertup))
		{
			Oid minus_opr = HeapTupleGetOid(opertup);
			Oid minusproc = get_opcode(minus_opr);
			/* Check permission to call "-" operator */
			aclresult = pg_proc_aclcheck(minusproc, GetUserId(), ACL_EXECUTE);
			if (aclresult != ACLCHECK_OK)
				aclcheck_error(aclresult, ACL_KIND_PROC, get_func_name(minusproc));
			/*Set up the fmgr lookup information */
			fmgr_info(minusproc, &(clause->minusfinfo));
			ReleaseSysCache(opertup);
		}
		else /*If it is not valid*/
			elog(ERROR, "Failed to find - (minus) operator  for Similarity Join");

		/* Check permission to call "<" operator or cmp function */
		aclresult = pg_proc_aclcheck(ltproc, GetUserId(), ACL_EXECUTE);
		if (aclresult != ACLCHECK_OK)
			aclcheck_error(aclresult, ACL_KIND_PROC,get_func_name(ltproc));

		/* Set up the fmgr lookup information */
		fmgr_info(ltproc, &(clause->cmpfinfo));

		iClause++;
	}

	return clauses;
}

/*
 * Added by S Pearson
 *
 * Called to set up for comparing strings in DBSimJoin.
 * We don't actually have any comparison operators we can load for Strings, but we
 * still need to do some setup.
 */
static MergeJoinClause
MJExamineQuals_Strings(List *qualList, PlanState *parent)
{
	MergeJoinClause clauses;
	int			nClauses = list_length(qualList);
	int			iClause;
	ListCell   *l;

	clauses = (MergeJoinClause) palloc0(nClauses * sizeof(MergeJoinClauseData));

	iClause = 0;
	foreach(l, qualList)
	{
		OpExpr	   *qual = (OpExpr *) lfirst(l);
		MergeJoinClause clause = &clauses[iClause];
		Oid			ltop;
		Oid			gtop;
		RegProcedure ltproc;
		RegProcedure gtproc;
		AclResult	aclresult;

		Oid loptype; /*oid of left operator's type*/
		Oid roptype; /*oid of right operator's type*/

		if (!IsA(qual, OpExpr))
			elog(ERROR, "mergejoin clause is not an OpExpr");

		/*
		 * Prepare the input expressions for execution.
		 */
		clause->lexpr = ExecInitExpr((Expr *) linitial(qual->args), parent);
		clause->rexpr = ExecInitExpr((Expr *) lsecond(qual->args), parent);

		/*
		 * Check permission to call the mergejoinable operator. For
		 * predictability, we check this even if we end up not using it.
		 */
		aclresult = pg_proc_aclcheck(qual->opfuncid, GetUserId(), ACL_EXECUTE);
		if (aclresult != ACLCHECK_OK)
			aclcheck_error(aclresult, ACL_KIND_PROC,
					get_func_name(qual->opfuncid));

		/* Set up the fmgr lookup information */
		fmgr_info(qual->opfuncid, &(clause->eqfinfo));

		/* And remember strictness */
		clause->mergestrict = clause->eqfinfo.fn_strict;

		op_mergejoin_crossops_SJ(qual->opno,
				&ltop,
				&gtop,
				&ltproc, /*SJ also uses this*/
				&gtproc,
				&loptype,  /*Added by Yasin SJ*/
				&roptype); /*Added by Yasin SJ*/
		clause->cmpstrategy = MERGEFUNC_LT; /* SJ uses this value*/

		/* Set up the fmgr lookup information */
		fmgr_info(ltproc, &(clause->cmpfinfo));

		iClause++;
	}

	return clauses;
}

/*
 * MJEvalOuterValues
 *
 * Compute the values of the mergejoined expressions for the current
 * outer tuple.  We also detect whether it's impossible for the current
 * outer tuple to match anything --- this is true if it yields a NULL
 * input for any strict mergejoin operator.
 *
 * We evaluate the values in OuterEContext, which can be reset each
 * time we move to a new tuple.
 */
static bool
MJEvalOuterValues(MergeJoinState *mergestate)
{
	ExprContext *econtext = mergestate->mj_OuterEContext;
	bool		canmatch = true;
	int			i;
	MemoryContext oldContext;

	ResetExprContext(econtext);

	oldContext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);

	econtext->ecxt_outertuple = mergestate->mj_OuterTupleSlot;

	for (i = 0; i < mergestate->mj_NumClauses; i++)
	{
		MergeJoinClause clause = &mergestate->mj_Clauses[i];

		clause->ldatum = ExecEvalExpr(clause->lexpr, econtext,
				&clause->lisnull, NULL);
		if (clause->lisnull && clause->mergestrict)
			canmatch = false;
	}

	MemoryContextSwitchTo(oldContext);

	return canmatch;
}

/*
 * MJEvalInnerValues
 *
 * Same as above, but for the inner tuple.	Here, we have to be prepared
 * to load data from either the true current inner, or the marked inner,
 * so caller must tell us which slot to load from.
 */
static bool
MJEvalInnerValues(MergeJoinState *mergestate, TupleTableSlot *innerslot)
{
	ExprContext *econtext = mergestate->mj_InnerEContext;
	bool		canmatch = true;
	int			i;
	MemoryContext oldContext;

	ResetExprContext(econtext);

	oldContext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);

	econtext->ecxt_innertuple = innerslot;

	for (i = 0; i < mergestate->mj_NumClauses; i++)
	{
		MergeJoinClause clause = &mergestate->mj_Clauses[i];

		clause->rdatum = ExecEvalExpr(clause->rexpr, econtext,
				&clause->risnull, NULL);
		if (clause->risnull && clause->mergestrict)
			canmatch = false;
	}

	MemoryContextSwitchTo(oldContext);

	return canmatch;
}


/*
 * MJEvalNextInnerValues  Added by Yasin SJ
 *
 */
static bool
MJEvalNextInnerValues(MergeJoinState *mergestate, TupleTableSlot *nextinnerslot)
{
	ExprContext *econtext = mergestate->mj_InnerEContext;
	bool		canmatch = true;
	int			i;
	MemoryContext oldContext;

	ResetExprContext(econtext);

	oldContext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);

	econtext->ecxt_innertuple = nextinnerslot;

	for (i = 0; i < mergestate->mj_NumClauses; i++)
	{
		MergeJoinClause clause = &mergestate->mj_Clauses[i];

		clause->rnextdatum = ExecEvalExpr(clause->rexpr, econtext,
				&clause->rnextisnull, NULL);
		if (clause->rnextisnull && clause->mergestrict)
			canmatch = false;
	}

	MemoryContextSwitchTo(oldContext);

	return canmatch;
}

/*
 * Added by S Pearson
 *
 * Same as above, but the caller must tell us which tuple to take.
 * We also must then figure out which plan it was originally from.
 */
static bool
MJEvalNextValue(MergeJoinState *mergestate, TupleTableSlot *slot)
{
	ExprContext *econtext;
	bool		canmatch = true, outer;
	int			i;
	MemoryContext oldContext;

	outer = isOuter(slot);
	econtext = outer ? mergestate->mj_OuterEContext : mergestate->mj_InnerEContext;

	ResetExprContext(econtext);

	oldContext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);

	if(outer)
		econtext->ecxt_outertuple = slot;
	else
		econtext->ecxt_innertuple = slot;

	for (i = 0; i < mergestate->mj_NumClauses; i++)
	{
		MergeJoinClause clause = &mergestate->mj_Clauses[i];

		if(outer)
		{
			 clause->ldatum = ExecEvalExpr(clause->lexpr, econtext,
				&clause->lisnull, NULL);
			if (clause->lisnull  && clause->mergestrict)
				canmatch = false;
		}
		else
		{
			clause->rdatum = ExecEvalExpr(clause->rexpr, econtext,
				&clause->risnull, NULL);
			if (clause->risnull && clause->mergestrict)
				canmatch = false;
		}

	}

	MemoryContextSwitchTo(oldContext);

	return canmatch;
}

/*
 * Added by S Pearson
 * Same as above, but takes a pivot.
 * We just use the already existing rnextdatum and rnextisnull variables in
 * MergeJoinClause to store it.
 */
static bool
MJEvalNextValuePivot(MergeJoinState *mergestate, MinimalTuple pivot)
{
	ExprContext *econtext;
	bool		canmatch = true, outer;
	int			i;
	MemoryContext oldContext;
	HeapTuple 	hPivot = heap_tuple_from_minimal_tuple(pivot);

	outer = pivotIsOuter(mergestate, hPivot);
	econtext = outer ? mergestate->mj_OuterEContext : mergestate->mj_InnerEContext;

	ResetExprContext(econtext);

	oldContext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);
	if(outer)
	{
		ExecStoreTuple(hPivot, mergestate->extraOuter, InvalidBuffer, true);
		econtext->ecxt_outertuple = mergestate->extraOuter;
	}
	else
	{
		ExecStoreTuple(hPivot, mergestate->extraInner, InvalidBuffer, true);
		econtext->ecxt_innertuple = mergestate->extraInner;
	}


	for (i = 0; i < mergestate->mj_NumClauses; i++)
	{
		MergeJoinClause clause = &mergestate->mj_Clauses[i];

		//Use rnextdatum and rnextisnull regardless.  Setting lexpr or rexpr is what matters.
		if(outer)
		{
			clause->rnextdatum = ExecEvalExpr(clause->lexpr, econtext,
					&clause->rnextisnull, NULL);
			if (clause->rnextisnull && clause->mergestrict)
				canmatch = false;
		}
		else
		{
			clause->rnextdatum = ExecEvalExpr(clause->rexpr, econtext,
					&clause->rnextisnull, NULL);
			if (clause->rnextisnull && clause->mergestrict)
				canmatch = false;
		}
	}

	MemoryContextSwitchTo(oldContext);

	return canmatch;
}


/*
 * MJCompare
 *
 * Compare the mergejoinable values of the current two input tuples
 * and return 0 if they are equal (ie, the mergejoin equalities all
 * succeed), +1 if outer > inner, -1 if outer < inner.
 *
 * MJEvalOuterValues and MJEvalInnerValues must already have been called
 * for the current outer and inner tuples, respectively.
 *
 * For Epsilon-Join: (Added by Yasin)
 * return 0 if dist(outer,inner)<=Epsilon,
 * else: +1 if outer > inner, -1 if outer < inner.
 *
 * For Join-Around: (Added by Yasin)
 * MJCompare is called only with (outer, inner, and next inner) not null
 * if (outer <= inner) returns 0
 * else
 *    if dist(outer,inner) < dist(outer,nextinner)  Return 0  (could be also <=)
 *    else Return +1
 */
static int
MJCompare(MergeJoinState *mergestate)
{
	int			result = 0;
	bool		nulleqnull = false;
	ExprContext *econtext = mergestate->js.ps.ps_ExprContext;
	int			i;
	MemoryContext oldContext;
	FunctionCallInfoData fcinfo;

	/* CHANGED BY Ahmed.M.Aly */
	double epsilonDbl;
	epsilonDbl = ((MergeJoin *) mergestate->js.ps.plan)->epsilon;
	Datum epsilonDatum = Float8GetDatum(epsilonDbl);

	/*
	 * Call the comparison functions in short-lived context, in case they leak
	 * memory.
	 */
	ResetExprContext(econtext);

	oldContext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);

	for (i = 0; i < mergestate->mj_NumClauses; i++)
	{
		MergeJoinClause clause = &mergestate->mj_Clauses[i];
		Datum		fresult, fresult2;

		/*
		 * Deal with null inputs.  We treat NULL as sorting after non-NULL.
		 *
		 * If both inputs are NULL, and the comparison function isn't strict,
		 * then we call it and check for a true result (this allows operators
		 * that behave like IS NOT DISTINCT to be mergejoinable). If the
		 * function is strict or returns false, we temporarily pretend NULL ==
		 * NULL and contine checking remaining columns.
		 */
		if (clause->lisnull)
		{
			if (clause->risnull)
			{
				if (!clause->eqfinfo.fn_strict)
				{
					InitFunctionCallInfoData(fcinfo, &(clause->eqfinfo), 2,
							NULL, NULL);
					fcinfo.arg[0] = clause->ldatum;
					fcinfo.arg[1] = clause->rdatum;
					fcinfo.argnull[0] = true;
					fcinfo.argnull[1] = true;
					fresult = FunctionCallInvoke(&fcinfo);
					if (!fcinfo.isnull && DatumGetBool(fresult))
					{
						/* treat nulls as really equal */
						continue;
					}
				}
				nulleqnull = true;
				continue;
			}
			/* NULL > non-NULL */
			result = 1;
			break;
		}
		if (clause->risnull)
		{
			/* non-NULL < NULL */
			result = -1;
			break;
		}

		if (clause->cmpstrategy == MERGEFUNC_LT) /*Epsilon-Join and Join-Around always use this strategy*/
		{
			if (mergestate->mj_SimjoinType == 0) { /*No SJ. Old Code.*/
				InitFunctionCallInfoData(fcinfo, &(clause->eqfinfo), 2,NULL, NULL);
				fcinfo.arg[0] = clause->ldatum;
				fcinfo.arg[1] = clause->rdatum;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull)
				{
					nulleqnull = true;
					continue;
				}
				else if (DatumGetBool(fresult))
				{/* equal */
					continue;
				}
				InitFunctionCallInfoData(fcinfo, &(clause->cmpfinfo), 2,NULL, NULL);
				fcinfo.arg[0] = clause->ldatum;
				fcinfo.arg[1] = clause->rdatum;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull)
				{
					nulleqnull = true;
					continue;
				}
				else if (DatumGetBool(fresult))
				{/* less than */
					result = -1;
					break;
				}
				else
				{/* greater than */
					result = 1;
					break;
				}
			}
			else if (mergestate->mj_SimjoinType == 1){/*Epsilon Join*/
				InitFunctionCallInfoData(fcinfo, &(clause->minusfinfo), 2, NULL, NULL); /*outer-inner*/
				fcinfo.arg[0] = clause->ldatum;
				fcinfo.arg[1] = clause->rdatum;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull) elog(ERROR, "Failed to perform minus operation in Similarity Join");

				InitFunctionCallInfoData(fcinfo, &(clause->minusfinfo), 2, NULL, NULL);/*inner-outer*/
				fcinfo.arg[0] = clause->rdatum;
				fcinfo.arg[1] = clause->ldatum;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult2 = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull) elog(ERROR, "Failed to perform minus operation in Similarity Join");

				InitFunctionCallInfoData(fcinfo, &(clause->cmpfinfo), 2, NULL, NULL);/*epsilon < (outer-inner)*/
				fcinfo.arg[0] = epsilonDatum;/*Epsilon*/
				fcinfo.arg[1] = fresult;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull) elog(ERROR, "Failed to perform < operation in Similarity Join");

				InitFunctionCallInfoData(fcinfo, &(clause->cmpfinfo), 2, NULL, NULL);/*epsilon < (inner-outer)*/
				fcinfo.arg[0] = epsilonDatum;/*Epsilon*/
				fcinfo.arg[1] = fresult2;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult2 = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull) elog(ERROR, "Failed to perform < operation in Similarity Join");

				if ( ! (DatumGetBool(fresult) ||  DatumGetBool(fresult2)) ) /* dist (outer,inner) <= Epsilon */
				{   /* SIMILAR */
					continue;

				}
				/*Not Similar*/
				InitFunctionCallInfoData(fcinfo, &(clause->cmpfinfo), 2,NULL, NULL);/*Outer < Inner*/
				fcinfo.arg[0] = clause->ldatum;
				fcinfo.arg[1] = clause->rdatum;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull)
				{
					nulleqnull = true;
					continue;
				}
				else if (DatumGetBool(fresult))
				{/* less than */
					result = -1;
					break;
				}
				else
				{/* greater than */
					result = 1;
					break;
				}
			}
			else if (mergestate->mj_SimjoinType == 2){/*Join-Around*/
				/* if (outer <= inner) returns 0 */
				InitFunctionCallInfoData(fcinfo, &(clause->cmpfinfo), 2, NULL, NULL);/*inner < outer*/
				fcinfo.arg[0] = clause->rdatum;/*inner*/
				fcinfo.arg[1] = clause->ldatum;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull) elog(ERROR, "Failed to perform < operation in Similarity Join");
				if (!DatumGetBool(fresult)) /*SIMILAR*/
					continue;

				/* if dist(outer,inner) < dist(outer,nextinner)  Return 0, else Return +1 */
				/* First we test if (outer > nextinner), if true return +1*/
				InitFunctionCallInfoData(fcinfo, &(clause->cmpfinfo), 2, NULL, NULL);/*nextinner < outer*/
				fcinfo.arg[0] = clause->rnextdatum;/*nextinner*/
				fcinfo.arg[1] = clause->ldatum;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull) elog(ERROR, "Failed to perform < operation in Similarity Join");
				if (DatumGetBool(fresult)){ /*outer is closer to nextinner since it is > than nextinner*/
					result = +1;
					break;
				}
				/*At this point d(o,i)=o-i, d(o,ni)=ni-o (since inner<outer<=nextinner)*/
				InitFunctionCallInfoData(fcinfo, &(clause->minusfinfo), 2, NULL, NULL); /*outer-inner*/
				fcinfo.arg[0] = clause->ldatum;
				fcinfo.arg[1] = clause->rdatum;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull) elog(ERROR, "Failed to perform minus operation in Similarity Join");

				InitFunctionCallInfoData(fcinfo, &(clause->minusfinfo), 2, NULL, NULL);/*nextinner-outer*/
				fcinfo.arg[0] = clause->rnextdatum;
				fcinfo.arg[1] = clause->ldatum;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult2 = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull) elog(ERROR, "Failed to perform minus operation in Similarity Join");

				InitFunctionCallInfoData(fcinfo, &(clause->cmpfinfo), 2, NULL, NULL);/*(outer-inner) < (nextinner-outer)*/
				fcinfo.arg[0] = fresult;
				fcinfo.arg[1] = fresult2;
				fcinfo.argnull[0] = false;
				fcinfo.argnull[1] = false;
				fresult = FunctionCallInvoke(&fcinfo);
				if (fcinfo.isnull) elog(ERROR, "Failed to perform < operation in Similarity Join");
				if (DatumGetBool(fresult)) /*dist(outer,inner) < dist(outer,nextinner)*/
				{   /* SIMILAR */
					continue;
				}
				else{/*Not Similar. We return +1 directly since we already know that outer>inner*/
					result = +1;
					break;
				}
			}
		}
		else
			/* must be MERGEFUNC_CMP */
		{
			InitFunctionCallInfoData(fcinfo, &(clause->cmpfinfo), 2,
					NULL, NULL);
			fcinfo.arg[0] = clause->ldatum;
			fcinfo.arg[1] = clause->rdatum;
			fcinfo.argnull[0] = false;
			fcinfo.argnull[1] = false;
			fresult = FunctionCallInvoke(&fcinfo);
			if (fcinfo.isnull)
			{
				nulleqnull = true;
				continue;
			}
			else if (DatumGetInt32(fresult) == 0)
			{
				/* equal */
				continue;
			}
			else if (DatumGetInt32(fresult) < 0)
			{
				/* less than */
				result = -1;
				break;
			}
			else
			{
				/* greater than */
				result = 1;
				break;
			}
		}
	}

	/*
	 * If we had any null comparison results or NULL-vs-NULL inputs, we do not
	 * want to report that the tuples are equal.  Instead, if result is still
	 * 0, change it to +1.	This will result in advancing the inner side of
	 * the join.
	 */
	if (nulleqnull && result == 0)
		result = 1;

	MemoryContextSwitchTo(oldContext);

	return result;
}


/*
 * MJCompare_InnerAndNextInner   Added by Yasin SJ
 *
 * If inner equals nextinner return 0
 * else return +1 (this value only means different)
 *
 * This method compares all the join-clause's RHS values of two inner tuples (inner, nextinner)
 * Currently it uses the equality functions (to compare RHS and LHS of the clauses).
 */
static int
MJCompare_InnerAndNextInner(MergeJoinState *mergestate)
{
	int			result = 0;
	ExprContext *econtext = mergestate->js.ps.ps_ExprContext;
	int			i;
	MemoryContext oldContext;
	FunctionCallInfoData fcinfo;

	/* Call the comparison functions in short-lived context, in case they leak memory. */
	ResetExprContext(econtext);
	oldContext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);

	for (i = 0; i < mergestate->mj_NumClauses; i++)
	{
		MergeJoinClause clause = &mergestate->mj_Clauses[i];
		Datum		fresult, fresult2;

		/* Deal with null inputs.  We treat NULL as sorting after non-NULL.
		 * If two tuples have all nulls, we consider that they are equal */
		if (clause->risnull)
		{
			if (clause->rnextisnull)
			{/* treat nulls as equal */
				continue;
			}
			result = 1;/* NULL != non-NULL */
			break;
		}
		if (clause->rnextisnull)
		{/* non-NULL != NULL */
			result = 1;
			break;
		}

		InitFunctionCallInfoData(fcinfo, &(clause->eqfinfo), 2,NULL, NULL);
		fcinfo.arg[0] = clause->rdatum;
		fcinfo.arg[1] = clause->rnextdatum;
		fcinfo.argnull[0] = false;
		fcinfo.argnull[1] = false;
		fresult = FunctionCallInvoke(&fcinfo);
		if (fcinfo.isnull){ /*This is not expected. Currently treated as equal*/
			continue;
		}
		else if (DatumGetBool(fresult)){
			continue; /* equal */
		}
		else{
			result = 1; /* not equal */
			break;
		}
	}

	MemoryContextSwitchTo(oldContext);

	return result;
}

/*
 * Added by S Pearson
 * Get the distance from the tuple to the target pivot
 */
static double
MJCompare_DistFuncPivot(MergeJoinState *mergestate,
		TupleTableSlot *slot, MinimalTuple pivot)
{
	double result = 0;
	bool slotOuter = isOuter(slot);
	//double clauseDistance;
	ExprContext *econtext = mergestate->js.ps.ps_ExprContext;
	int			i;
	MemoryContext oldContext;
	FunctionCallInfoData fcinfo;

	//Load values and check for nulls.  Always will be strict.
	if(!(MJEvalNextValue(mergestate, slot) && MJEvalNextValuePivot(mergestate, pivot)))
		return -1;

	/*
	 * Call the comparison functions in short-lived context, in case they leak
	 * memory.
	 */
	ResetExprContext(econtext);

	oldContext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);

	for (i = 0; i < mergestate->mj_NumClauses; i++)
	{
		MergeJoinClause clause = &mergestate->mj_Clauses[i];
		Datum fresult;

		InitFunctionCallInfoData(fcinfo, &(clause->minusfinfo), 2,NULL, NULL);
		fcinfo.arg[0] = slotOuter ? clause->ldatum : clause->rdatum;
		fcinfo.arg[1] = clause->rnextdatum;
		fcinfo.argnull[0] = false;
		fcinfo.argnull[1] = false;
		fresult = FunctionCallInvoke(&fcinfo);
		if (fcinfo.isnull) elog(ERROR, "Failed to perform minus operation in DBSimJoin");

		/*InitFunctionCallInfoData(fcinfo, &(clause->productfinfo), 2,NULL, NULL);
		fcinfo.arg[0] = fresult;
		fcinfo.arg[1] = fresult;
		fcinfo.argnull[0] = false;
		fcinfo.argnull[1] = false;
		clauseDistance = FunctionCallInvoke(&fcinfo);
		if (fcinfo.isnull) elog(ERROR, "Failed to perform product operation in DBSimJoin");*/

		//XXX Revert to commented code in case of problems squaring results.
		result+=  DatumGetFloat4(fresult) * DatumGetFloat4(fresult);//DatumGetInt32(fresult) * DatumGetInt32(fresult);
	}

	MemoryContextSwitchTo(oldContext);
	return sqrt(result);
}
/*
 * Added by S Pearson
 * Get the distance from the tuple to the target tuple
 */
static double
MJCompare_DistFunc(MergeJoinState *mergestate,
		TupleTableSlot *slot1, TupleTableSlot *slot2)
{
	double result = 0;
	//Datum clauseDistance;
	ExprContext *econtext = mergestate->js.ps.ps_ExprContext;
	int			i;
	MemoryContext oldContext;
	FunctionCallInfoData fcinfo;

	//Load values and check for nulls.  Always will be strict.
	if(!(MJEvalNextValue(mergestate, slot1) && MJEvalNextValue(mergestate, slot2)))
		return -1;

	/*
	 * Call the comparison functions in short-lived context, in case they leak
	 * memory.
	 */
	ResetExprContext(econtext);

	oldContext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);

	for (i = 0; i < mergestate->mj_NumClauses; i++)
	{
		MergeJoinClause clause = &mergestate->mj_Clauses[i];
		Datum fresult;

		InitFunctionCallInfoData(fcinfo, &(clause->minusfinfo), 2,NULL, NULL);
		fcinfo.arg[0] = clause->ldatum;
		fcinfo.arg[1] = clause->rdatum;
		fcinfo.argnull[0] = false;
		fcinfo.argnull[1] = false;
		fresult = FunctionCallInvoke(&fcinfo);
		if (fcinfo.isnull) elog(ERROR, "Failed to perform minus operation in DBSimJoin");

		/*InitFunctionCallInfoData(fcinfo, &(clause->productfinfo), 2,NULL, NULL);
		fcinfo.arg[0] = fresult;
		fcinfo.arg[1] = fresult;
		fcinfo.argnull[0] = false;
		fcinfo.argnull[1] = false;
		clauseDistance = FunctionCallInvoke(&fcinfo);
		if (fcinfo.isnull) elog(ERROR, "Failed to perform product operation in DBSimJoin");*/

		//XXX Revert to commented code in case of problems squaring results.
		result+= DatumGetFloat4(fresult) * DatumGetFloat4(fresult);//fresult * fresult;
	}

	MemoryContextSwitchTo(oldContext);
	return sqrt(result);
}

/*
 * Added by S Pearson
 * Get the distance from the tuple to the target pivot
 */
static double
MJCompare_EditDistPivot(MergeJoinState *mergestate,
		TupleTableSlot *slot, MinimalTuple pivot)
{
	double result = 0;
	bool slotOuter = isOuter(slot);
	//double clauseDistance;
	ExprContext *econtext = mergestate->js.ps.ps_ExprContext;
	int			i;
	MemoryContext oldContext;

	//Load values and check for nulls.  Always will be strict.
	if(!(MJEvalNextValue(mergestate, slot) && MJEvalNextValuePivot(mergestate, pivot)))
		return -1;

	/*
	 * Call the comparison functions in short-lived context, in case they leak
	 * memory.
	 */
	ResetExprContext(econtext);

	oldContext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);

	char* lString;
	char* rString;
	int clauseDistance;
	bool flag;
	for (i = 0; i < mergestate->mj_NumClauses; i++)
	{
		MergeJoinClause clause = &mergestate->mj_Clauses[i];

		lString = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(slotOuter ? clause->ldatum : clause->rdatum)));
		rString = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(clause->rnextdatum)));
		if(strcmp(lString,"Ying Li") == 0 || strcmp(lString,"Ying Liu") == 0 ||
				strcmp(rString,"Ying Li") == 0 || strcmp(rString,"Ying Liu") == 0)
			flag = true;

		if (mergestate->mj_NumClauses > 1)
		{
			clauseDistance = levenshteinDistance(mergestate, lString,
					strlen(lString), rString, strlen(rString));
			result += clauseDistance * clauseDistance;
		}
		else
			result += levenshteinDistance(mergestate, lString, strlen(lString), rString, strlen(rString));
	}

	MemoryContextSwitchTo(oldContext);
	return mergestate->mj_NumClauses > 1 ? sqrt((double)result) : result;
}
/*
 * Added by S Pearson
 * Get the distance from the tuple to the target tuple
 */
static double
MJCompare_EditDist(MergeJoinState *mergestate,
		TupleTableSlot *slot1, TupleTableSlot *slot2)
{
	int result = 0;
	//Datum clauseDistance;
	ExprContext *econtext = mergestate->js.ps.ps_ExprContext;
	int			i;
	MemoryContext oldContext;

	//Load values and check for nulls.  Always will be strict.
	if(!(MJEvalNextValue(mergestate, slot1) && MJEvalNextValue(mergestate, slot2)))
		return -1;

	/*
	 * Call the comparison functions in short-lived context, in case they leak
	 * memory.
	 */
	ResetExprContext(econtext);
	oldContext = MemoryContextSwitchTo(econtext->ecxt_per_tuple_memory);

	char* lString;
	char* rString;
	int lStrLen, rStrLen;
	int clauseDistance;
	bool flag;
	for (i = 0; i < mergestate->mj_NumClauses; i++)
	{
		MergeJoinClause clause = &mergestate->mj_Clauses[i];

		lString = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(clause->ldatum)));
		rString = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(clause->rdatum)));
		if(strcmp(lString,"Ying Li") == 0 || strcmp(lString,"Ying Liu") == 0 ||
				strcmp(rString,"Ying Li") == 0 || strcmp(rString,"Ying Liu") == 0)
			flag = true;
		lStrLen = strlen(lString);
		rStrLen = strlen(rString);

		//Get rid of obviously dis-similar tuples
		if (abs(lStrLen - rStrLen) > ((MergeJoin *) mergestate->js.ps.plan)->epsilon)
		{
			MemoryContextSwitchTo(oldContext);
			return ((MergeJoin *) mergestate->js.ps.plan)->epsilon + 1;
		}

		if (mergestate->mj_NumClauses > 1)
		{
			clauseDistance = levenshteinDistance(mergestate, lString,
					lStrLen, rString, rStrLen);
			result += clauseDistance * clauseDistance;
		}
		else
			result+= levenshteinDistance(mergestate, lString, lStrLen, rString, rStrLen);
	}

	MemoryContextSwitchTo(oldContext);
	return mergestate->mj_NumClauses > 1 ? sqrt((double)result) : result;
}

/*
 * Added by S Pearson
 * Calculates the Levenshtein Distance between two Strings.
 */
int levenshteinDistance(MergeJoinState* node, char* c1, int c1Len, char* c2, int c2Len)
{
	int distance;
	int rowLen = c2Len+1;
	int i, j;
	int *d = palloc((c1Len + 1) * (c2Len + 1) * sizeof(int));
	for(i=0; i < (c1Len+1) * (c2Len+1); i++)
		d[i] = 0;
	USEMEM(node, GetMemoryChunkSpace(d));

	for(i = 1; i <= c1Len; i++)
//		//d[i,0]
		d[i * rowLen] = i;
	for(j = 1; j <= c2Len; j++)
//		//d[0,j]
		d[j] = j;


	for(i = 1; i <= c1Len; i++)
	{
		for(j = 1; j <= c2Len; j++)
		{
			d[i * rowLen + j] = fmin(
					d[(i-1) * rowLen + j] + 1, fmin(	//insertion
							d[i * rowLen + j - 1] + 1,	//deletion
							d[(i-1) * rowLen + j - 1] + (c1[i-1] == c2[j-1] ? 0 : 1)) //substitution
							);
		}
	}
	distance = d[c1Len * rowLen + c2Len];
	FREEMEM(node, GetMemoryChunkSpace(d));
	pfree(d);
	return distance;
}

/*
 * Generate a fake join tuple with nulls for the inner tuple,
 * and return it if it passes the non-join quals.
 */
static TupleTableSlot *
MJFillOuter(MergeJoinState *node)
{
	ExprContext *econtext = node->js.ps.ps_ExprContext;
	List	   *otherqual = node->js.ps.qual;

	ResetExprContext(econtext);

	econtext->ecxt_outertuple = node->mj_OuterTupleSlot;
	econtext->ecxt_innertuple = node->mj_NullInnerTupleSlot;

	if (ExecQual(otherqual, econtext, false))
	{
		/*
		 * qualification succeeded.  now form the desired projection tuple and
		 * return the slot containing it.
		 */
		TupleTableSlot *result;
		ExprDoneCond isDone;

		MJ_printf("ExecMergeJoin: returning outer fill tuple\n");

		result = ExecProject(node->js.ps.ps_ProjInfo, &isDone);

		if (isDone != ExprEndResult)
		{
			node->js.ps.ps_TupFromTlist =
					(isDone == ExprMultipleResult);
			return result;
		}
	}

	return NULL;
}

/*
 * Generate a fake join tuple with nulls for the outer tuple,
 * and return it if it passes the non-join quals.
 */
static TupleTableSlot *
MJFillInner(MergeJoinState *node)
{
	ExprContext *econtext = node->js.ps.ps_ExprContext;
	List	   *otherqual = node->js.ps.qual;

	ResetExprContext(econtext);

	econtext->ecxt_outertuple = node->mj_NullOuterTupleSlot;
	econtext->ecxt_innertuple = node->mj_InnerTupleSlot;

	if (ExecQual(otherqual, econtext, false))
	{
		/*
		 * qualification succeeded.  now form the desired projection tuple and
		 * return the slot containing it.
		 */
		TupleTableSlot *result;
		ExprDoneCond isDone;

		MJ_printf("ExecMergeJoin: returning inner fill tuple\n");

		result = ExecProject(node->js.ps.ps_ProjInfo, &isDone);

		if (isDone != ExprEndResult)
		{
			node->js.ps.ps_TupFromTlist =
					(isDone == ExprMultipleResult);
			return result;
		}
	}

	return NULL;
}

/*
 * Added by S Pearson
 * Gets the index offset for a window between two different partitions in the
 * window array
 */
int getWindowOffset(int numPartitions, int partition1, int partition2)
{
	int x, y, z; //z differentiates between the two windows
	if(partition1 < partition2)
	{
		x = partition1;
		y = partition2;
		z = 0;
	}
	else
	{
		x = partition2;
		y = partition1;
		z = 1;
	}

	return 2 * (x * numPartitions - (x * (x+1))/2 + y - (x+1)) + z;
}

/*
 * Added by S Pearson
 * Function for finding out whether a partition is in memory.
 */
bool isPartitionInMemory(MergeJoinState *node, Tuplestorestate *partition)
{
	return tuplestore_inMem(partition) && (tuplestore_availMem(partition) >= 0);
}

/*
 * Added by S Pearson
 * Function for finding out whether a window set is in memory
 * Since the available memory of a Tuplestore can be negative, we can add the
 * available memory of both Tuplestores together to determine how much total
 * available memory there is.  If the available memory is >= the partition size,
 * then the combination of both windows will be in memory.
 */
bool areWindowsInMemory(MergeJoinState *node, Tuplestorestate *win1, Tuplestorestate *win2)
{
	/*
	 * Array to modify partitionSize by amount the initial memtuples array inside
	 * a Tuplestorestate uses.  Needed to check whether a set will really be in-memory.
	 */
	void **tempArray = palloc(1024 * sizeof(void *));
	if(tuplestore_inMem(win1) && tuplestore_inMem(win2) &&
			(tuplestore_availMem(win1) + tuplestore_availMem(win2) >=
			(node->partitionSize * 1024 - GetMemoryChunkSpace(tempArray))))
	{
		pfree(tempArray);
		return true;
	}
	pfree(tempArray);
	return false;
}

/*
 * Added by S Pearson
 * Returns true if the partition has few enough tuples to go to the nested join loops.
 * This is the case if the number of tuples is smaller than the node->maxTuples value.
 */
bool isPartitionJoinReady(MergeJoinState *node, Tuplestorestate *partition)
{
	return tuplestore_numTuples(partition) <=
			(node->numPartitions >= node->maxTuples ? node->numPartitions : node->maxTuples);
}

/*
 * Added by S Pearson
 * Returns true if the windows have few enough tuples to go to the nested join loops.
 * This is the case if the sum of the number of tuples from both windows is smaller
 * than the node->maxTuples value.
 */
bool areWindowsJoinReady(MergeJoinState *node, Tuplestorestate *win1, Tuplestorestate *win2)
{
	return tuplestore_numTuples(win1) + tuplestore_numTuples(win2) <=
			(node->numPartitions >= node->maxTuples ? node->numPartitions : node->maxTuples);
}

/*
 * Added by S Pearson
 *
 *Hibernate all partitions that can be hibernated
 */
void hibernateNewPartitions(MergeJoinState *node)
{
	int i;

	for(i = node->currentPartition; i < node->oldNumPartitions; i++)
		if(!(isPartitionJoinReady(node, node->partitions[i]) ||
				isPartitionInMemory(node, node->partitions[i])))
		{
			FREEMEM(node, GetMemoryChunkSpace(node->partitions[i]));
			hibernateBufFile(node->partitions[i]);
			USEMEM(node, GetMemoryChunkSpace(node->partitions[i]));
		}
}

/*
 * Added by S Pearson
 *
 * Hibernate all windows that can be hibernated
 */
void hibernateNewWindows(MergeJoinState *node)
{
	int i;

	for(i = node->currentWindow; i < node->oldNumWindows; i+=2)
		if(!(areWindowsJoinReady(node, node->windows[i], node->windows[i+1]) ||
				areWindowsInMemory(node, node->windows[i], node->windows[i+1])))
		{
			FREEMEM(node, GetMemoryChunkSpace(node->windows[i]));
			FREEMEM(node, GetMemoryChunkSpace(node->windows[i+1]));
			hibernateBufFile(node->windows[i]);
			hibernateBufFile(node->windows[i+1]);
			USEMEM(node, GetMemoryChunkSpace(node->windows[i]));
			USEMEM(node, GetMemoryChunkSpace(node->windows[i+1]));
		}
}

/*
 * Added by S Pearson
 *
 * Hibernate all wPartitions that can be hibernated
 */
void hibernateNewWPartitions(MergeJoinState *node)
{
	int i;

	for(i = node->currentWPartition; i < node->oldNumWPartitions; i++)
		if(!(areWindowsJoinReady(node, node->wPartitions0[i], node->wPartitions1[i]) ||
				areWindowsInMemory(node, node->wPartitions0[i], node->wPartitions1[i])))
		{
			FREEMEM(node, GetMemoryChunkSpace(node->wPartitions0[i]));
			FREEMEM(node, GetMemoryChunkSpace(node->wPartitions1[i]));
			hibernateBufFile(node->wPartitions0[i]);
			hibernateBufFile(node->wPartitions1[i]);
			USEMEM(node, GetMemoryChunkSpace(node->wPartitions0[i]));
			USEMEM(node, GetMemoryChunkSpace(node->wPartitions1[i]));
		}
}

/*
 * Added by S Pearson
 *
 * Hibernate all wWindows that can be hibernated
 */
void hibernateNewWWindows(MergeJoinState *node)
{
	int i;

	for(i = node->currentWWindow; i < node->oldNumWWindows; i+=2)
	{
		//First pair of set
		if(!(areWindowsJoinReady(node, node->wWindows0[i], node->wWindows1[i+1]) ||
				areWindowsInMemory(node, node->wWindows0[i], node->wWindows1[i+1])))
		{
			FREEMEM(node, GetMemoryChunkSpace(node->wWindows0[i]));
			FREEMEM(node, GetMemoryChunkSpace(node->wWindows1[i+1]));
			hibernateBufFile(node->wWindows0[i]);
			hibernateBufFile(node->wWindows1[i+1]);
			USEMEM(node, GetMemoryChunkSpace(node->wWindows0[i]));
			USEMEM(node, GetMemoryChunkSpace(node->wWindows1[i+1]));
		}
		//Second pair of set
		if(!(areWindowsJoinReady(node, node->wWindows0[i+1], node->wWindows1[i]) ||
				areWindowsInMemory(node, node->wWindows0[i+1], node->wWindows1[i])))
		{
			FREEMEM(node, GetMemoryChunkSpace(node->wWindows0[i+1]));
			FREEMEM(node, GetMemoryChunkSpace(node->wWindows1[i]));
			hibernateBufFile(node->wWindows0[i+1]);
			hibernateBufFile(node->wWindows1[i]);
			USEMEM(node, GetMemoryChunkSpace(node->wWindows0[i+1]));
			USEMEM(node, GetMemoryChunkSpace(node->wWindows1[i]));
		}
	}
}

/*
 * Added by S Pearson
 *
 * Get random pivots from a partition.
 */
void
getPivotsFromPartition(MergeJoinState *node, Tuplestorestate *partition, int numPivots)
{
	int pivot, range, i;
	Assert(tuplestore_numTuples(partition) >= numPivots);
	range = floor(tuplestore_numTuples(partition) / numPivots);
	resetPivots(node);

	for(node->currentPivot = 0; node->currentPivot < numPivots; node->currentPivot++)
	{
		pivot = rand() % range;
		for(i = 1; i < pivot; i++)
		{
			tuplestore_advance(partition, true);
		}
		node->nextTupleSlot = getTuple(node, partition);
		node->pivots[node->currentPivot] = ExecCopySlotMinimalTuple(node->nextTupleSlot);
		USEMEM(node, GetMemoryChunkSpace(node->pivots[node->currentPivot]));
	}
	tuplestore_rescan(partition);
}

/*
 * Added by S Pearson
 *
 * Get random pivots from windows.
 */
void
getPivotsFromWindows(MergeJoinState *node, Tuplestorestate *window0, Tuplestorestate *window1, int numPivots)
{
	int pivot, range, i;
	Assert(tuplestore_numTuples(window0) + tuplestore_numTuples(window1) >= numPivots);
	range = floor((tuplestore_numTuples(window0) + tuplestore_numTuples(window1)) / numPivots);
	resetPivots(node);

	for(node->currentPivot = 0; node->currentPivot < numPivots; node->currentPivot++)
	{
		pivot = rand() % range;
		for(i = 1; i < pivot; i++)
		{
			//Switch over to traversing through the second window if the first is done
			if(!tuplestore_ateof(window0))
			{
				node->nextTupleSlot = getTuple(node, window0);
				if(node->nextTupleSlot == NULL)
					tuplestore_advance(window1, true);
			}
			else
				tuplestore_advance(window1, true);
		}
		//Make sure we grab the tuple from the second window if the first is done
		if(!tuplestore_ateof(window0))
		{
			node->nextTupleSlot = getTuple(node, window0);
			if(node->nextTupleSlot == NULL)
				node->nextTupleSlot = getTuple(node, window1);
		}
		else
			node->nextTupleSlot = getTuple(node, window1);

		node->pivots[node->currentPivot] = ExecCopySlotMinimalTuple(node->nextTupleSlot);
		USEMEM(node, GetMemoryChunkSpace(node->pivots[node->currentPivot]));
	}
	tuplestore_rescan(window0);
	tuplestore_rescan(window1);
}

/*
 * Added by S Pearson
 * Partitions a partition into the given partition and window arrays.
 * The caller must ensure these arrays have been sized and/or rebuilt properly.
 */
void partition(MergeJoinState *node, Tuplestorestate *partition,
		int numPartitions, int partitionOffset, int windowOffset,
		Tuplestorestate **partitionArray, Tuplestorestate **windowArray)
{
	int i;
	getPivotsFromPartition(node, partition, numPartitions);

	for(node->currentPivot = 0; node->currentPivot < numPartitions; node->currentPivot++)
	{
		//Make partition
		partitionArray[partitionOffset + node->currentPivot] =
				tuplestore_begin_heap(false, false, node->partitionSize);
		USEMEM(node, GetMemoryChunkSpace(
			partitionArray[partitionOffset + node->currentPivot]));

		/*
		 * Create windows
		 */
		for(i = node->currentPivot + 1; i < numPartitions; i++)
		{
			int wOffset = getWindowOffset(numPartitions,
					node->currentPivot, i) + windowOffset;
			windowArray[wOffset] = tuplestore_begin_heap(false,
					false, node->partitionSize);
			USEMEM(node, GetMemoryChunkSpace(windowArray[wOffset]));
			windowArray[wOffset + 1] = tuplestore_begin_heap(false,
					false, node->partitionSize);
			USEMEM(node, GetMemoryChunkSpace(windowArray[wOffset + 1]));
		}
	}

	//All partitions/windows created, need to add tuples to them now
	while(!tuplestore_ateof(partition))
	{
		node->nextTupleSlot = getTuple(node, partition);
		if(node->nextTupleSlot != NULL)
			addTuple(node, node->nextTupleSlot, numPartitions,
				partitionArray, partitionOffset,
				windowArray, windowOffset);
	}
}

/*
 * Added by S Pearson
 * Partitions any two windows, window-partitions, or window-windows into the given
 * wPartition and wWindow arrays.  The caller must ensure these arrays have been sized
 * and/or rebuilt properly.
 */
void
partitionWindows(MergeJoinState *node, Tuplestorestate *window0, Tuplestorestate *window1,
		int numPartitions, int wPartitionOffset, int wWindowOffset,
		Tuplestorestate **wPartitionArray0, Tuplestorestate **wPartitionArray1,
		Tuplestorestate **wWindowArray0, Tuplestorestate **wWindowArray1)
{
	int i;
	getPivotsFromWindows(node, window0, window1, numPartitions);

	for(node->currentPivot = 0; node->currentPivot < numPartitions; node->currentPivot++)
	{
		//Make wPartitions
		wPartitionArray0[wPartitionOffset + node->currentPivot] =
				tuplestore_begin_heap(false, false, node->partitionSize);
		wPartitionArray1[wPartitionOffset + node->currentPivot] =
				tuplestore_begin_heap(false, false, node->partitionSize);
		USEMEM(node, GetMemoryChunkSpace(wPartitionArray0[wPartitionOffset + node->currentPivot]));
		USEMEM(node, GetMemoryChunkSpace(wPartitionArray1[wPartitionOffset + node->currentPivot]));

		//Make wWindows
		for(i = node->currentPivot+1; i < numPartitions; i++)
		{
			int windowOffset = getWindowOffset(numPartitions, node->currentPivot, i) + wWindowOffset;
			wWindowArray0[windowOffset] =
					tuplestore_begin_heap(false, false, node->partitionSize);
			wWindowArray1[windowOffset] =
					tuplestore_begin_heap(false, false, node->partitionSize);
			USEMEM(node, GetMemoryChunkSpace(wWindowArray0[windowOffset]));
			USEMEM(node, GetMemoryChunkSpace(wWindowArray1[windowOffset]));
			wWindowArray0[windowOffset+1] =
					tuplestore_begin_heap(false,false, node->partitionSize);
			wWindowArray1[windowOffset+1] =
					tuplestore_begin_heap(false,false, node->partitionSize);
			USEMEM(node, GetMemoryChunkSpace(wWindowArray0[windowOffset+1]));
			USEMEM(node, GetMemoryChunkSpace(wWindowArray1[windowOffset+1]));
		}
	}

	/*Add tuples to partitions and windows*/
	while(!(tuplestore_ateof(window0) &&
			tuplestore_ateof(window1)))
	{
		if(!tuplestore_ateof(window0))
		{
			node->nextTupleSlot = getTuple(node, window0);
			if(node->nextTupleSlot != NULL)
				addTuple(node, node->nextTupleSlot, numPartitions, wPartitionArray0,
					wPartitionOffset, wWindowArray0, wWindowOffset);
		}
		if(!tuplestore_ateof(window1))
		{
			node->nextTupleSlot = getTuple(node, window1);
			if(node->nextTupleSlot != NULL)
				addTuple(node, node->nextTupleSlot, numPartitions, wPartitionArray1,
					wPartitionOffset, wWindowArray1, wWindowOffset);
		}
	}
}

/*
 * Added by S Pearson
 * De-allocate any remaining pivots
 */
void
resetPivots(MergeJoinState *state)
{
	int pivot;
	for(pivot = 0; pivot < state->numPartitions; pivot++)
	{
		if(state->pivots[pivot] == NULL) break;
		FREEMEM(state, GetMemoryChunkSpace(state->pivots[pivot]));
		pfree(state->pivots[pivot]);
		state->pivots[pivot] = NULL;
	}
}

/***************************************************************
 * Functions relating to adding/processing tuples in DBSimJoin *
 ***************************************************************/

/*
 * Added by S Pearson
 * Creates a new TupleDesc with a tableID column.
 * Used for DBSimJoin, where the original table ID must be stored so that
 * tuples can be correctly retrieved from Tuplestores
 */
TupleDesc initNewTupleDescTableID(TupleDesc origDesc)
{
	/*Initialize new TupleDesc*/
	TupleDesc newDesc = CreateTupleDescCopy(origDesc);
	int i;
	char	   *stg;
	int			attroffset;
	Form_pg_attribute * attrs;
	newDesc->natts++;
	attroffset = sizeof(struct tupleDesc) + newDesc->natts * sizeof(Form_pg_attribute);
	attroffset = MAXALIGN(attroffset);
	stg = palloc(attroffset + newDesc->natts * MAXALIGN(ATTRIBUTE_TUPLE_SIZE));
	attrs = (Form_pg_attribute *) (stg + sizeof(struct tupleDesc));
	stg += attroffset;
	for(i = 0; i < origDesc->natts; i++)
	{
		attrs[i] = origDesc->attrs[i];
		stg += MAXALIGN(ATTRIBUTE_TUPLE_SIZE);
	}
	attrs[newDesc->natts-1] = (Form_pg_attribute) stg;
	newDesc->attrs = attrs;
	TupleDescInitEntry(newDesc, newDesc->natts, "TableID", INT4OID, -1, 0);
	return newDesc;
}

/*
 * Added by S Pearson
 * Adds the TableID column to the tuple.
 * Assume the correct TupleDesc has already been applied to the newSlot
 */
void
initTupleFromPlan(TupleTableSlot *origSlot, TupleTableSlot *newSlot, bool isOuter)
{
	HeapTuple hTup = ExecMaterializeSlot(origSlot);
	//Deform the tuple inside the TupleTableSlot
	heap_deform_tuple(hTup, newSlot->tts_tupleDescriptor,
			newSlot->tts_values, newSlot->tts_isnull);
	//Set the end value of the values/isNull arrays to the proper tableID
	newSlot->tts_values[newSlot->tts_tupleDescriptor->natts - 1] = isOuter ? Int32GetDatum(OUTER_PLAN) : Int32GetDatum(INNER_PLAN);
	newSlot->tts_isnull[newSlot->tts_tupleDescriptor->natts - 1] = false;

	//Reform the tuple with the new values
	hTup = heap_form_tuple(newSlot->tts_tupleDescriptor, newSlot->tts_values, newSlot->tts_isnull);
	ExecStoreTuple(hTup, newSlot, InvalidBuffer, origSlot->tts_shouldFree);
}

/*
 * Added by S Pearson
 * Function to get a tuple and determine its original table from a Tuplestore.
 * Returns a valid TupleTableSlot pointer if successful, NULL otherwise.
 */
TupleTableSlot *
getTuple(MergeJoinState *mergestate, Tuplestorestate *partition)
{
	bool shouldFree, outer;
	TupleTableSlot *slot;

	/*
	 * Get the tuple, put it in a slot.  The TupleDesc of the slot might not
	 * be correct, so we have to fix it.
	 */
	if(tuplestore_gettupleslot(partition, true, mergestate->tempSlot))
	{
		shouldFree = mergestate->tempSlot->tts_shouldFree;
		/*
		 * Make sure something didn't go drastically wrong.
		 * At least one TupleDesc must have the same natts as the tuple
		 */
		Assert(mergestate->tempSlot->tts_mintuple->t_natts == mergestate->newOuterDesc->natts ||
			mergestate->tempSlot->tts_mintuple->t_natts == mergestate->newInnerDesc->natts);
		/*
		 * Check possible TupleDescs to see if one is a match in length.  If only one
		 * matches in length, this should be the correct TupleDesc.
		 * If both match in length, we can then check which table
		 * the tuple originally came from as both TupleDescs should now include a Table ID column at the end.
		 */
		if(mergestate->newOuterDesc->natts == mergestate->newInnerDesc->natts)
			/*
			 * Either TupleDesc will have the correct length to access the Table ID
			 * column, even if it isn't the correct TupleDesc for that tuple.
			 */
			outer = isOuter(mergestate->tempSlot);
		/*
		 * If the tuple only matches one TupleDesc in length, that TupleDesc is
		 * the correct one for this tuple.
		 * Set the slot's TupleDesc to that one.
		 */
		else outer = mergestate->tempSlot->tts_mintuple->t_natts == mergestate->newOuterDesc->natts;

		slot = outer ? mergestate->mj_OuterTupleSlot : mergestate->mj_InnerTupleSlot;
		//Store the tuple in the slot
		ExecStoreMinimalTuple(ExecCopySlotMinimalTuple(mergestate->tempSlot), slot, shouldFree);
		return slot;
	}
	return NULL;
}

/*
 * Added by S Pearson
 * Same as above, except used for the inner nested loops in DBSimJoin.
 * Therefore, we might end up with two values from the same plan, so we need to add
 * into different slots than used in the previous function.
 */
TupleTableSlot *
getNextTuple(MergeJoinState *mergestate, Tuplestorestate *partition)
{
	bool shouldFree, outer;
	TupleTableSlot *slot;

	/*
	 * Get the tuple, put it in a slot.  The TupleDesc of the slot might not
	 * be correct, so we have to fix it.
	 */
	if(tuplestore_gettupleslot(partition, true, mergestate->tempSlot))
	{
		shouldFree = mergestate->tempSlot->tts_shouldFree;
		/*
		 * Make sure something didn't go drastically wrong.
		 * At least one TupleDesc must have the same natts as the tuple
		 */
		Assert(mergestate->tempSlot->tts_mintuple->t_natts == mergestate->newOuterDesc->natts ||
			mergestate->tempSlot->tts_mintuple->t_natts == mergestate->newInnerDesc->natts);
		/*
		 * Check possible TupleDescs to see if one is a match in length.  If only one
		 * matches in length, this should be the correct TupleDesc.
		 * If both match in length, we can then check which table
		 * the tuple originally came from as both TupleDescs should now include a Table ID column at the end.
		 */
		if(mergestate->newOuterDesc->natts == mergestate->newInnerDesc->natts)
			/*
			 * Either TupleDesc will have the correct length to access the Table ID
			 * column, even if it isn't the correct TupleDesc for that tuple.
			 */
			outer = isOuter(mergestate->tempSlot);
		/*
		 * If the tuple only matches one TupleDesc in length, that TupleDesc is
		 * the correct one for this tuple.
		 * Set the slot's TupleDesc to that one.
		 */
		else outer = mergestate->tempSlot->tts_mintuple->t_natts == mergestate->newOuterDesc->natts;

		slot = outer ? mergestate->extraOuter : mergestate->extraInner;

		//Store the tuple in the slot
		ExecStoreMinimalTuple(ExecCopySlotMinimalTuple(mergestate->tempSlot), slot, shouldFree);

		return slot;
	}
	return NULL;
}

/*
 * Added by S Pearson
 * Gets the proper TupleDesc for the pivot so that we can get the correct attributes
 * from it.
 * Pivot must already have been transformed into a HeapTuple.
 */
bool pivotIsOuter(MergeJoinState *mergestate, HeapTuple pivot)
{
	bool shouldFree = false;
	/*
	 * Make sure something didn't go drastically wrong.
	 * At least one TupleDesc must have the same natts as the tuple
	 */
	Assert(pivot->t_data->t_natts == mergestate->newOuterDesc->natts ||
		pivot->t_data->t_natts == mergestate->newInnerDesc->natts);
	/*
	 * Check possible TupleDescs to see if one is a match in length.  If only one
	 * matches in length, this should be the correct TupleDesc.
	 * If both match in length, we can then check which table
	 * the tuple originally came from as both TupleDescs should now include a Table ID column at the end.
	 */
	return mergestate->newOuterDesc->natts == mergestate->newInnerDesc->natts ?
			/*
			 * Either TupleDesc will have the correct length to access the Table ID column,
			 * even if it isn't the correct TupleDesc for that tuple.
			 */
			(DatumGetInt32(heap_getattr(pivot, mergestate->newOuterDesc->natts,
				mergestate->newOuterDesc, &shouldFree)) == OUTER_PLAN?
				true : false) :
			/*
			 * If the tuple only matches one TupleDesc in length, that TupleDesc is the
			 * correct one for this tuple.
			 * Set the slot's TupleDesc to that one.
			 */
			pivot->t_data->t_natts == mergestate->newOuterDesc->natts ?
				true : false;
}

/*
 * Added by S Pearson
 * Determines whether a given TupleTableSlot is from the outer or inner plan
 * for the purposes of determining the correct TupleDesc or the original table
 * for tuples.
 * This should only be used for DBSimJoin, so we can assume tuples will have a
 * TableID attribute at the end.
 */
bool isOuter(TupleTableSlot *tuple)
{
	return DatumGetInt32(slot_getattr(tuple, tuple->tts_tupleDescriptor->natts,
			&tuple->tts_isnull[tuple->tts_tupleDescriptor->natts - 1])) == OUTER_PLAN;
}

/*
 * Added by S Pearson
 * Returns true if the tuples are from the same table.
 * This should only be used for DBSimJoin, so we can assume tuples will have a
 * TableID attribute at the end.
 */
bool fromSameTable(TupleTableSlot *slot1, TupleTableSlot *slot2)
{
	return isOuter(slot1) == isOuter(slot2);
}

/*
 * Added by S Pearson
 * Takes a given tuple and adds it to the proper partition and
 * windows, assuming pivots are already chosen.
 *
 * tuple - the tuple to add to the partition and windows
 * numPivots - the number of pivots we're going to make
 * partitionArray - the array of partitions to add to
 * partitionOffset - the index to offset the partition array by
 * windowArray - the array of windows to add to
 * numOldWindows - the index to offset the window array by
 */
void
addTuple(MergeJoinState *mergestate, TupleTableSlot *tuple, int numPivots,
		Tuplestorestate **partitionArray, int partitionOffset,
		Tuplestorestate **windowArray, int windowOffset)
{
	double minDist = DBL_MAX;
	double distPivot, distWindow;
	int i, minPivot;
	for(i = 0; i < numPivots; i++)
	{
		distPivot =
				//MJCompare_EditDistPivot	//Uncomment this line for String data
				MJCompare_DistFuncPivot		//Uncomment this line for vector data
				(mergestate, tuple,
				mergestate->pivots[i]);
		if(distPivot >= 0 && distPivot < minDist)
		{
			minDist = distPivot;
			minPivot = i;
		}
	}
	tuplestore_puttupleslot(partitionArray[partitionOffset + minPivot],
			tuple);

	/*Add to windows*/
	for(i = 0; i < numPivots; i++)
	{
		if(i != minPivot)
		{
			distWindow =
					//MJCompare_EditDistPivot	//Uncomment this line for String data
					MJCompare_DistFuncPivot		//Uncomment this line for vector data
					(mergestate, tuple,
					mergestate->pivots[i]);
			if(distWindow >= 0 && (distWindow - minDist)/2 <=
					((MergeJoin *)mergestate->js.ps.plan)->epsilon)
			{
				tuplestore_puttupleslot(
					windowArray[getWindowOffset(numPivots, minPivot, i) + windowOffset],
					tuple);
			}
		}
	}
}

/*
 * Added by S Pearson
 * Puts the tuple stored in the slot into the proper return slot.
 * We will only ever be returning tuples from different plans, so we don't
 * have to worry about accidentally overwriting a slot.
 */
void resetTuple(MergeJoinState *mergestate, TupleTableSlot *slot1)
{
	HeapTuple hTup = heap_copytuple(ExecMaterializeSlot(slot1));
	bool outer = isOuter(slot1);
	//Deform the tuple inside the slot
	heap_deform_tuple(hTup, slot1->tts_tupleDescriptor, slot1->tts_values, slot1->tts_isnull);
	//Reform tuple with the original TupleDescriptor.  This should remove the TableID column.
	hTup = heap_form_tuple(outer ? mergestate->returnOuter->tts_tupleDescriptor :
		mergestate->returnInner->tts_tupleDescriptor, slot1->tts_values, slot1->tts_isnull);
	//Put into proper slot
	ExecStoreTuple(hTup, outer ? mergestate->returnOuter : mergestate->returnInner,
			slot1->tts_buffer, slot1->tts_shouldFree);
}

/*************************************
 * End of tuple-processing functions *
 *************************************/

/* ----------------------------------------------------------------
 *		ExecMergeTupleDump
 *
 *		This function is called through the MJ_dump() macro
 *		when EXEC_MERGEJOINDEBUG is defined
 * ----------------------------------------------------------------
 */
#ifdef EXEC_MERGEJOINDEBUG

static void
ExecMergeTupleDumpOuter(MergeJoinState *mergestate)
{
	TupleTableSlot *outerSlot = mergestate->mj_OuterTupleSlot;

	printf("==== outer tuple ====\n");
	if (TupIsNull(outerSlot))
		printf("(nil)\n");
	else
		MJ_debugtup(outerSlot);
}

static void
ExecMergeTupleDumpInner(MergeJoinState *mergestate)
{
	TupleTableSlot *innerSlot = mergestate->mj_InnerTupleSlot;

	printf("==== inner tuple ====\n");
	if (TupIsNull(innerSlot))
		printf("(nil)\n");
	else
		MJ_debugtup(innerSlot);
}

static void
ExecMergeTupleDumpMarked(MergeJoinState *mergestate)
{
	TupleTableSlot *markedSlot = mergestate->mj_MarkedTupleSlot;

	printf("==== marked tuple ====\n");
	if (TupIsNull(markedSlot))
		printf("(nil)\n");
	else
		MJ_debugtup(markedSlot);
}

static void
ExecMergeTupleDump(MergeJoinState *mergestate)
{
	printf("******** ExecMergeTupleDump ********\n");

	ExecMergeTupleDumpOuter(mergestate);
	ExecMergeTupleDumpInner(mergestate);
	ExecMergeTupleDumpMarked(mergestate);

	printf("******** \n");
}
#endif

/* ----------------------------------------------------------------
 *		ExecMergeJoin
 * ----------------------------------------------------------------
 */
TupleTableSlot *
ExecMergeJoin(MergeJoinState *node)
{
	EState	   *estate;
	List	   *joinqual;
	List	   *otherqual;
	bool		qualResult;
	int			compareResult;
	int			compareResult2; /*Changed by Yasin SJ*/
	PlanState  *innerPlan;
	TupleTableSlot *innerTupleSlot;
	PlanState  *outerPlan;
	TupleTableSlot *outerTupleSlot;
	ExprContext *econtext;
	bool		doFillOuter;
	bool		doFillInner;

	//Added by S Pearson
	Tuplestorestate **newPartitions, **newWPartitions0, **newWPartitions1,
					**newWWindows0, **newWWindows1;
	int i, newIndex;
	int windowOffset;
	double epsilonDist;
	epsilonDist = ((MergeJoin *) node->js.ps.plan)->epsilon;

	/*
	 * get information from node
	 */
	estate = node->js.ps.state;
	innerPlan = innerPlanState(node);
	outerPlan = outerPlanState(node);
	econtext = node->js.ps.ps_ExprContext;
	joinqual = node->js.joinqual;
	otherqual = node->js.ps.qual;
	doFillOuter = node->mj_FillOuter;
	doFillInner = node->mj_FillInner;


	/*
	 * Check to see if we're still projecting out tuples from a previous join
	 * tuple (because there is a function-returning-set in the projection
	 * expressions).  If so, try to project another one.
	 */
	if (node->js.ps.ps_TupFromTlist)
	{
		TupleTableSlot *result;
		ExprDoneCond isDone;

		result = ExecProject(node->js.ps.ps_ProjInfo, &isDone);
		if (isDone == ExprMultipleResult)
			return result;
		/* Done with that source tuple... */
		node->js.ps.ps_TupFromTlist = false;
	}

	/*
	 * Reset per-tuple memory context to free any expression evaluation
	 * storage allocated in the previous tuple cycle.  Note this can't happen
	 * until we're done projecting out tuples from a join tuple.
	 */
	ResetExprContext(econtext);

	/*
	 * ok, everything is setup.. let's go to work
	 */
	for (;;)
	{
		MJ_dump(node);
		/*
		 * get the current state of the join and do things accordingly.
		 */
		switch (node->mj_JoinState)
		{

		/****************************
		 * DBSimJoin specific cases *
		 * Added by S Pearson		*
		 ****************************/

		/*XXX EXEC_DBSIMJOIN_INIT
		 * First time ExecMergeJoin() has been called.
		 * Get pivots, build the partitions and windows.
		 */
		case EXEC_DBSIMJOIN_INIT:
		{
			outerTupleSlot = ExecProcNode(outerPlan);
			if(TupIsNull(outerTupleSlot))
			{
				/*If outer subplan exhausted, select pivots from inner*/
				innerTupleSlot = ExecProcNode(innerPlan);
				if(TupIsNull(innerTupleSlot))
				{
					/*
					 * Everything exhausted, but not all pivots used.
					 * small enough to solve with nested-loop
					 * already
					 */
					ExecReScan(innerPlan, innerPlan->ps_ExprContext);
					ExecReScan(outerPlan, outerPlan->ps_ExprContext);
					/*
					 * De-allocate all partitions/windows/pivots made.  We don't need them
					 * for small input, we're going straight to a nestloop
					 */
					for(i = 0; i < node->currentPivot; i++)
					{
						int j;
						FREEMEM(node, GetMemoryChunkSpace(node->pivots[i]));
						pfree(node->pivots[i]);
						FREEMEM(node, GetMemoryChunkSpace(node->partitions[i]));
						tuplestore_end(node->partitions[i]);
						pfree(node->partitions[i]);
						for(j = i + 1; j < node->numPartitions; j++)
						{
							windowOffset = getWindowOffset(node->numPartitions, i, j);
							FREEMEM(node, GetMemoryChunkSpace(node->windows[windowOffset]));
							FREEMEM(node, GetMemoryChunkSpace(node->windows[windowOffset + 1]));
							tuplestore_end(node->windows[windowOffset]);
							tuplestore_end(node->windows[windowOffset + 1]);
							pfree(node->windows[windowOffset]);
							pfree(node->windows[windowOffset + 1]);
						}
					}
					pfree(node->pivots);
					pfree(node->partitions);
					pfree(node->windows);
					node->partitionToJoin0 = tuplestore_begin_heap(false, false, node->partitionSize);
					node->mj_JoinState = EXEC_DBSIMJOIN_INIT_SMALL_INPUT_NESTLOOP;
					break;
				}
				initTupleFromPlan(innerTupleSlot, node->mj_InnerTupleSlot, false);

				/*
				 * Create MinimalTuple to store pivot in memory
				 */
				node->pivots[node->currentPivot] = ExecCopySlotMinimalTuple(node->mj_InnerTupleSlot);
				USEMEM(node, GetMemoryChunkSpace(node->pivots[node->currentPivot]));
			}
			else
			{
				initTupleFromPlan(outerTupleSlot, node->mj_OuterTupleSlot, true);

				/*
				 * Create MinimalTuple to store pivot in memory
				 */
				node->pivots[node->currentPivot] = ExecCopySlotMinimalTuple(node->mj_OuterTupleSlot);
				USEMEM(node, GetMemoryChunkSpace(node->pivots[node->currentPivot]));
			}
			/*
			 * Create partition
			 */
			node->partitions[node->currentPivot] = tuplestore_begin_heap(false,
					false, node->partitionSize);
			USEMEM(node, GetMemoryChunkSpace(node->partitions[node->currentPivot]));
			/*
			 * Create the windows
			 */
			for(i=node->currentPivot+1; i < node->numPartitions; i++)
			{
				windowOffset = getWindowOffset(node->numPartitions, node->currentPivot, i);
				node->windows[windowOffset] = tuplestore_begin_heap(false,
						false, node->partitionSize);
				USEMEM(node, GetMemoryChunkSpace(node->windows[windowOffset]));
				node->windows[windowOffset+1] = tuplestore_begin_heap(false,
						false, node->partitionSize);
				USEMEM(node, GetMemoryChunkSpace(node->windows[windowOffset+1]));
			}

			/*
			 * If all partitions and windows have been created, move on to adding
			 * tuples into them.
			 *
			 * Reset the nodes to the marked values, as we have to re-traverse
			 * the pivots we already selected in order to add them to the correct
			 * partitions and windows.
			 */
			if(++node->currentPivot >= node->numPartitions) {
				node->mj_JoinState = EXEC_DBSIMJOIN_NEXT_OUTER;
				ExecReScan(innerPlan, innerPlan->ps_ExprContext);
				ExecReScan(outerPlan, outerPlan->ps_ExprContext);
			}
		}
		break;

		/*XXX EXEC_DBSIMJOIN_NEXT_OUTER
		 * Add tuples in the outer plan into the partitions/windows
		 */
		case EXEC_DBSIMJOIN_NEXT_OUTER:
		{
			outerTupleSlot = ExecProcNode(outerPlan);
			if(TupIsNull(outerTupleSlot))
			{
				/*Done with outer subplan*/
				node->mj_JoinState = EXEC_DBSIMJOIN_NEXT_INNER;
				break;
			}
			initTupleFromPlan(outerTupleSlot, node->mj_OuterTupleSlot, true);
			addTuple(node, node->mj_OuterTupleSlot, node->numPartitions,
					node->partitions, 0, node->windows, 0);
		}
		break;

		/*XXX EXEC_DBSIMJOIN_NEXT_INNER
		 * Add tuples in the inner plan into partitions/windows
		 */
		case EXEC_DBSIMJOIN_NEXT_INNER:
		{
			innerTupleSlot = ExecProcNode(innerPlan);
			if(TupIsNull(innerTupleSlot))
			{
				/*Done with inner subplan*/
				node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_PROCESS_PARTITIONS;

				hibernateNewPartitions(node);
				hibernateNewWindows(node);

				break;
			}
			initTupleFromPlan(innerTupleSlot, node->mj_InnerTupleSlot, false);
			addTuple(node, node->mj_InnerTupleSlot, node->numPartitions,
					node->partitions, 0, node->windows, 0);
		}
		break;

		/*XXX EXEC_DBSIMJOIN_EXTER_PROCESS_PARTITIONS
		 * Finds partitions that can currently be completed
		 * in-memory and joins them
		 */
		case EXEC_DBSIMJOIN_EXTERN_PROCESS_PARTITIONS:
		{
			/*
			 * Check to see if we've processed all new partitions that
			 * we can process in-memory
			 */
			if(node->currentPartition < node->oldNumPartitions)
			{
				/*
				 * Test for whether or not the partition is too small to have a next iteration.
				 * If so, go immediately to nested loop join.
				 */
				if(isPartitionJoinReady(node, node->partitions[node->currentPartition]))
				{
					//Small enough to join
					node->partitionToJoin0 = node->partitions[node->currentPartition];
					node->partitions[node->currentPartition++] = NULL;
					node->numPartitionsLeft--;
					//Perform nest-loop on the partition
					node->returnState = EXEC_DBSIMJOIN_EXTERN_PROCESS_PARTITIONS;
					node->mj_JoinState = EXEC_DBSIMJOIN_INIT_NESTLOOP;
					break;
				}
				//Test for whether to go into internal-memory algorithm
				else if(isPartitionInMemory(node, node->partitions[node->currentPartition]))
				{
					/*
					 * Build arrays for internal-memory portion of the algorithm.
					 * Always use only 2 partitions/windows for internal-memory.
					 */
					node->internPartitions = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * 2);
					node->internWindows = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * 2);
					USEMEM(node, GetMemoryChunkSpace(node->internPartitions));
					USEMEM(node, GetMemoryChunkSpace(node->internWindows));

					partition(node, node->partitions[node->currentPartition], 2, 0, 0,
							node->internPartitions, node->internWindows);

					//Done with partition, set it to NULL
					FREEMEM(node, GetMemoryChunkSpace(node->partitions[node->currentPartition]));
					tuplestore_end(node->partitions[node->currentPartition]);
					pfree(node->partitions[node->currentPartition]);
					node->partitions[node->currentPartition++] = NULL;
					node->numPartitionsLeft--;

					//Set up counter variables
					node->oldNumInternPartitions = 2;
					node->numInternPartitionsLeft = 2;
					node->currentInternPartition = 0;

					node->oldNumInternWindows = 2;
					node->numInternWindowsLeft = 2;
					node->currentInternWindow = 0;

					//Start internal-memory algorithm
					node->internReturnState = EXEC_DBSIMJOIN_EXTERN_PROCESS_PARTITIONS;
					node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_PARTITIONS;
					break;
				}
				else
					//Go to next partition
					node->currentPartition++;
			}
			else
				node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOWS;
		}
		break;

		/*XXX EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOWS
		 * Process all windows that can be done in memory
		 * Then move on to processing windows externally
		 */
		case EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOWS:
		{
			if(node->numWindowsLeft > 0)
			{
				//Check to see if we've processed all new windows
				if(node->currentWindow < node->oldNumWindows)
				{
					/*
					 * Test for whether or not the windows are too small to have a next
					 * iteration.  If so, go immediately to nested loop join.
					 */
					if(areWindowsJoinReady(node, node->windows[node->currentWindow],
							node->windows[node->currentWindow + 1]))
					{
						//Small enough to join
						node->partitionToJoin0 = node->windows[node->currentWindow];
						node->windows[node->currentWindow++] = NULL;
						node->partitionToJoin1 = node->windows[node->currentWindow];
						node->windows[node->currentWindow++] = NULL;
						node->numWindowsLeft-= 2;
						//Perform nest-loop on the windows
						node->returnState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOWS;
						node->mj_JoinState = EXEC_DBSIMJOIN_INIT_WINLOOP;
						break;
					}
					//Test for whether to go into internal-memory algorithm
					else if(areWindowsInMemory(node, node->windows[node->currentWindow],
							node->windows[node->currentWindow + 1]))
					{
						/*
						 * Build arrays for internal-memory portion of the algorithm.
						 * Always use only 2 partitions/windows for internal-memory.
						 */
						node->internWPartitions0 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						node->internWPartitions1 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						node->internWWindows0 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						node->internWWindows1 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						USEMEM(node, GetMemoryChunkSpace(node->internWPartitions0));
						USEMEM(node, GetMemoryChunkSpace(node->internWPartitions1));
						USEMEM(node, GetMemoryChunkSpace(node->internWWindows0));
						USEMEM(node, GetMemoryChunkSpace(node->internWWindows1));

						partitionWindows(node, node->windows[node->currentWindow],
								node->windows[node->currentWindow + 1], 2, 0, 0,
								node->internWPartitions0, node->internWPartitions1,
								node->internWWindows0, node->internWWindows1);

						//Done with windows, set them to NULL
						FREEMEM(node, GetMemoryChunkSpace(node->windows[node->currentWindow]));
						tuplestore_end(node->windows[node->currentWindow]);
						pfree(node->windows[node->currentWindow]);
						node->windows[node->currentWindow++] = NULL;

						FREEMEM(node, GetMemoryChunkSpace(node->windows[node->currentWindow]));
						tuplestore_end(node->windows[node->currentWindow]);
						pfree(node->windows[node->currentWindow]);
						node->windows[node->currentWindow++] = NULL;
						node->numWindowsLeft-= 2;

						//Set up counter variables
						node->oldNumInternWPartitions = 2;
						node->numInternWPartitionsLeft = 2;
						node->currentInternWPartition = 0;

						node->oldNumInternWWindows = 2;
						node->numInternWWindowsLeft = 2;
						node->currentInternWWindow = 0;

						//Start internal-memory algorithm
						node->internReturnState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOWS;
						node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_PARTITIONS;
						break;
					}
					else
					{
						//Go to next window set
						node->currentWindow+=2;
					}
				}
				else
				{
					Tuplestorestate *tempWindow0, *tempWindow1;
					//De-hibernate the first not-null window set so we can work on it.
					for(i = 0;;i+=2)
						if(node->windows[i] != NULL && node->windows[i+1] != NULL)
						{
							tempWindow0 = node->windows[i];
							node->windows[i++] = NULL;
							tempWindow1 = node->windows[i];
							node->windows[i++] = NULL;
							node->numWindowsLeft-=2;
							break;
						}

					//De-hibernate the windows
					FREEMEM(node, GetMemoryChunkSpace(tempWindow0));
					FREEMEM(node, GetMemoryChunkSpace(tempWindow1));
					reenableBufFile(tempWindow0);
					reenableBufFile(tempWindow1);
					USEMEM(node, GetMemoryChunkSpace(tempWindow0));
					USEMEM(node, GetMemoryChunkSpace(tempWindow1));

					/*Set up arrays for the wWindow iteration*/
					//wPartition arrays
					node->wPartitions0 = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * node->numPartitions);
					node->wPartitions1 = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * node->numPartitions);
					USEMEM(node, GetMemoryChunkSpace(node->wPartitions0));
					USEMEM(node, GetMemoryChunkSpace(node->wPartitions1));

					//wWindowArrays
					node->wWindows0 = (Tuplestorestate**)
							palloc(sizeof(Tuplestorestate *)* node->numWindows);
					node->wWindows1 = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * node->numWindows);
					USEMEM(node, GetMemoryChunkSpace(node->wWindows0));
					USEMEM(node, GetMemoryChunkSpace(node->wWindows1));

					partitionWindows(node, tempWindow0, tempWindow1, node->numPartitions,
							0,0, node->wPartitions0, node->wPartitions1, node->wWindows0,
							node->wWindows1);

					//Remove first window set, we're done with it
					FREEMEM(node, GetMemoryChunkSpace(tempWindow0));
					FREEMEM(node, GetMemoryChunkSpace(tempWindow1));
					tuplestore_end(tempWindow0);
					tuplestore_end(tempWindow1);
					pfree(tempWindow0);
					pfree(tempWindow1);

					//Initialize variables for window processing
					node->oldNumWPartitions = node->numPartitions;
					node->numWPartitionsLeft = node->numPartitions;
					node->currentWPartition = 0;

					node->oldNumWWindows = node->numWindows;
					node->numWWindowsLeft = node->numWindows;
					node->currentWWindow = 0;

					hibernateNewWPartitions(node);
					hibernateNewWWindows(node);

					node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_PARTITIONS;
				}
			}
			/*
			 * All windows processed.  Go on to the next external partition
			 * iteration
			 */
			else
			{
				FREEMEM(node, GetMemoryChunkSpace(node->windows));
				pfree(node->windows);
				node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_ITERATE_PARTITIONS;
			}
		}
		break;

		/*XXX EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_PARTITIONS
		 * Process window-partitions that can be done in-memory
		 */
		case EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_PARTITIONS:
		{
			//Check to see if we've processed all new wPartitions
			if(node->currentWPartition < node->oldNumWPartitions)
			{
				/*
				 * Test for whether or not the wPartitions are too small to have a next
				 * iteration.  If so, go immediately to nested loop join.
				 */
				if(areWindowsJoinReady(node, node->wPartitions0[node->currentWPartition],
						node->wPartitions1[node->currentWPartition]))
				{
					//Small enough to join
					node->partitionToJoin0 = node->wPartitions0[node->currentWPartition];
					node->partitionToJoin1 = node->wPartitions1[node->currentWPartition];
					node->wPartitions0[node->currentWPartition] = NULL;
					node->wPartitions1[node->currentWPartition++] = NULL;
					node->numWPartitionsLeft--;
					//Perform nest-loop on the wPartition
					node->returnState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_PARTITIONS;
					node->mj_JoinState = EXEC_DBSIMJOIN_INIT_WINLOOP;
					break;
				}
				//Test for whether to go into internal-memory algorithm
				else if(areWindowsInMemory(node, node->wPartitions0[node->currentWPartition],
						node->wPartitions1[node->currentWPartition]))
				{
					/*
					 * Build arrays for internal-memory portion of the algorithm.
					 * Always use only 2 partitions/windows for internal-memory.
					 */
					node->internWPartitions0 = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * 2);
					node->internWPartitions1 = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * 2);
					node->internWWindows0 = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * 2);
					node->internWWindows1 = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * 2);
					USEMEM(node, GetMemoryChunkSpace(node->internWPartitions0));
					USEMEM(node, GetMemoryChunkSpace(node->internWPartitions1));
					USEMEM(node, GetMemoryChunkSpace(node->internWWindows0));
					USEMEM(node, GetMemoryChunkSpace(node->internWWindows1));

					partitionWindows(node, node->wPartitions0[node->currentWPartition],
							node->wPartitions1[node->currentWPartition], 2, 0, 0,
							node->internWPartitions0, node->internWPartitions1,
							node->internWWindows0, node->internWWindows1);

					//Done with wPartitions, set them to NULL
					FREEMEM(node, GetMemoryChunkSpace(node->wPartitions0[node->currentWPartition]));
					FREEMEM(node, GetMemoryChunkSpace(node->wPartitions1[node->currentWPartition]));
					tuplestore_end(node->wPartitions0[node->currentWPartition]);
					tuplestore_end(node->wPartitions1[node->currentWPartition]);
					pfree(node->wPartitions0[node->currentWPartition]);
					pfree(node->wPartitions1[node->currentWPartition]);
					node->wPartitions0[node->currentWPartition] = NULL;
					node->wPartitions1[node->currentWPartition++] = NULL;
					node->numWPartitionsLeft--;

					//Set up counter variables
					node->oldNumInternWPartitions = 2;
					node->numInternWPartitionsLeft = 2;
					node->currentInternWPartition = 0;

					node->oldNumInternWWindows = 2;
					node->numInternWWindowsLeft = 2;
					node->currentInternWWindow = 0;

					//Start internal-memory algorithm
					node->internReturnState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_PARTITIONS;
					node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_PARTITIONS;
					break;
				}
				else
				{
					//Go to next wPartition
					node->currentWPartition++;
				}
			}
			/*
			 * Check to see if we've processed all new wPartitions that
			 * we can process in-memory
			 */
			else
				node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_WINDOWS;
		}
		break;

		/*XXX EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_WINDOWS
		 * Process all window-windows that can be done in-memory
		 */
		case EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_WINDOWS:
		{
			if(node->currentWWindow < node->oldNumWWindows)
			{
				//First pair of the set has not already been processed
				if((node->wWindows0[node->currentWWindow] != NULL &&
					node->wWindows1[node->currentWWindow + 1] != NULL) &&
					(!isHibernated(node->wWindows0[node->currentWWindow]) &&
					!isHibernated(node->wWindows1[node->currentWWindow + 1])))
				{
					/*
					 * Test for whether or not the first pair of the window set is too
					 * small to have a next iteration.  If so, go immediately to nested loop join.
					 */
					if(areWindowsJoinReady(node, node->wWindows0[node->currentWWindow],
							node->wWindows1[node->currentWWindow + 1]))
					{
						//Small enough to join
						node->partitionToJoin0 = node->wWindows0[node->currentWWindow];
						node->partitionToJoin1 = node->wWindows1[node->currentWWindow + 1];
						node->wWindows0[node->currentWWindow] = NULL;
						node->wWindows1[node->currentWWindow + 1] = NULL;
						//Perform nest-loop on the wWindows
						node->returnState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_WINDOWS;
						node->mj_JoinState = EXEC_DBSIMJOIN_INIT_WINLOOP;
						break;
					}
					//Test for whether to go into internal-memory algorithm
					else if(areWindowsInMemory(node, node->wWindows0[node->currentWWindow],
						node->wWindows1[node->currentWWindow + 1]))
					{
						/*
						 * Build arrays for internal-memory portion of the algorithm.
						 * Always use only 2 partitions/windows for internal-memory.
						 */
						node->internWPartitions0 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						node->internWPartitions1 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						node->internWWindows0 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						node->internWWindows1 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						USEMEM(node, GetMemoryChunkSpace(node->internWPartitions0));
						USEMEM(node, GetMemoryChunkSpace(node->internWPartitions1));
						USEMEM(node, GetMemoryChunkSpace(node->internWWindows0));
						USEMEM(node, GetMemoryChunkSpace(node->internWWindows1));

						partitionWindows(node, node->wWindows0[node->currentWWindow],
								node->wWindows1[node->currentWWindow + 1], 2, 0, 0,
								node->internWPartitions0, node->internWPartitions1,
								node->internWWindows0, node->internWWindows1);

						/*
						 * Done with wWindows, set them to null. But we can't decrement the
						 * counter for number left until we're sure the whole set has been processed
						 */
						FREEMEM(node, GetMemoryChunkSpace(node->wWindows0[node->currentWWindow]));
						FREEMEM(node, GetMemoryChunkSpace(node->wWindows1[node->currentWWindow + 1]));
						tuplestore_end(node->wWindows0[node->currentWWindow]);
						tuplestore_end(node->wWindows1[node->currentWWindow + 1]);
						pfree(node->wWindows0[node->currentWWindow]);
						pfree(node->wWindows1[node->currentWWindow + 1]);
						node->wWindows0[node->currentWWindow] = NULL;
						node->wWindows1[node->currentWWindow + 1] = NULL;

						//Set up counter variables
						node->oldNumInternWPartitions = 2;
						node->numInternWPartitionsLeft = 2;
						node->currentInternWPartition = 0;

						node->oldNumInternWWindows = 2;
						node->numInternWWindowsLeft = 2;
						node->currentInternWWindow = 0;

						//Start internal-memory algorithm
						node->internReturnState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_WINDOWS;
						node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_PARTITIONS;
						break;
					}
				}
				//Second pair of the set has not already been processed
				if((node->wWindows0[node->currentWWindow + 1] != NULL &&
					node->wWindows1[node->currentWWindow] != NULL) &&
					(!isHibernated(node->wWindows0[node->currentWWindow  + 1]) &&
					!isHibernated(node->wWindows1[node->currentWWindow])))
				{
					/*
					 * Test for whether or not the second pair of the window set is too
					 * small to have a next iteration.  If so, go immediately to nested loop join.
					 */
					if(areWindowsJoinReady(node, node->wWindows0[node->currentWWindow + 1],
							node->wWindows1[node->currentWWindow]))
					{
						//Small enough to join
						node->partitionToJoin0 = node->wWindows0[node->currentWWindow + 1];
						node->partitionToJoin1 = node->wWindows1[node->currentWWindow ];
						node->wWindows0[node->currentWWindow + 1] = NULL;
						node->wWindows1[node->currentWWindow] = NULL;
						//Perform nest-loop on the wWindows
						node->returnState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_WINDOWS;
						node->mj_JoinState = EXEC_DBSIMJOIN_INIT_WINLOOP;
						break;
					}
					//Second pair of the set is in-memory
					else if(areWindowsInMemory(node, node->wWindows0[node->currentWWindow + 1],
						node->wWindows1[node->currentWWindow]))
					{
						/*
						 * Build arrays for internal-memory portion of the algorithm.
						 * Always use only 2 partitions/windows for internal-memory.
						 */
						node->internWPartitions0 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						node->internWPartitions1 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						node->internWWindows0 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						node->internWWindows1 = (Tuplestorestate **)
								palloc(sizeof(Tuplestorestate *) * 2);
						USEMEM(node, GetMemoryChunkSpace(node->internWPartitions0));
						USEMEM(node, GetMemoryChunkSpace(node->internWPartitions1));
						USEMEM(node, GetMemoryChunkSpace(node->internWWindows0));
						USEMEM(node, GetMemoryChunkSpace(node->internWWindows1));

						partitionWindows(node, node->wWindows0[node->currentWWindow + 1],
								node->wWindows1[node->currentWWindow], 2, 0, 0,
								node->internWPartitions0, node->internWPartitions1,
								node->internWWindows0, node->internWWindows1);

						/*
						 * Done with wWindows, set them to null. But we can't decrement the
						 * counter for number left until we're sure the whole set has been processed
						 */
						FREEMEM(node, GetMemoryChunkSpace(node->wWindows0[node->currentWWindow + 1]));
						FREEMEM(node, GetMemoryChunkSpace(node->wWindows1[node->currentWWindow]));
						tuplestore_end(node->wWindows0[node->currentWWindow + 1]);
						tuplestore_end(node->wWindows1[node->currentWWindow]);
						pfree(node->wWindows0[node->currentWWindow + 1]);
						pfree(node->wWindows1[node->currentWWindow]);
						node->wWindows0[node->currentWWindow + 1] = NULL;
						node->wWindows1[node->currentWWindow ] = NULL;

						//Set up counter variables
						node->oldNumInternWPartitions = 2;
						node->numInternWPartitionsLeft = 2;
						node->currentInternWPartition = 0;

						node->oldNumInternWWindows = 2;
						node->numInternWWindowsLeft = 2;
						node->currentInternWWindow = 0;

						//Start internal-memory algorithm
						node->internReturnState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_WINDOWS;
						node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_PARTITIONS;
						break;
					}
				}
				//Check whether both pairs of a set have been completed
				if(node->wWindows0[node->currentWWindow] == NULL && node->wWindows1[node->currentWWindow] == NULL &&
					node->wWindows0[node->currentWWindow + 1] == NULL && node->wWindows1[node->currentWWindow + 1] == NULL)
				{
					node->numWWindowsLeft -= 2;
				}
				node->currentWWindow+=2;
			}
			else
				//Next external wPartition
				node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_ITERATE_WINDOW_PARTITIONS;
		}
		break;

		/*XXX EXEC_DBSIMJOIN_EXTERN_ITERATE_WINDOW_PARTITIONS
		 * Process window-partitions externally
		 */
		case EXEC_DBSIMJOIN_EXTERN_ITERATE_WINDOW_PARTITIONS:
		{
			if(node->numWPartitionsLeft > 0) //Still partitions left to process
			{
				Tuplestorestate *tempWPartition0, *tempWPartition1;
				/*
				 * De-hibernate the first not-null wPartitions
				 */
				for(i=0;;i++)
					if(node->wPartitions0[i] != NULL && node->wPartitions1[i] != NULL)
					{
						tempWPartition0 = node->wPartitions0[i];
						tempWPartition1 = node->wPartitions1[i];
						node->wPartitions0[i] = NULL;
						node->wPartitions1[i++] = NULL;
						node->numWPartitionsLeft--;
						break;
					}
				FREEMEM(node, GetMemoryChunkSpace(tempWPartition0));
				FREEMEM(node, GetMemoryChunkSpace(tempWPartition1));
				reenableBufFile(tempWPartition0);
				reenableBufFile(tempWPartition1);
				USEMEM(node, GetMemoryChunkSpace(tempWPartition0));
				USEMEM(node, GetMemoryChunkSpace(tempWPartition1));

				/*
				 * Rebuild wPartition array without already-processed partitions,
				 * including space for new window-partitions
				 */
				node->newNumWPartitions = node->numWPartitionsLeft + node->numPartitions;
				newWPartitions0 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumWPartitions);
				newWPartitions1 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumWPartitions);
				USEMEM(node, GetMemoryChunkSpace(newWPartitions0));
				USEMEM(node, GetMemoryChunkSpace(newWPartitions1));
				//Put wPartitions into new array
				newIndex = 0;
				for(i = 0; i < node->oldNumWPartitions; i++)
				{
					if(node->wPartitions0[i] != NULL &&
						node->wPartitions1[i] != NULL)
					{
						newWPartitions0[newIndex] = node->wPartitions0[i];
						newWPartitions1[newIndex++] = node->wPartitions1[i];
					}
				}
				FREEMEM(node, GetMemoryChunkSpace(node->wPartitions0));
				FREEMEM(node, GetMemoryChunkSpace(node->wPartitions1));
				pfree(node->wPartitions0);
				pfree(node->wPartitions1);
				node->wPartitions0 = newWPartitions0;
				node->wPartitions1 = newWPartitions1;

				/*
				 * Rebuild window-window array without already-processed window-windows
				 * We can only remove full window-sets at a time, so if only one of the
				 * pairs has been done in-memory, we can't remove the set.
				 * Also include space for new window-windows
				 */
				node->newNumWWindows = node->numWWindowsLeft + node->numWindows;
				newWWindows0 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumWWindows);
				newWWindows1 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumWWindows);
				USEMEM(node, GetMemoryChunkSpace(newWWindows0));
				USEMEM(node, GetMemoryChunkSpace(newWWindows1));
				//Put wWindows into new array
				newIndex = 0;
				for(i = 0; i < node->oldNumWWindows; i++)
				{
					if((node->wWindows0[i] != NULL && node->wWindows1[i + 1] != NULL) ||
						(node->wWindows0[i + 1] != NULL && node->wWindows1[i] != NULL))
					{
						newWWindows0[newIndex] = node->wWindows0[i];
						newWWindows1[newIndex++] = node->wWindows1[i++];
						newWWindows0[newIndex] = node->wWindows0[i];
						newWWindows1[newIndex++] = node->wWindows1[i];
					}
					else i++;
				}
				FREEMEM(node, GetMemoryChunkSpace(node->wWindows0));
				FREEMEM(node, GetMemoryChunkSpace(node->wWindows1));
				pfree(node->wWindows0);
				pfree(node->wWindows1);
				node->wWindows0 = newWWindows0;
				node->wWindows1 = newWWindows1;

				//Set wPartition counter variables
				node->currentWPartition = node->numWPartitionsLeft;
				node->oldNumWPartitions = node->newNumWPartitions;

				//Set wWindow counter variables
				node->currentWWindow = node->numWWindowsLeft;
				node->oldNumWWindows = node->newNumWWindows;

				partitionWindows(node, tempWPartition0, tempWPartition1, node->numPartitions,
						node->currentWPartition, node->currentWWindow, node->wPartitions0,
						node->wPartitions1, node->wWindows0, node->wWindows1);

				node->numWPartitionsLeft+= node->numPartitions;
				node->numWWindowsLeft+= node->numWindows;

				//Done with the original wPartitions, can get rid of them
				FREEMEM(node, GetMemoryChunkSpace(tempWPartition0));
				FREEMEM(node, GetMemoryChunkSpace(tempWPartition1));
				tuplestore_end(tempWPartition0);
				tuplestore_end(tempWPartition1);
				pfree(tempWPartition0);
				pfree(tempWPartition1);

				/*
				 * Go process all new wPartitions and wWindows and do the ones that
				 * can be done in-memory
				 */
				hibernateNewWPartitions(node);
				hibernateNewWWindows(node);

				node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_PARTITIONS;
			}
			else //No more wPartitions exist.  Process the next external wWindow.
				node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_ITERATE_WINDOW_WINDOWS;
		}
		break;

		/*XXX EXEC_DBSIMJOIN_EXTERN_ITERATE_WINDOW_WINDOWS
		 * Process a window-window externally and check to see whether all windows
		 * partitions and windows have been processed
		 */
		case EXEC_DBSIMJOIN_EXTERN_ITERATE_WINDOW_WINDOWS:
		{
			//Check to see if any window-windows still need processed
			if(node->numWWindowsLeft > 0)
			{
				Tuplestorestate *tempWWindow0, *tempWWindow1;
				/*
				 * Try to dehibernate the wWindow0[0] - wWindow1[1] window pair.
				 * If we can't, dehibernate the wWindow0[1] - wWindow1[0] pair.
				 */
				for(i=0;;i+=2)
					if(node->wWindows0[i] != NULL && node->wWindows1[i+1] != NULL)
					{
						tempWWindow0= node->wWindows0[i];
						tempWWindow1 = node->wWindows1[i+1];
						node->wWindows0[i] = NULL;
						node->wWindows1[i+1] = NULL;
						break;
					}
					else if(node->wWindows0[i+1] != NULL && node->wWindows1[i] != NULL)
					{
						tempWWindow0 = node->wWindows0[i+1];
						tempWWindow1 = node->wWindows1[i];
						node->wWindows0[i+1] = NULL;
						node->wWindows1[i] = NULL;
						break;
					}

				/*
				 * Check entire first window set to see if we've processed both
				 * possible window-window pairs. If we have, decrement the counter
				 * for the total number of window-window pairs.
				 */
				if(node->wWindows0[i] == NULL && node->wWindows1[i] == NULL &&
					node->wWindows0[i+1] == NULL && node->wWindows1[i+1] == NULL)
				{
					node->numWWindowsLeft-=2;
				}

				FREEMEM(node, GetMemoryChunkSpace(tempWWindow0));
				FREEMEM(node, GetMemoryChunkSpace(tempWWindow1));
				reenableBufFile(tempWWindow0);
				reenableBufFile(tempWWindow1);
				USEMEM(node, GetMemoryChunkSpace(tempWWindow0));
				USEMEM(node, GetMemoryChunkSpace(tempWWindow1));

				/*
				 * Rebuild wPartition array without already-processed partitions,
				 * including space for new window-partitions
				 */
				node->newNumWPartitions = node->numWPartitionsLeft + node->numPartitions;
				newWPartitions0 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumWPartitions);
				newWPartitions1 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumWPartitions);
				USEMEM(node, GetMemoryChunkSpace(newWPartitions0));
				USEMEM(node, GetMemoryChunkSpace(newWPartitions1));
				//Put wPartitions into new array
				newIndex = 0;
				for(i = 0; i < node->oldNumWPartitions; i++)
				{
					if(node->wPartitions0[i] != NULL &&
						node->wPartitions1[i] != NULL)
					{
						newWPartitions0[newIndex] = node->wPartitions0[i];
						newWPartitions1[newIndex++] = node->wPartitions1[i];
					}
				}
				FREEMEM(node, GetMemoryChunkSpace(node->wPartitions0));
				FREEMEM(node, GetMemoryChunkSpace(node->wPartitions1));
				pfree(node->wPartitions0);
				pfree(node->wPartitions1);
				node->wPartitions0 = newWPartitions0;
				node->wPartitions1 = newWPartitions1;

				/*
				 * Rebuild window-window array without already-processed window-windows
				 * We can only remove full window-sets at a time, so if only one of the
				 * pairs has been done in-memory, we can't remove the set.
				 * Also include space for new window-windows
				 */
				node->newNumWWindows = node->numWWindowsLeft + node->numWindows;
				newWWindows0 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumWWindows);
				newWWindows1 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumWWindows);
				USEMEM(node, GetMemoryChunkSpace(newWWindows0));
				USEMEM(node, GetMemoryChunkSpace(newWWindows1));
				//Put wWindows into new array
				newIndex = 0;
				for(i = 0; i < node->oldNumWWindows; i++)
				{
					if(!((node->wWindows0[i] == NULL || node->wWindows1[i + 1] == NULL) &&
						(node->wWindows0[i + 1] == NULL || node->wWindows1[i] == NULL)))
					{
						newWWindows0[newIndex] = node->wWindows0[i];
						newWWindows1[newIndex++] = node->wWindows1[i++];
						newWWindows0[newIndex] = node->wWindows0[i];
						newWWindows1[newIndex++] = node->wWindows1[i];
					}
					else i++;
				}
				FREEMEM(node, GetMemoryChunkSpace(node->wWindows0));
				FREEMEM(node, GetMemoryChunkSpace(node->wWindows1));
				pfree(node->wWindows0);
				pfree(node->wWindows1);
				node->wWindows0 = newWWindows0;
				node->wWindows1 = newWWindows1;

				//Set wPartition counter variables
				node->currentWPartition = node->numWPartitionsLeft;
				node->oldNumWPartitions = node->newNumWPartitions;

				//Set wWindow counter variables
				node->currentWWindow = node->numWWindowsLeft;
				node->oldNumWWindows = node->newNumWWindows;

				partitionWindows(node, tempWWindow0, tempWWindow1, node->numPartitions,
						node->currentWPartition, node->currentWWindow, node->wPartitions0,
						node->wPartitions1, node->wWindows0, node->wWindows1);

				node->numWPartitionsLeft+= node->numPartitions;
				node->numWWindowsLeft+= node->numWindows;
				/*
				 * Done with original wWindows, can remove them
				 */
				FREEMEM(node, GetMemoryChunkSpace(tempWWindow0));
				FREEMEM(node, GetMemoryChunkSpace(tempWWindow1));
				tuplestore_end(tempWWindow0);
				tuplestore_end(tempWWindow1);
				pfree(tempWWindow0);
				pfree(tempWWindow1);

				//Process new wPartitions and wWindows that can be done in-memory
				hibernateNewWPartitions(node);
				hibernateNewWWindows(node);

				node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_PARTITIONS;
			}
			/*
			 * No more wWindows but still wPartitions
			 * Need this test to avoid a situation where we skip processing the
			 * remaining window-partitions
			 */
			else if(node->numWPartitionsLeft > 0)
			{
				node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOW_PARTITIONS;
			}
			else //Original window fully processed.  Go process the next external window
			{
				/*
				 * We can de-allocate memory for the window-partition arrays,
				 * as they will be re-made each time we go into window processing
				 */
				FREEMEM(node, GetMemoryChunkSpace(node->wPartitions0));
				FREEMEM(node, GetMemoryChunkSpace(node->wPartitions1));
				FREEMEM(node, GetMemoryChunkSpace(node->wWindows0));
				FREEMEM(node, GetMemoryChunkSpace(node->wWindows1));
				pfree(node->wPartitions0);
				pfree(node->wPartitions1);
				pfree(node->wWindows0);
				pfree(node->wWindows1);
				i = 0;
				node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_PROCESS_WINDOWS;
			}
		}
		break;

		/*XXX EXEC_DBSIMJOIN_EXTERN_ITERATE_PARTITIONS
		 * Start processing partitions externally
		 */
		case EXEC_DBSIMJOIN_EXTERN_ITERATE_PARTITIONS:
		{
			//Still have partitions to process
			if(node->numPartitionsLeft > 0)
			{
				Tuplestorestate *tempPartition;
				//De-hibernate the first not-null partition so we can work on it
				for(i = 0;;i++)
					if(node->partitions[i] != NULL)
					{
						tempPartition = node->partitions[i];
						node->partitions[i++] = NULL;
						node->numPartitionsLeft--;
						break;
					}

				FREEMEM(node, GetMemoryChunkSpace(tempPartition));
				reenableBufFile(tempPartition);
				USEMEM(node, GetMemoryChunkSpace(tempPartition));

				/*
				 * Rebuild partition array without already-processed partitions,
				 * including space for new partitions
				 */
				node->newNumPartitions = node->numPartitionsLeft + node->numPartitions;
				newPartitions = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumPartitions);
				USEMEM(node, GetMemoryChunkSpace(newPartitions));

				//Put partitions into new array
				newIndex = 0;
				for(i = 0; i < node->oldNumPartitions; i++)
				{
					if(node->partitions[i] != NULL)
						newPartitions[newIndex++] = node->partitions[i];
				}
				FREEMEM(node, GetMemoryChunkSpace(node->partitions));
				pfree(node->partitions);
				node->partitions = newPartitions;

				/*
				 * Create a new window array
				 */
				node->newNumWindows = node->numWindowsLeft + node->numWindows;
				node->windows = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumWindows);
				USEMEM(node, GetMemoryChunkSpace(node->windows));


				//Reset partition counter variables
				node->oldNumPartitions = node->newNumPartitions;
				node->currentPartition = node->numPartitionsLeft;

				//Reset window counter variables to their initial values.
				node->oldNumWindows = node->numWindows;
				node->currentWindow = 0;

				partition(node, tempPartition, node->numPartitions, node->currentPartition,
						node->currentWindow, node->partitions, node->windows);

				node->numPartitionsLeft+= node->numPartitions;
				node->numWindowsLeft = node->numWindows;

				//Done adding tuples, can get rid of original partition
				FREEMEM(node, GetMemoryChunkSpace(tempPartition));
				tuplestore_end(tempPartition);
				pfree(tempPartition);

				/*
				 * Now we have new partitions/windows to process, we should head
				 * back to PROCESS_PARTITIONS_IN_MEMORY to continue.
				 */
				hibernateNewPartitions(node);
				hibernateNewWindows(node);

				node->mj_JoinState = EXEC_DBSIMJOIN_EXTERN_PROCESS_PARTITIONS;
			}
			/*
			 * Done with processing all partitions.  Should be done.
			 */
			else
			{
				//Deallocate memory for arrays
				FREEMEM(node, GetMemoryChunkSpace(node->partitions));
				pfree(node->partitions);
				resetPivots(node);
				return NULL;
			}
		}
		break;

		/*XXX EXEC_DBSIMJOIN_INTERN_PROCESS_PARTITIONS
		 * Process internPartitions with the internal-memory algorithm
		 */
		case EXEC_DBSIMJOIN_INTERN_PROCESS_PARTITIONS:
		{
			 //Check to see if we've processed all new internPartitions
			if(node->currentInternPartition < node->oldNumInternPartitions)
			{
				/*
				 * Test for whether or not the internPartition is small enough to
				 * perform a nest-loop on.
				 */
				if(isPartitionJoinReady(node, node->internPartitions[node->currentInternPartition]))
				{
					//Small enough to join
					node->partitionToJoin0 = node->internPartitions[node->currentInternPartition];
					node->internPartitions[node->currentInternPartition++] = NULL;
					node->numInternPartitionsLeft--;
					//Perform nest-loop on the internPartition
					node->returnState = EXEC_DBSIMJOIN_INTERN_PROCESS_PARTITIONS;
					node->mj_JoinState = EXEC_DBSIMJOIN_INIT_NESTLOOP;
					break;
				}
				else
					node->currentInternPartition++;
			}
			else
				node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOWS;
		}
		break;

		/*XXX EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOWS
		 * Process internWindows
		 */
		case EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOWS:
		{
			if(node->numInternWindowsLeft > 0)
			{
				//Check to see if we've processed all new internWindows
				if(node->currentInternWindow < node->oldNumInternWindows)
				{
					/*
					 * Test for whether or not the internWindows are small enough to
					 * perform a nest-loop on
					 */
					if(areWindowsJoinReady(node, node->internWindows[node->currentInternWindow],
							node->internWindows[node->currentInternWindow + 1]))
					{
						//Small enough to join
						node->partitionToJoin0 = node->internWindows[node->currentInternWindow];
						node->internWindows[node->currentInternWindow++] = NULL;
						node->partitionToJoin1 = node->internWindows[node->currentInternWindow];
						node->internWindows[node->currentInternWindow++] = NULL;
						node->numInternWindowsLeft-= 2;
						//Perform nest-loop on the internWindows
						node->returnState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOWS;
						node->mj_JoinState = EXEC_DBSIMJOIN_INIT_WINLOOP;
						break;
					}
					else
						node->currentInternWindow+=2;
				}
				else
				{
					Tuplestorestate *tempInternWindow0, *tempInternWindow1;
					//Get the first not-null internWindow set so we can work on it.
					for(i = 0;;i+=2)
						if(node->internWindows[i] != NULL && node->internWindows[i+1] != NULL)
						{
							tempInternWindow0 = node->internWindows[i];
							node->internWindows[i++] = NULL;
							tempInternWindow1 = node->internWindows[i];
							node->internWindows[i++] = NULL;
							node->numInternWindowsLeft-=2;
							break;
						}

					/*Set up arrays for the internWWindow iteration*/
					//internWPartition arrays
					node->internWPartitions0 = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * 2);
					node->internWPartitions1 = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * 2);
					USEMEM(node, GetMemoryChunkSpace(node->internWPartitions0));
					USEMEM(node, GetMemoryChunkSpace(node->internWPartitions1));

					//internWWindowArrays
					node->internWWindows0 = (Tuplestorestate**)
							palloc(sizeof(Tuplestorestate *)* 2);
					node->internWWindows1 = (Tuplestorestate **)
							palloc(sizeof(Tuplestorestate *) * 2);
					USEMEM(node, GetMemoryChunkSpace(node->internWWindows0));
					USEMEM(node, GetMemoryChunkSpace(node->internWWindows1));

					partitionWindows(node, tempInternWindow0, tempInternWindow1, 2,
							0,0, node->internWPartitions0, node->internWPartitions1,
							node->internWWindows0, node->internWWindows1);

					//Remove first window set, we're done with it
					FREEMEM(node, GetMemoryChunkSpace(tempInternWindow0));
					FREEMEM(node, GetMemoryChunkSpace(tempInternWindow1));
					tuplestore_end(tempInternWindow0);
					tuplestore_end(tempInternWindow1);
					pfree(tempInternWindow0);
					pfree(tempInternWindow1);

					//Initialize variables for internal window processing
					node->oldNumInternWPartitions = 2;
					node->numInternWPartitionsLeft = 2;
					node->currentInternWPartition = 0;

					node->oldNumInternWWindows = 2;
					node->numInternWWindowsLeft = 2;
					node->currentInternWWindow = 0;

					node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_PARTITIONS;
				}
			}
			/*
			 * All internWindows processed.  Go on to the next internal partition
			 * iteration
			 */
			else
				node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_ITERATE_PARTITIONS;
		}
		break;

		/*XXX EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_PARTITIONS
		 * Process internWPartitions
		 */
		case EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_PARTITIONS:
		{
			//Check to see if we've processed all new internWPartitions
			if(node->currentInternWPartition < node->oldNumInternWPartitions)
			{
				/*
				 * Test for whether or not the internWPartitions are small enough to
				 * perform a nest-loop on
				 */
				if(areWindowsJoinReady(node, node->internWPartitions0[node->currentInternWPartition],
						node->internWPartitions1[node->currentInternWPartition]))
				{
					//Small enough to join
					node->partitionToJoin0 = node->internWPartitions0[node->currentInternWPartition];
					node->partitionToJoin1 = node->internWPartitions1[node->currentInternWPartition];
					node->internWPartitions0[node->currentInternWPartition] = NULL;
					node->internWPartitions1[node->currentInternWPartition++] = NULL;
					node->numInternWPartitionsLeft--;
					//Perform nest-loop on the wPartition
					node->returnState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_PARTITIONS;
					node->mj_JoinState = EXEC_DBSIMJOIN_INIT_WINLOOP;
					break;
				}
				else
					node->currentInternWPartition++;
			}
			/*
			 * Check to see if we've processed all new internWPartitions that
			 * we can process in-memory
			 */
			else
				node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_WINDOWS;
		}
		break;

		/*XXX EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_WINDOWS
		 * Process internWWindows
		 */
		case EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_WINDOWS:
		{
			if(node->currentInternWWindow < node->oldNumInternWWindows)
			{
				/*
				 * Test for whether or not the first pair of the window set is small
				 * enough to perform a nest-loop on and has not been processed yet
				 */
				if((node->internWWindows0[node->currentInternWWindow] != NULL &&
				   node->internWWindows1[node->currentInternWWindow + 1] != NULL) &&
				   areWindowsJoinReady(node, node->internWWindows0[node->currentInternWWindow],
				   node->internWWindows1[node->currentInternWWindow + 1]))
				{
					//Small enough to join
					node->partitionToJoin0 = node->internWWindows0[node->currentInternWWindow];
					node->partitionToJoin1 = node->internWWindows1[node->currentInternWWindow + 1];
					node->internWWindows0[node->currentInternWWindow] = NULL;
					node->internWWindows1[node->currentInternWWindow + 1] = NULL;
					//Perform nest-loop on the wWindows
					node->returnState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_WINDOWS;
					node->mj_JoinState = EXEC_DBSIMJOIN_INIT_WINLOOP;
					break;
				}
				/*
				 * Test for whether or not the second pair of the window set is small
				 * enough to perform a nest-loop on and has not been processed yet
				 */
				if((node->internWWindows0[node->currentInternWWindow + 1] != NULL &&
					node->internWWindows1[node->currentInternWWindow] != NULL) &&
					areWindowsJoinReady(node, node->internWWindows0[node->currentInternWWindow + 1],
					node->internWWindows1[node->currentInternWWindow]))
				{
					//Small enough to join
					node->partitionToJoin0 = node->internWWindows0[node->currentInternWWindow + 1];
					node->partitionToJoin1 = node->internWWindows1[node->currentInternWWindow ];
					node->internWWindows0[node->currentInternWWindow + 1] = NULL;
					node->internWWindows1[node->currentInternWWindow] = NULL;
					//Perform nest-loop on the wWindows
					node->returnState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_WINDOWS;
					node->mj_JoinState = EXEC_DBSIMJOIN_INIT_WINLOOP;
					break;
				}
				//Check whether both pairs of a set have been completed
				if(node->internWWindows0[node->currentInternWWindow] == NULL && node->internWWindows1[node->currentInternWWindow] == NULL &&
					node->internWWindows0[node->currentInternWWindow + 1] == NULL && node->internWWindows1[node->currentInternWWindow + 1] == NULL)
				{
					node->numInternWWindowsLeft -= 2;
				}
				node->currentInternWWindow+=2;
			}
			else
				//Next internal wPartition
				node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_ITERATE_WINDOW_PARTITIONS;
		}
		break;

		/*XXX EXEC_DBSIMJOIN_INTERN_ITERATE_WINDOW_PARTITIONS
		 * Perform next iteration on the next hibernated internWPartition
		 */
		case EXEC_DBSIMJOIN_INTERN_ITERATE_WINDOW_PARTITIONS:
		{
			if(node->numInternWPartitionsLeft > 0) //Still partitions left to process
			{
				Tuplestorestate *tempInternWPartition0, *tempInternWPartition1;
				/*
				 * Get the first not-null internWPartitions
				 */
				for(i=0;;i++)
					if(node->internWPartitions0[i] != NULL && node->internWPartitions1[i] != NULL)
					{
						tempInternWPartition0 = node->internWPartitions0[i];
						tempInternWPartition1 = node->internWPartitions1[i];
						node->internWPartitions0[i] = NULL;
						node->internWPartitions1[i++] = NULL;
						node->numInternWPartitionsLeft--;
						break;
					}

				/*
				 * Rebuild internal wPartition array without already-processed partitions,
				 * including space for new window-partitions
				 */
				node->newNumInternWPartitions = node->numInternWPartitionsLeft + 2;
				newWPartitions0 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumInternWPartitions);
				newWPartitions1 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumInternWPartitions);
				USEMEM(node, GetMemoryChunkSpace(newWPartitions0));
				USEMEM(node, GetMemoryChunkSpace(newWPartitions1));
				//Put internWPartitions into new array
				newIndex = 0;
				for(i = 0; i < node->oldNumInternWPartitions; i++)
				{
					if(node->internWPartitions0[i] != NULL &&
						node->internWPartitions1[i] != NULL)
					{
						newWPartitions0[newIndex] = node->internWPartitions0[i];
						newWPartitions1[newIndex++] = node->internWPartitions1[i];
					}
				}
				FREEMEM(node, GetMemoryChunkSpace(node->internWPartitions0));
				FREEMEM(node, GetMemoryChunkSpace(node->internWPartitions1));
				pfree(node->internWPartitions0);
				pfree(node->internWPartitions1);
				node->internWPartitions0 = newWPartitions0;
				node->internWPartitions1 = newWPartitions1;

				/*
				 * Rebuild internal window-window array without already-processed window-windows
				 * We can only remove full window-sets at a time, so if only one of the
				 * pairs has been done in-memory, we can't remove the set.
				 * Also include space for new window-windows
				 */
				node->newNumInternWWindows = node->numInternWWindowsLeft + 2;
				newWWindows0 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumInternWWindows);
				newWWindows1 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumInternWWindows);
				USEMEM(node, GetMemoryChunkSpace(newWWindows0));
				USEMEM(node, GetMemoryChunkSpace(newWWindows1));
				//Put internWWindows into new array
				newIndex = 0;
				for(i = 0; i < node->oldNumInternWWindows; i++)
				{
					if((node->internWWindows0[i] != NULL && node->internWWindows1[i + 1] != NULL) ||
						(node->internWWindows0[i + 1] != NULL && node->internWWindows1[i] != NULL))
					{
						newWWindows0[newIndex] = node->internWWindows0[i];
						newWWindows1[newIndex++] = node->internWWindows1[i++];
						newWWindows0[newIndex] = node->internWWindows0[i];
						newWWindows1[newIndex++] = node->internWWindows1[i];
					}
					else i++;
				}
				FREEMEM(node, GetMemoryChunkSpace(node->internWWindows0));
				FREEMEM(node, GetMemoryChunkSpace(node->internWWindows1));
				pfree(node->internWWindows0);
				pfree(node->internWWindows1);
				node->internWWindows0 = newWWindows0;
				node->internWWindows1 = newWWindows1;

				//Set internWPartition counter variables
				node->currentInternWPartition = node->numInternWPartitionsLeft;
				node->oldNumInternWPartitions = node->newNumInternWPartitions;

				//Set internWWindow counter variables
				node->currentInternWWindow = node->numInternWWindowsLeft;
				node->oldNumInternWWindows = node->newNumInternWWindows;

				partitionWindows(node, tempInternWPartition0, tempInternWPartition1, 2,
						node->currentInternWPartition, node->currentInternWWindow, node->internWPartitions0,
						node->internWPartitions1, node->internWWindows0, node->internWWindows1);

				node->numInternWPartitionsLeft+= 2;
				node->numInternWWindowsLeft+= 2;

				//Done with the original internWPartitions, can get rid of them
				FREEMEM(node, GetMemoryChunkSpace(tempInternWPartition0));
				FREEMEM(node, GetMemoryChunkSpace(tempInternWPartition1));
				tuplestore_end(tempInternWPartition0);
				tuplestore_end(tempInternWPartition1);
				pfree(tempInternWPartition0);
				pfree(tempInternWPartition1);

				/*
				 * Go process all new internWPartitions and internWWindows and do the ones that
				 * can be done in-memory
				 */
				node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_PARTITIONS;
			}
			else //No more internWPartitions exist.  Process the next external wWindow.
			{
				i = 0;
				node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_ITERATE_WINDOW_WINDOWS;
			}
		}
		break;

		//XXX EXEC_DBSIMJOIN_INTERN_ITERATE_WINDOW_WINDOWS
		case EXEC_DBSIMJOIN_INTERN_ITERATE_WINDOW_WINDOWS:
		{
			//Check to see if any window-windows still need processed
			if(node->numInternWWindowsLeft > 0)
			{
				Tuplestorestate *tempInternWWindow0, *tempInternWWindow1;
				/*
				 * Try to process the wWindow0[0] - wWindow1[1] window pair.
				 * If we can't, process the wWindow0[1] - wWindow1[0] pair.
				 */
				for(i=0;;i+=2)
					if(node->internWWindows0[i] != NULL && node->internWWindows1[i+1] != NULL)
					{
						tempInternWWindow0= node->internWWindows0[i];
						tempInternWWindow1 = node->internWWindows1[i+1];
						node->internWWindows0[i] = NULL;
						node->internWWindows1[i+1] = NULL;
						break;
					}
					else if(node->internWWindows0[i+1] != NULL && node->internWWindows1[i] != NULL)
					{
						tempInternWWindow0 = node->internWWindows0[i+1];
						tempInternWWindow1 = node->internWWindows1[i];
						node->internWWindows0[i+1] = NULL;
						node->internWWindows1[i] = NULL;
						break;
					}

				/*
				 * Check entire first window set to see if we've processed both
				 * possible window-window pairs. If we have, decrement the counter
				 * for the total number of window-window pairs.
				 */
				if(node->internWWindows0[i] == NULL && node->internWWindows1[i] == NULL &&
					node->internWWindows0[i+1] == NULL && node->internWWindows1[i+1] == NULL)
				{
					node->numInternWWindowsLeft-=2;
				}

				/*
				 * Rebuild wPartition array without already-processed partitions,
				 * including space for new window-partitions
				 */
				node->newNumInternWPartitions = node->numInternWPartitionsLeft + 2;
				newWPartitions0 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumInternWPartitions);
				newWPartitions1 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumInternWPartitions);
				USEMEM(node, GetMemoryChunkSpace(newWPartitions0));
				USEMEM(node, GetMemoryChunkSpace(newWPartitions1));
				//Put wPartitions into new array
				newIndex = 0;
				for(i = 0; i < node->oldNumInternWPartitions; i++)
				{
					if(node->internWPartitions0[i] != NULL &&
						node->internWPartitions1[i] != NULL)
					{
						newWPartitions0[newIndex] = node->internWPartitions0[i];
						newWPartitions1[newIndex++] = node->internWPartitions1[i];
					}
				}
				FREEMEM(node, GetMemoryChunkSpace(node->internWPartitions0));
				FREEMEM(node, GetMemoryChunkSpace(node->internWPartitions1));
				pfree(node->internWPartitions0);
				pfree(node->internWPartitions1);
				node->internWPartitions0 = newWPartitions0;
				node->internWPartitions1 = newWPartitions1;

				/*
				 * Rebuild window-window array without already-processed window-windows
				 * We can only remove full window-sets at a time, so if only one of the
				 * pairs has been done in-memory, we can't remove the set.
				 * Also include space for new window-windows
				 */
				node->newNumInternWWindows = node->numInternWWindowsLeft + 2;
				newWWindows0 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumInternWWindows);
				newWWindows1 = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumInternWWindows);
				USEMEM(node, GetMemoryChunkSpace(newWWindows0));
				USEMEM(node, GetMemoryChunkSpace(newWWindows1));
				//Put wWindows into new array
				newIndex = 0;
				for(i = 0; i < node->oldNumInternWWindows; i++)
				{
					if(!((node->internWWindows0[i] == NULL || node->internWWindows1[i + 1] == NULL) &&
						(node->internWWindows0[i + 1] == NULL || node->internWWindows1[i] == NULL)))
					{
						newWWindows0[newIndex] = node->internWWindows0[i];
						newWWindows1[newIndex++] = node->internWWindows1[i++];
						newWWindows0[newIndex] = node->internWWindows0[i];
						newWWindows1[newIndex++] = node->internWWindows1[i];
					}
					else i++;
				}
				FREEMEM(node, GetMemoryChunkSpace(node->internWWindows0));
				FREEMEM(node, GetMemoryChunkSpace(node->internWWindows1));
				pfree(node->internWWindows0);
				pfree(node->internWWindows1);
				node->internWWindows0 = newWWindows0;
				node->internWWindows1 = newWWindows1;

				//Set wPartition counter variables
				node->currentInternWPartition = node->numInternWPartitionsLeft;
				node->oldNumInternWPartitions = node->newNumInternWPartitions;

				//Set wWindow counter variables
				node->currentInternWWindow = node->numInternWWindowsLeft;
				node->oldNumInternWWindows = node->newNumInternWWindows;

				partitionWindows(node, tempInternWWindow0, tempInternWWindow1, 2,
						node->currentInternWPartition, node->currentInternWWindow,
						node->internWPartitions0, node->internWPartitions1, node->internWWindows0, node->internWWindows1);

				node->numInternWPartitionsLeft+= 2;
				node->numInternWWindowsLeft+= 2;
				/*
				 * Done with original wWindows, can remove them
				 */
				FREEMEM(node, GetMemoryChunkSpace(tempInternWWindow0));
				FREEMEM(node, GetMemoryChunkSpace(tempInternWWindow1));
				tuplestore_end(tempInternWWindow0);
				tuplestore_end(tempInternWWindow1);
				pfree(tempInternWWindow0);
				pfree(tempInternWWindow1);

				node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_PARTITIONS;
			}
			/*
			 * No more wWindows but still wPartitions
			 * Need this test to avoid a situation where we skip processing the
			 * remaining window-partitions
			 */
			else if(node->numInternWPartitionsLeft > 0)
			{
				node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOW_PARTITIONS;
			}
			else //Original window fully processed.  Go process the next external window
			{
				/*
				 * We can de-allocate memory for the window-partition arrays,
				 * as they will be re-made each time we go into window processing
				 */
				FREEMEM(node, GetMemoryChunkSpace(node->internWPartitions0));
				FREEMEM(node, GetMemoryChunkSpace(node->internWPartitions1));
				FREEMEM(node, GetMemoryChunkSpace(node->internWWindows0));
				FREEMEM(node, GetMemoryChunkSpace(node->internWWindows1));
				pfree(node->internWPartitions0);
				pfree(node->internWPartitions1);
				pfree(node->internWWindows0);
				pfree(node->internWWindows1);
				i = 0;
				/*
				 * If we originally came to the internal algorithm from processing
				 * a window, we need to return there now to avoid problems with
				 * the non-existent internPartition and internWindow arrays.
				 */
				node->mj_JoinState = node->internReturnState !=
						EXEC_DBSIMJOIN_EXTERN_PROCESS_PARTITIONS ? node->internReturnState :
						EXEC_DBSIMJOIN_INTERN_PROCESS_WINDOWS;
			}
		}
		break;

		//XXX EXEC_DBSIMJOIN_INTERN_ITERATE_PARTITIONS
		case EXEC_DBSIMJOIN_INTERN_ITERATE_PARTITIONS:
		{
			//Still have partitions to process
			if(node->numInternPartitionsLeft > 0)
			{
				Tuplestorestate *tempInternPartition;
				//De-hibernate the first not-null partition so we can work on it
				for(i = 0;;i++)
					if(node->internPartitions[i] != NULL)
					{
						tempInternPartition = node->internPartitions[i];
						node->internPartitions[i++] = NULL;
						node->numInternPartitionsLeft--;
						break;
					}

				/*
				 * Rebuild partition array without already-processed partitions,
				 * including space for new partitions
				 */
				node->newNumInternPartitions = node->numInternPartitionsLeft + 2;
				newPartitions = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumInternPartitions);
				USEMEM(node, GetMemoryChunkSpace(newPartitions));

				//Put partitions into new array
				newIndex = 0;
				for(i = 0; i < node->oldNumInternPartitions; i++)
				{
					if(node->internPartitions[i] != NULL)
						newPartitions[newIndex++] = node->internPartitions[i];
				}
				FREEMEM(node, GetMemoryChunkSpace(node->internPartitions));
				pfree(node->internPartitions);
				node->internPartitions = newPartitions;

				/*
				 * Create a new window array
				 */
				node->newNumInternWindows = node->numInternWindowsLeft + 2;
				node->internWindows = (Tuplestorestate **)
						palloc(sizeof(Tuplestorestate *) * node->newNumInternWindows);
				USEMEM(node, GetMemoryChunkSpace(node->internWindows));


				//Reset partition counter variables
				node->oldNumInternPartitions = node->newNumInternPartitions;
				node->currentInternPartition = node->numInternPartitionsLeft;

				//Reset window counter variables to their initial values.
				node->oldNumInternWindows = 2;
				node->currentInternWindow = 0;

				partition(node, tempInternPartition, 2, node->currentInternPartition,
						node->currentInternWindow, node->internPartitions, node->internWindows);

				node->numInternPartitionsLeft+= 2;
				node->numInternWindowsLeft+= 2;

				//Done adding tuples, can get rid of original partition
				FREEMEM(node, GetMemoryChunkSpace(tempInternPartition));
				tuplestore_end(tempInternPartition);
				pfree(tempInternPartition);

				/*
				 * Now we have new partitions/windows to process, we should head
				 * back to PROCESS_PARTITIONS_IN_MEMORY to continue.
				 */
				node->mj_JoinState = EXEC_DBSIMJOIN_INTERN_PROCESS_PARTITIONS;
			}
			/*
			 * Done with processing all partitions.  Should be done.
			 * Return to external portion of the algorithm.
			 */
			else
			{
				//Deallocate memory for arrays
				FREEMEM(node, GetMemoryChunkSpace(node->internPartitions));
				pfree(node->internPartitions);
				FREEMEM(node, GetMemoryChunkSpace(node->internWindows));
				pfree(node->internWindows);
				node->mj_JoinState = node->internReturnState;
			}
		}
		break;

		/*XXX EXEC_DBSIMJOIN_INIT_SMALL_INPUT_NESTLOOP
		 * Set up to do nestloop of very small initial input (smaller than #pivots)
		 */
		case EXEC_DBSIMJOIN_INIT_SMALL_INPUT_NESTLOOP:
		{
			outerTupleSlot = ExecProcNode(outerPlan);
			if(TupIsNull(outerTupleSlot))
			{
				/*If outer subplan exhausted, get tuples from inner*/
				innerTupleSlot = ExecProcNode(innerPlan);
				if(TupIsNull(innerTupleSlot))
				{
					//All tuples added, start the join
					node->returnState = EXEC_DBSIMJOIN_FINISH_SMALL_INPUT;
					node->mj_JoinState = EXEC_DBSIMJOIN_INIT_NESTLOOP;
					break;
				}
				initTupleFromPlan(innerTupleSlot, node->mj_InnerTupleSlot, false);
				tuplestore_puttupleslot(node->partitionToJoin0, node->mj_InnerTupleSlot);
			}
			else
			{
				initTupleFromPlan(outerTupleSlot, node->mj_OuterTupleSlot, true);
				tuplestore_puttupleslot(node->partitionToJoin0, node->mj_OuterTupleSlot);
			}
		}
		break;

		/*XXX EXEC_DBSIMJOIN_INIT_NESTLOOP
		 * Initial marking of the tuplestore partition we are going to join
		 */
		case EXEC_DBSIMJOIN_INIT_NESTLOOP:
		{
			tuplestore_markpos(node->partitionToJoin0);
			node->processingPartition = true;
			node->mj_JoinState = EXEC_DBSIMJOIN_ITER_NESTLOOP;
		}
		break;

		/*XXX EXEC_DBSIMJOIN_ITER_NESTLOOP
		 * Advances the outer iteration of the nested join loop
		 */
		case EXEC_DBSIMJOIN_ITER_NESTLOOP:
		{
			//Go back to last position of outer loop
			tuplestore_restorepos(node->partitionToJoin0);
			//Advance one position
			node->nestedJoinOuter = getTuple(node, node->partitionToJoin0);
			if(node->nestedJoinOuter != NULL)
			{
				//Continue joining
				tuplestore_markpos(node->partitionToJoin0);
				node->mj_JoinState = EXEC_DBSIMJOIN_NESTLOOP;
			}
			else
			{
				//Done joining partition
				node->mj_JoinState = node->returnState;
				//End the tuplestore, freeing up the partition
				FREEMEM(node, GetMemoryChunkSpace(node->partitionToJoin0));
				tuplestore_end(node->partitionToJoin0);
				pfree(node->partitionToJoin0);
			}
		}
		break;

		/*XXX EXEC_DBSIMJOIN_NESTLOOP
		 * Performs the inner nested loop join for a partition
		 */
		case EXEC_DBSIMJOIN_NESTLOOP:
		{
			node->nestedJoinInner = getNextTuple(node, node->partitionToJoin0);
			if(node->nestedJoinInner != NULL)
			{
				bool sameTable = fromSameTable(node->nestedJoinOuter, node->nestedJoinInner);
				//bool sameTable = false; //DEBUG
				if(!sameTable)
				{
					double dist =
							//MJCompare_EditDist	//Uncomment this line for String data
							MJCompare_DistFunc		//Uncomment this line for vector data
							(node, node->nestedJoinInner,
						node->nestedJoinOuter);
					if(dist >= 0 && dist <= epsilonDist)
					{
						//Tuples are similar and from different tables, join them
						resetTuple(node, node->nestedJoinOuter);
						resetTuple(node, node->nestedJoinInner);
						node->mj_JoinState = EXEC_MJ_JOINTUPLES;
					}
				}
			}
			else
				//This iteration is done, go to the next one
				node->mj_JoinState = EXEC_DBSIMJOIN_ITER_NESTLOOP;
		}
		break;

		/*XXX EXEC_DBSIMJOIN_INIT_WINLOOP
		 * Initial marking of the tuplestores we join for window join
		 */
		case EXEC_DBSIMJOIN_INIT_WINLOOP:
		{
			node->processingPartition = false;
			node->mj_JoinState = EXEC_DBSIMJOIN_ITER_WINLOOP;
		}
		break;

		/*XXX EXEC_DBSIMJOIN_ITER_WINLOOP
		 * Advances the outer iteration of the nested join loop
		 */
		case EXEC_DBSIMJOIN_ITER_WINLOOP:
		{
			//Check to make sure that there are tuples we still need to check
			node->nestedJoinOuter = getTuple(node, node->partitionToJoin0);
			if(node->nestedJoinOuter != NULL)
			{
				//Reset inner loop
				tuplestore_rescan(node->partitionToJoin1);
				//Continue joining
				node->mj_JoinState = EXEC_DBSIMJOIN_NESTWINLOOP;
			}
			else
			{
				//Done joining windows
				node->mj_JoinState = node->returnState;
				//End the tuplestores, freeing up the windows
				FREEMEM(node, GetMemoryChunkSpace(node->partitionToJoin0));
				FREEMEM(node, GetMemoryChunkSpace(node->partitionToJoin1));
				tuplestore_end(node->partitionToJoin0);
				tuplestore_end(node->partitionToJoin1);
				pfree(node->partitionToJoin0);
				pfree(node->partitionToJoin1);
			}
		}
		break;

		/*XXX EXEC_DBSIMJOIN_NESTWINLOOP
		 * Performs inner nested loop join for windows
		 */
		case EXEC_DBSIMJOIN_NESTWINLOOP:
		{
			node->nestedJoinInner = getNextTuple(node, node->partitionToJoin1);
			if(node->nestedJoinInner != NULL)
			{
				bool sameTable = fromSameTable(node->nestedJoinOuter, node->nestedJoinInner);
				if(!sameTable)
				{
					double dist =
							//MJCompare_EditDist	//Uncomment this line for String data
							MJCompare_DistFunc		//Uncomment this line for vector data
							(node, node->nestedJoinInner,
						node->nestedJoinOuter);
					if(dist >= 0 && dist <= epsilonDist)
					{
						//Tuples are similar and from different tables, join them
						resetTuple(node, node->nestedJoinOuter);
						resetTuple(node, node->nestedJoinInner);
						node->mj_JoinState = EXEC_MJ_JOINTUPLES;
					}
				}
			}
			else
			{
				//This iteration is done, go to the next one
				node->mj_JoinState = EXEC_DBSIMJOIN_ITER_WINLOOP;
			}
		}
		break;

		/*XXX EXEC_DBSIMJOIN_FINISH_SMALL_INPUT
		 * Finish up after dealing with a very small initial input
		 */
		case EXEC_DBSIMJOIN_FINISH_SMALL_INPUT:
		{
			return NULL;
		}
		break;

		/********************
		 * End of DBSimJoin *
		 * specific cases	*
		 ********************/

		/*
		 * EXEC_MJ_INITIALIZE_OUTER means that this is the first time
		 * ExecMergeJoin() has been called and so we have to fetch the
		 * first matchable tuple for both outer and inner subplans. We
		 * do the outer side in INITIALIZE_OUTER state, then advance
		 * to INITIALIZE_INNER state for the inner subplan.
		 */
		case EXEC_MJ_INITIALIZE_OUTER:
			MJ_printf("ExecMergeJoin: EXEC_MJ_INITIALIZE_OUTER\n");

			outerTupleSlot = ExecProcNode(outerPlan);
			node->mj_OuterTupleSlot = outerTupleSlot;
			if (TupIsNull(outerTupleSlot))
			{
				MJ_printf("ExecMergeJoin: nothing in outer subplan\n");
				if (doFillInner)
				{
					/*
					 * Need to emit right-join tuples for remaining inner
					 * tuples.	We set MatchedInner = true to force the
					 * ENDOUTER state to advance inner.
					 */
					node->mj_JoinState = EXEC_MJ_ENDOUTER;
					node->mj_MatchedInner = true;
					break;
				}
				/* Otherwise we're done. */
				return NULL;
			}

			/* Compute join values and check for unmatchability */
			if (MJEvalOuterValues(node))
			{
				/* OK to go get the first inner tuple */
				node->mj_JoinState = EXEC_MJ_INITIALIZE_INNER;
			}
			else
			{
				/* Stay in same state to fetch next outer tuple */
				if (doFillOuter)
				{
					/*
					 * Generate a fake join tuple with nulls for the inner
					 * tuple, and return it if it passes the non-join
					 * quals.
					 */
					TupleTableSlot *result;

					result = MJFillOuter(node);
					if (result)
						return result;
				}
			}
			break;

		case EXEC_MJ_INITIALIZE_INNER:
			MJ_printf("ExecMergeJoin: EXEC_MJ_INITIALIZE_INNER\n");

			innerTupleSlot = ExecProcNode(innerPlan);
			node->mj_InnerTupleSlot = innerTupleSlot;

			print_slot(innerTupleSlot);
			if (TupIsNull(innerTupleSlot))
			{
				MJ_printf("ExecMergeJoin: nothing in inner subplan\n");
				if (doFillOuter)
				{
					/*
					 * Need to emit left-join tuples for all outer tuples,
					 * including the one we just fetched.  We set
					 * MatchedOuter = false to force the ENDINNER state to
					 * emit first tuple before advancing outer.
					 */
					node->mj_JoinState = EXEC_MJ_ENDINNER;
					node->mj_MatchedOuter = false;
					break;
				}
				/* Otherwise we're done. */
				return NULL;
			}

			/* Compute join values and check for unmatchability */
			if (MJEvalInnerValues(node, innerTupleSlot)) /*Not null*/
			{
				/*Changed by Yasin SJ*/
				if (node->mj_SimjoinType == 2){
					ExecCopySlot(node->mj_InnerCopyTupleSlot, node->mj_InnerTupleSlot);/*make an independent copy of inner copyinner <-inner*/ /*not null*/

					/*We get the next inner tuple*/
					innerTupleSlot = ExecProcNode(innerPlan); /*This updates automatically node->mj_InnerTupleSlot*/
					node->mj_InnerTupleSlot = innerTupleSlot; /*This could be null*/
					print_slot(innerTupleSlot);
					if (!TupIsNull(innerTupleSlot))
						(void) MJEvalNextInnerValues(node, innerTupleSlot); /*We assume result=true. Extend this to handle result=false*/ /*Not null*/
				}

				/*
				 * OK, we have the initial tuples.	Begin by skipping
				 * non-matching tuples.
				 */
				node->mj_JoinState = EXEC_MJ_SKIP_TEST;
			}
			else
			{
				/* Stay in same state to fetch next inner tuple */
				if (doFillInner)
				{
					/*
					 * Generate a fake join tuple with nulls for the outer
					 * tuple, and return it if it passes the non-join
					 * quals.
					 */
					TupleTableSlot *result;

					result = MJFillInner(node);
					if (result)
						return result;
				}
			}
			break;

			/*
			 * EXEC_MJ_JOINTUPLES means we have two tuples which satisfied
			 * the merge clause so we join them and then proceed to get
			 * the next inner tuple (EXEC_MJ_NEXTINNER).
			 */
		case EXEC_MJ_JOINTUPLES:
			MJ_printf("ExecMergeJoin: EXEC_MJ_JOINTUPLES\n");

			/*
			 * Set the next state machine state.  The right things will
			 * happen whether we return this join tuple or just fall
			 * through to continue the state machine execution.
			 */
			if(node->mj_SimjoinType == 3) //changed by S Pearson
			{
				outerTupleSlot = node->returnOuter;
				innerTupleSlot = node->returnInner;

				node->mj_JoinState = node->processingPartition ?
					EXEC_DBSIMJOIN_NESTLOOP :
					EXEC_DBSIMJOIN_NESTWINLOOP;
			}
			else
			{
				node->mj_JoinState = EXEC_MJ_NEXTINNER;

				/*
				 * Check the extra qual conditions to see if we actually want
				 * to return this join tuple.  If not, can proceed with merge.
				 * We must distinguish the additional joinquals (which must
				 * pass to consider the tuples "matched" for outer-join logic)
				 * from the otherquals (which must pass before we actually
				 * return the tuple).
				 *
				 * We don't bother with a ResetExprContext here, on the
				 * assumption that we just did one while checking the merge
				 * qual.  One per tuple should be sufficient.  We do have to
				 * set up the econtext links to the tuples for ExecQual to
				 * use.
				 */

				outerTupleSlot = node->mj_OuterTupleSlot;
			}
			econtext->ecxt_outertuple = outerTupleSlot;
			if (node->mj_SimjoinType == 2)/*Join Around uses copy of current tuple*/
				innerTupleSlot = node->mj_InnerCopyTupleSlot;
			econtext->ecxt_innertuple = innerTupleSlot;

			if (node->js.jointype == JOIN_IN &&
					node->mj_MatchedOuter)
				qualResult = false;
			else
			{
				qualResult = (joinqual == NIL ||
						ExecQual(joinqual, econtext, false));
				MJ_DEBUG_QUAL(joinqual, qualResult);
			}

			if (qualResult)
			{
				node->mj_MatchedOuter = true;
				node->mj_MatchedInner = true;

				qualResult = (otherqual == NIL ||
						ExecQual(otherqual, econtext, false));
				MJ_DEBUG_QUAL(otherqual, qualResult);

				if (qualResult)
				{
					/*
					 * qualification succeeded.  now form the desired
					 * projection tuple and return the slot containing it.
					 */
					TupleTableSlot *result;
					ExprDoneCond isDone;

					MJ_printf("ExecMergeJoin: returning tuple\n");

					result = ExecProject(node->js.ps.ps_ProjInfo,
							&isDone);

					if (isDone != ExprEndResult)
					{
						node->js.ps.ps_TupFromTlist =
								(isDone == ExprMultipleResult);
						return result;
					}
				}
			}
			break;

			/*
			 * EXEC_MJ_NEXTINNER means advance the inner scan to the next
			 * tuple. If the tuple is not nil, we then proceed to test it
			 * against the join qualification.
			 *
			 * Before advancing, we check to see if we must emit an
			 * outer-join fill tuple for this inner tuple.
			 */
		case EXEC_MJ_NEXTINNER:
			MJ_printf("ExecMergeJoin: EXEC_MJ_NEXTINNER\n");

			if (doFillInner && !node->mj_MatchedInner)
			{
				/*
				 * Generate a fake join tuple with nulls for the outer
				 * tuple, and return it if it passes the non-join quals.
				 */
				TupleTableSlot *result;

				node->mj_MatchedInner = true;		/* do it only once */

				result = MJFillInner(node);
				if (result)
					return result;
			}

			/*Changed by Yasin SJ*/
			if (node->mj_SimjoinType == 1){ /*Epsilon -Join*/
				ExecCopySlot(node->mj_PreviousInnerTupleSlot, node->mj_InnerTupleSlot);/*gets an independent copy. prev <-inner*/ /*not null*/
			}
			else if (node->mj_SimjoinType == 2){ /*Join-Around*/

				if (!TupIsNull(node->mj_InnerTupleSlot)){
					ExecCopySlot(node->mj_PreviousInnerTupleSlot, node->mj_InnerCopyTupleSlot);/*gets an independent copy. previnner <-copyinner*//*Not null*/
					ExecCopySlot(node->mj_InnerCopyTupleSlot, node->mj_InnerTupleSlot);/*gets an independent copy. copyinner <-inner(last)*//*Not null*/
				}
				else{/*mj_InnerTupleSlot is Null*/
					/*In this case does not need previnner*/
					ExecClearTuple(node->mj_InnerCopyTupleSlot); /*This will free memory and make it point to null tuple*/
				}
			}


			/*
			 * now we get the next inner tuple, if any.  If there's none,
			 * advance to next outer tuple (which may be able to join to
			 * previously marked tuples).
			 */

			/*Changed by Yasin SJ*/
			if (node->mj_SimjoinType == 2){ /*Join Around*/

				if (TupIsNull(node->mj_InnerTupleSlot)){/*We already found end of inner*/
					node->mj_JoinState = EXEC_MJ_NEXTOUTER;
					break;
				}
				else{/*There are still inner tuples to be read*/
					innerTupleSlot = ExecProcNode(innerPlan);
					node->mj_InnerTupleSlot = innerTupleSlot;
					node->mj_MatchedInner = false;

					(void) MJEvalInnerValues(node, node->mj_PreviousInnerTupleSlot);/*not null*/
					(void) MJEvalNextInnerValues(node, node->mj_InnerCopyTupleSlot);/*not null*/
					compareResult = MJCompare_InnerAndNextInner(node);/*=0 if equal, !=0 otherwise*/
					if (compareResult == 0)
						node->mj_JoinState = EXEC_MJ_JOINTUPLES;
					else
						node->mj_JoinState = EXEC_MJ_NEXTOUTER;
					break;
				}
			}

			innerTupleSlot = ExecProcNode(innerPlan);
			node->mj_InnerTupleSlot = innerTupleSlot;
			MJ_DEBUG_PROC_NODE(innerTupleSlot);
			node->mj_MatchedInner = false;
			if (TupIsNull(innerTupleSlot))
			{
				node->mj_JoinState = EXEC_MJ_NEXTOUTER;
				break;
			}

			/*
			 * Load up the new inner tuple's comparison values.  If we see
			 * that it contains a NULL and hence can't match any outer
			 * tuple, we can skip the comparison and assume the new tuple
			 * is greater than current outer.
			 */
			if (!MJEvalInnerValues(node, innerTupleSlot)) /*Not null*/
			{
				node->mj_JoinState = EXEC_MJ_NEXTOUTER;
				break;
			}

			/*
			 * Test the new inner tuple to see if it matches outer.
			 *
			 * If they do match, then we join them and move on to the next
			 * inner tuple (EXEC_MJ_JOINTUPLES).
			 *
			 * If they do not match then advance to next outer tuple.
			 */

			compareResult = MJCompare(node);

			MJ_DEBUG_COMPARE(compareResult);

			if (compareResult == 0)
				node->mj_JoinState = EXEC_MJ_JOINTUPLES;
			else
			{
				Assert(compareResult < 0);
				node->mj_JoinState = EXEC_MJ_NEXTOUTER;
			}
			break;

			/*-------------------------------------------
			 * EXEC_MJ_NEXTOUTER means
			 *
			 *				outer inner
			 * outer tuple -  5		5  - marked tuple
			 *				  5		5
			 *				  6		6  - inner tuple
			 *				  7		7
			 *
			 * we know we just bumped into the
			 * first inner tuple > current outer tuple (or possibly
			 * the end of the inner stream)
			 * so get a new outer tuple and then
			 * proceed to test it against the marked tuple
			 * (EXEC_MJ_TESTOUTER)
			 *
			 * Before advancing, we check to see if we must emit an
			 * outer-join fill tuple for this outer tuple.
			 *------------------------------------------------
			 */
		case EXEC_MJ_NEXTOUTER:
			MJ_printf("ExecMergeJoin: EXEC_MJ_NEXTOUTER\n");

			if (doFillOuter && !node->mj_MatchedOuter)
			{
				/*
				 * Generate a fake join tuple with nulls for the inner
				 * tuple, and return it if it passes the non-join quals.
				 */
				TupleTableSlot *result;

				node->mj_MatchedOuter = true;		/* do it only once */

				result = MJFillOuter(node);
				if (result)
					return result;
			}

			/*
			 * now we get the next outer tuple, if any
			 */
			outerTupleSlot = ExecProcNode(outerPlan);
			node->mj_OuterTupleSlot = outerTupleSlot;
			MJ_DEBUG_PROC_NODE(outerTupleSlot);
			node->mj_MatchedOuter = false;

			/*
			 * if the outer tuple is null then we are done with the join,
			 * unless we have inner tuples we need to null-fill.
			 */
			if (TupIsNull(outerTupleSlot))
			{
				MJ_printf("ExecMergeJoin: end of outer subplan\n");
				innerTupleSlot = node->mj_InnerTupleSlot;
				if (doFillInner && !TupIsNull(innerTupleSlot))
				{
					/*
					 * Need to emit right-join tuples for remaining inner
					 * tuples.
					 */
					node->mj_JoinState = EXEC_MJ_ENDOUTER;
					break;
				}
				/* Otherwise we're done. */
				return NULL;
			}

			/* Compute join values and check for unmatchability */
			if (MJEvalOuterValues(node))
			{
				/* Go test the new tuple against the marked tuple */
				node->mj_JoinState = EXEC_MJ_TESTOUTER;
			}
			else
			{
				/* Can't match, so fetch next outer tuple */
				node->mj_JoinState = EXEC_MJ_NEXTOUTER;
			}
			break;

			/*--------------------------------------------------------
			 * EXEC_MJ_TESTOUTER If the new outer tuple and the marked
			 * tuple satisfy the merge clause then we know we have
			 * duplicates in the outer scan so we have to restore the
			 * inner scan to the marked tuple and proceed to join the
			 * new outer tuple with the inner tuples.
			 *
			 * This is the case when
			 *						  outer inner
			 *							4	  5  - marked tuple
			 *			 outer tuple -	5	  5
			 *		 new outer tuple -	5	  5
			 *							6	  8  - inner tuple
			 *							7	 12
			 *
			 *				new outer tuple == marked tuple
			 *
			 * If the outer tuple fails the test, then we are done
			 * with the marked tuples, and we have to look for a
			 * match to the current inner tuple.  So we will
			 * proceed to skip outer tuples until outer >= inner
			 * (EXEC_MJ_SKIP_TEST).
			 *
			 *		This is the case when
			 *
			 *						  outer inner
			 *							5	  5  - marked tuple
			 *			 outer tuple -	5	  5
			 *		 new outer tuple -	6	  8  - inner tuple
			 *							7	 12
			 *
			 *				new outer tuple > marked tuple
			 *
			 *---------------------------------------------------------
			 */
		case EXEC_MJ_TESTOUTER:
			MJ_printf("ExecMergeJoin: EXEC_MJ_TESTOUTER\n");

			/*
			 * Here we must compare the outer tuple with the marked inner
			 * tuple.  (We can ignore the result of MJEvalInnerValues,
			 * since the marked inner tuple is certainly matchable.)
			 */

			/*Changed by Yasin SJ*/
			if (node->mj_SimjoinType == 0){/*No SJ (Regular Join)*/
				innerTupleSlot = node->mj_MarkedTupleSlot; /*Old code*/
				(void) MJEvalInnerValues(node, innerTupleSlot);
				compareResult = MJCompare(node);
			}
			else if (node->mj_SimjoinType == 1){ /*Epsilon-Join*/ /*MJCompare uses outer and prev(instead of inner)*/
				innerTupleSlot = node->mj_PreviousInnerTupleSlot;
				(void) MJEvalInnerValues(node, innerTupleSlot); /*not null*/
				compareResult = MJCompare(node);
			}
			else if (node->mj_SimjoinType == 2){ /*Join-Around*/ /*MJCompare uses outer, prev(instead of inner), and copyinner(instead of next)*/
				if (TupIsNull(node->mj_InnerCopyTupleSlot))
					compareResult = 0; /*If we got to the end of inner, outer matches marked (and previous)*/
				else{/*Both prev and innercopy are not null*/
					(void) MJEvalNextInnerValues(node, node->mj_InnerCopyTupleSlot); /*not null*/
					innerTupleSlot = node->mj_PreviousInnerTupleSlot;
					(void) MJEvalInnerValues(node, innerTupleSlot); /*not null*/
					compareResult = MJCompare(node);
				}
			}


			MJ_DEBUG_COMPARE(compareResult);

			if (compareResult == 0)
			{
				/*Changed by Yasin SJ*/
				if (node->mj_SimjoinType == 0){/*Old code*/
					/*
					 * the merge clause matched so now we restore the inner
					 * scan position to the first mark, and go join that tuple
					 * (and any following ones) to the new outer.
					 *
					 * NOTE: we do not need to worry about the MatchedInner
					 * state for the rescanned inner tuples.  We know all of
					 * them will match this new outer tuple and therefore
					 * won't be emitted as fill tuples.  This works *only*
					 * because we require the extra joinquals to be nil when
					 * doing a right or full join --- otherwise some of the
					 * rescanned tuples might fail the extra joinquals.
					 */
					ExecRestrPos(innerPlan);

					/*
					 * ExecRestrPos probably should give us back a new Slot,
					 * but since it doesn't, use the marked slot.  (The
					 * previously returned mj_InnerTupleSlot cannot be assumed
					 * to hold the required tuple.)
					 */
					node->mj_InnerTupleSlot = innerTupleSlot;
					/* we need not do MJEvalInnerValues again */

					node->mj_JoinState = EXEC_MJ_JOINTUPLES;
				}
				if (node->mj_SimjoinType == 1){ /*Epsilon-Join*/
					innerTupleSlot = node->mj_MarkedTupleSlot; /*inner should point to markedtuple*/
					(void) MJEvalInnerValues(node, innerTupleSlot); /*marked tuple cannot be null*/
					ExecRestrPos(innerPlan);
					node->mj_InnerTupleSlot = innerTupleSlot;

					node->mj_JoinState = EXEC_MJ_SKIP_TEST; /*Because some tuples after mark cannot match*/
				}
				else if (node->mj_SimjoinType == 2){ /*Join-Around*/
					ExecRestrPos(innerPlan);

					ExecCopySlot(node->mj_InnerCopyTupleSlot, node->mj_FirstMarkedSlot);/*gets an independent copy*/ /*Not null*/
					(void) MJEvalInnerValues(node, node->mj_InnerCopyTupleSlot); /*Evaluate new inner*/ /*1st Marked cannot be null*/

					if (!TupIsNull(node->mj_MarkedTupleSlot)){/*If it is null, node->mj_InnerTupleSlot is already null*/
						ExecCopySlot(node->mj_InnerTupleSlot, node->mj_MarkedTupleSlot);/*gets an independent copy*/ /*not null*/
						(void) MJEvalNextInnerValues(node, node->mj_InnerTupleSlot); /*Evaluate new nextinner*/ /*not null*/
					}
					node->mj_JoinState = EXEC_MJ_JOINTUPLES;
				}
			}
			else/*No match*/ /*notice that for J-Around if node->mj_InnerTupleSlot, it will never enter this case*/
			{
				/* ----------------
				 *	if the new outer tuple didn't match the marked inner
				 *	tuple then we have a case like:
				 *
				 *			 outer inner
				 *			   4	 4	- marked tuple
				 * new outer - 5	 4
				 *			   6	 5	- inner tuple
				 *			   7
				 *
				 *	which means that all subsequent outer tuples will be
				 *	larger than our marked inner tuples.  So we need not
				 *	revisit any of the marked tuples but can proceed to
				 *	look for a match to the current inner.	If there's
				 *	no more inners, we are done.
				 * ----------------
				 */

				/*Changed by Yasin SJ*/
				if (node->mj_SimjoinType == 0)/*No SJ (Regular Join)*/
					Assert(compareResult > 0); /*Old code. In SJ new outer could be smaller or bigger than mark*/
				if (node->mj_SimjoinType == 2)
					innerTupleSlot = node->mj_InnerCopyTupleSlot;
				else
					innerTupleSlot = node->mj_InnerTupleSlot;


				if (TupIsNull(innerTupleSlot))
				{
					if (doFillOuter)
					{
						/*
						 * Need to emit left-join tuples for remaining
						 * outer tuples.
						 */
						node->mj_JoinState = EXEC_MJ_ENDINNER;
						break;
					}
					/* Otherwise we're done. */
					return NULL;
				}

				/* reload comparison data for current inner */
				if (MJEvalInnerValues(node, innerTupleSlot)) /*not null*/
				{
					/*Changed by Yasin SJ*/
					if (node->mj_SimjoinType == 2){ /*Join-Around*/
						if (!TupIsNull(node->mj_InnerTupleSlot))
							(void) MJEvalNextInnerValues(node, node->mj_InnerTupleSlot);/*reload comparison data for next inner*/ /*not null*/
					}
					/* proceed to compare it to the current outer */
					node->mj_JoinState = EXEC_MJ_SKIP_TEST;
				}
				else
				{
					/*
					 * current inner can't possibly match any outer;
					 * better to advance the inner scan than the outer.
					 */
					node->mj_JoinState = EXEC_MJ_SKIPINNER_ADVANCE;
				}
			}
			break;

			/*----------------------------------------------------------
			 * EXEC_MJ_SKIP means compare tuples and if they do not
			 * match, skip whichever is lesser.
			 *
			 * For example:
			 *
			 *				outer inner
			 *				  5		5
			 *				  5		5
			 * outer tuple -  6		8  - inner tuple
			 *				  7    12
			 *				  8    14
			 *
			 * we have to advance the outer scan
			 * until we find the outer 8.
			 *
			 * On the other hand:
			 *
			 *				outer inner
			 *				  5		5
			 *				  5		5
			 * outer tuple - 12		8  - inner tuple
			 *				 14    10
			 *				 17    12
			 *
			 * we have to advance the inner scan
			 * until we find the inner 12.
			 *----------------------------------------------------------
			 */
		case EXEC_MJ_SKIP_TEST:
			MJ_printf("ExecMergeJoin: EXEC_MJ_SKIP_TEST\n");

			/*
			 * before we advance, make sure the current tuples do not
			 * satisfy the mergeclauses.  If they do, then we update the
			 * marked tuple position and go join them.
			 */

			/*Changed by Yasin SJ*/
			if (node->mj_SimjoinType == 2){/*Join-Around*/
				if (TupIsNull(node->mj_InnerTupleSlot)){ /*If (next)Inner is null, process MATCH with current inner (copy)*/
					/*Mark: 1st marked is stored separately, remaining is marked as usual*/
					ExecCopySlot(node->mj_FirstMarkedSlot, node->mj_InnerCopyTupleSlot);/*Make an independent copy. fisrtmarked <-innercopy*//*Not null*/
					ExecMarkPos(innerPlan);
					ExecClearTuple(node->mj_MarkedTupleSlot); /*Free memory and make mj_MarkedTupleSlot point to Null tuple*/
					node->mj_JoinState = EXEC_MJ_JOINTUPLES;
					break;
				}

				compareResult = MJCompare(node);/*Uses outer, inner, next (all should be already evaluated)*/
				compareResult2 = MJCompare_InnerAndNextInner(node);/*=0 if equal, !=0 otherwise*/ /*Uses inner and nextinner (all should be already evaluated)*/

				if ((compareResult2 != 0) && (compareResult != 0)){/*No match with current inner*/
					node->mj_JoinState = EXEC_MJ_SKIPINNER_ADVANCE;
					break;
				}
				/*Passes this point if match was found or if we cannot tell if match exists*/
				/*Mark: 1st marked is stored separately, remaining is marked as usual*/
				ExecCopySlot(node->mj_FirstMarkedSlot, node->mj_InnerCopyTupleSlot);/*Make an independent copy. fisrtmarked <-innercopy*/ /*not null*/
				ExecMarkPos(innerPlan);
				MarkInnerTuple(node->mj_InnerTupleSlot, node);


				if ((compareResult2 == 0) && (compareResult > 0)){/*Cannot tell if outer matches inner (inner=nextinner)*/
					do{/*advance inner, nextinner*/
						ExecCopySlot(node->mj_InnerCopyTupleSlot, node->mj_InnerTupleSlot);/*independent copy*/ /*Not null*/
						(void) MJEvalInnerValues(node, node->mj_InnerTupleSlot); /*Not null*/
						/*Load next inner*/
						innerTupleSlot = ExecProcNode(innerPlan);
						node->mj_InnerTupleSlot = innerTupleSlot;

						if (TupIsNull(innerTupleSlot)) /*Outer matches inner*/
							break;

						/* Compute join values and check for unmatchability */
						(void) MJEvalNextInnerValues(node, innerTupleSlot); /*not null*/
					}
					while(MJCompare_InnerAndNextInner(node) == 0);

					if (TupIsNull(innerTupleSlot)){/*Match*/
						ExecRestrPos(innerPlan);
						ExecCopySlot(node->mj_InnerCopyTupleSlot, node->mj_FirstMarkedSlot);/*gets an independent copy*/ /*not null*/
						(void) MJEvalInnerValues(node, node->mj_InnerCopyTupleSlot); /*Evaluate new inner*/ /*not null*/

						ExecCopySlot(node->mj_InnerTupleSlot, node->mj_MarkedTupleSlot);/*gets an independent copy*/ /*not null*/
						(void) MJEvalNextInnerValues(node, node->mj_InnerTupleSlot); /*Evaluate new nextinner*/ /*Cannot be null since original inner(marked)=original nextinner*/
					}
					else if(MJCompare(node)) {/*Match*/ /*This block should be equal to the previous one*/
						ExecRestrPos(innerPlan);
						ExecCopySlot(node->mj_InnerCopyTupleSlot, node->mj_FirstMarkedSlot);/*gets an independent copy*/ /*Not null*/
						(void) MJEvalInnerValues(node, node->mj_InnerCopyTupleSlot); /*Evaluate new inner*/ /*Not null*/

						ExecCopySlot(node->mj_InnerTupleSlot, node->mj_MarkedTupleSlot);/*gets an independent copy*/ /*Not null*/
						(void) MJEvalNextInnerValues(node, node->mj_InnerTupleSlot); /*Evaluate new nextinner*/ /*Cannot be null since original inner(marked)=original nextinner*/
					}
					else{/*No match was found. Obs: nextinner cannot be null*/
						node->mj_JoinState = EXEC_MJ_SKIPINNER_ADVANCE; /*We already know that o does not match current inner*/
						break; /*Evaluation of inner and next inner already made*/
					}
				}
				node->mj_JoinState = EXEC_MJ_JOINTUPLES;
				break;
			}
			else{ /*Old Code*/
				compareResult = MJCompare(node);
				MJ_DEBUG_COMPARE(compareResult);
				if (compareResult == 0)
				{
					ExecMarkPos(innerPlan);
					MarkInnerTuple(node->mj_InnerTupleSlot, node);
					node->mj_JoinState = EXEC_MJ_JOINTUPLES;
				}
				else if (compareResult < 0)
					node->mj_JoinState = EXEC_MJ_SKIPOUTER_ADVANCE;
				else/* compareResult > 0 */
					node->mj_JoinState = EXEC_MJ_SKIPINNER_ADVANCE;
				break;
			}

			/*
			 * Before advancing, we check to see if we must emit an
			 * outer-join fill tuple for this outer tuple.
			 */
		case EXEC_MJ_SKIPOUTER_ADVANCE:
			MJ_printf("ExecMergeJoin: EXEC_MJ_SKIPOUTER_ADVANCE\n");

			if (doFillOuter && !node->mj_MatchedOuter)
			{
				/*
				 * Generate a fake join tuple with nulls for the inner
				 * tuple, and return it if it passes the non-join quals.
				 */
				TupleTableSlot *result;

				node->mj_MatchedOuter = true;		/* do it only once */

				result = MJFillOuter(node);
				if (result)
					return result;
			}

			/*
			 * now we get the next outer tuple, if any
			 */
			outerTupleSlot = ExecProcNode(outerPlan);
			node->mj_OuterTupleSlot = outerTupleSlot;
			MJ_DEBUG_PROC_NODE(outerTupleSlot);
			node->mj_MatchedOuter = false;

			/*
			 * if the outer tuple is null then we are done with the join,
			 * unless we have inner tuples we need to null-fill.
			 */
			if (TupIsNull(outerTupleSlot))
			{
				MJ_printf("ExecMergeJoin: end of outer subplan\n");
				innerTupleSlot = node->mj_InnerTupleSlot;
				if (doFillInner && !TupIsNull(innerTupleSlot))
				{
					/*
					 * Need to emit right-join tuples for remaining inner
					 * tuples.
					 */
					node->mj_JoinState = EXEC_MJ_ENDOUTER;
					break;
				}
				/* Otherwise we're done. */
				return NULL;
			}

			/* Compute join values and check for unmatchability */
			if (MJEvalOuterValues(node))
			{
				/* Go test the new tuple against the current inner */
				node->mj_JoinState = EXEC_MJ_SKIP_TEST;
			}
			else
			{
				/* Can't match, so fetch next outer tuple */
				node->mj_JoinState = EXEC_MJ_SKIPOUTER_ADVANCE;
			}
			break;

			/*
			 * Before advancing, we check to see if we must emit an
			 * outer-join fill tuple for this inner tuple.
			 */
		case EXEC_MJ_SKIPINNER_ADVANCE:
			MJ_printf("ExecMergeJoin: EXEC_MJ_SKIPINNER_ADVANCE\n");

			if (doFillInner && !node->mj_MatchedInner)
			{
				/*
				 * Generate a fake join tuple with nulls for the outer
				 * tuple, and return it if it passes the non-join quals.
				 */
				TupleTableSlot *result;

				node->mj_MatchedInner = true;		/* do it only once */

				result = MJFillInner(node);
				if (result)
					return result;
			}

			if (node->mj_SimjoinType == 2){/*Join-Around*/
				ExecCopySlot(node->mj_InnerCopyTupleSlot, node->mj_InnerTupleSlot);/*make an independent copy*/ /*Cannot be null*/
				(void) MJEvalInnerValues(node, node->mj_InnerCopyTupleSlot); /*Not null*/
			}

			/*
			 * now we get the next inner tuple, if any
			 */
			innerTupleSlot = ExecProcNode(innerPlan);
			node->mj_InnerTupleSlot = innerTupleSlot;

			MJ_DEBUG_PROC_NODE(innerTupleSlot);
			node->mj_MatchedInner = false;

			/*
			 * if the inner tuple is null then we are done with the join (except in Join-Around),
			 * unless we have outer tuples we need to null-fill.
			 */
			if (TupIsNull(innerTupleSlot))
			{
				/*Changed by Yasin SJ*/
				if (node->mj_SimjoinType ==2){/*Join-Around*/
					node->mj_JoinState = EXEC_MJ_SKIP_TEST;/*Outer matches current inner since nextinner is null*/
					/*We send to SKIP first to mark*/
					break;
				}

				MJ_printf("ExecMergeJoin: end of inner subplan\n");
				outerTupleSlot = node->mj_OuterTupleSlot;
				if (doFillOuter && !TupIsNull(outerTupleSlot))
				{
					/*
					 * Need to emit left-join tuples for remaining outer
					 * tuples.
					 */
					node->mj_JoinState = EXEC_MJ_ENDINNER;
					break;
				}

				/* Otherwise we're done. */
				return NULL;
			}

			/* Compute join values and check for unmatchability */
			/*Changed by Yasin SJ*/
			if (node->mj_SimjoinType ==2){/*Join-Around*/
				(void) MJEvalNextInnerValues(node, innerTupleSlot);/*We assume result=true. This can be extended to handle result=false*/ /*Not null*/
				node->mj_JoinState = EXEC_MJ_SKIP_TEST;
			}
			else{/*Old code*/
				if (MJEvalInnerValues(node, innerTupleSlot))
				{/* proceed to compare it to the current outer */
					node->mj_JoinState = EXEC_MJ_SKIP_TEST;
				}
				else
				{
					/*
					 * current inner can't possibly match any outer; better to
					 * advance the inner scan than the outer.
					 */
					node->mj_JoinState = EXEC_MJ_SKIPINNER_ADVANCE;
				}
			}
			break;

			/*
			 * EXEC_MJ_ENDOUTER means we have run out of outer tuples, but
			 * are doing a right/full join and therefore must null-fill
			 * any remaining unmatched inner tuples.
			 */
		case EXEC_MJ_ENDOUTER:
			MJ_printf("ExecMergeJoin: EXEC_MJ_ENDOUTER\n");

			Assert(doFillInner);

			if (!node->mj_MatchedInner)
			{
				/*
				 * Generate a fake join tuple with nulls for the outer
				 * tuple, and return it if it passes the non-join quals.
				 */
				TupleTableSlot *result;

				node->mj_MatchedInner = true;		/* do it only once */

				result = MJFillInner(node);
				if (result)
					return result;
			}

			/*
			 * now we get the next inner tuple, if any
			 */
			innerTupleSlot = ExecProcNode(innerPlan);
			node->mj_InnerTupleSlot = innerTupleSlot;
			MJ_DEBUG_PROC_NODE(innerTupleSlot);
			node->mj_MatchedInner = false;

			if (TupIsNull(innerTupleSlot))
			{
				MJ_printf("ExecMergeJoin: end of inner subplan\n");
				return NULL;
			}

			/* Else remain in ENDOUTER state and process next tuple. */
			break;

			/*
			 * EXEC_MJ_ENDINNER means we have run out of inner tuples, but
			 * are doing a left/full join and therefore must null- fill
			 * any remaing unmatched outer tuples.
			 */
		case EXEC_MJ_ENDINNER:
			MJ_printf("ExecMergeJoin: EXEC_MJ_ENDINNER\n");

			Assert(doFillOuter);

			if (!node->mj_MatchedOuter)
			{
				/*
				 * Generate a fake join tuple with nulls for the inner
				 * tuple, and return it if it passes the non-join quals.
				 */
				TupleTableSlot *result;

				node->mj_MatchedOuter = true;		/* do it only once */

				result = MJFillOuter(node);
				if (result)
					return result;
			}

			/*
			 * now we get the next outer tuple, if any
			 */
			outerTupleSlot = ExecProcNode(outerPlan);
			node->mj_OuterTupleSlot = outerTupleSlot;
			MJ_DEBUG_PROC_NODE(outerTupleSlot);
			node->mj_MatchedOuter = false;

			if (TupIsNull(outerTupleSlot))
			{
				MJ_printf("ExecMergeJoin: end of outer subplan\n");
				return NULL;
			}

			/* Else remain in ENDINNER state and process next tuple. */
			break;

			/*
			 * broken state value?
			 */
		default:
			elog(ERROR, "unrecognized mergejoin state: %d",
					(int) node->mj_JoinState);
			break;
		}
	}
}

/* ----------------------------------------------------------------
 *		ExecInitMergeJoin
 * ----------------------------------------------------------------
 */
MergeJoinState *
ExecInitMergeJoin(MergeJoin *node, EState *estate, int eflags)
{

	MergeJoinState *mergestate;

	/* check for unsupported flags */
	Assert(!(eflags & (EXEC_FLAG_BACKWARD | EXEC_FLAG_MARK)));

	MJ1_printf("ExecInitMergeJoin: %s\n",
			"initializing node");

	/*
	 * create state structure
	 */
	mergestate = makeNode(MergeJoinState);
	mergestate->js.ps.plan = (Plan *) node;
	mergestate->js.ps.state = estate;


	/* CHANGED BY YASIN */
	if (eflags & EXEC_FLAG_INAROUND) mergestate->js.ps.ps_InAround = true;

	/* CHANGED BY Ahmed.M.Aly */
	/************************
	 * XXX SET TO DBSimJoin *
	 ************************/
	mergestate->mj_SimjoinType = 3;
	//mergestate->mj_SimjoinType = node->SJ_Type; /* 0=No SJ, 1=Epsilon-Join, 2=Join-Around */ /*3=DBSimJoin*/

	/*
	 * Miscellaneous initialization
	 *
	 * create expression context for node
	 */
	ExecAssignExprContext(estate, &mergestate->js.ps);

	/*
	 * we need two additional econtexts in which we can compute the join
	 * expressions from the left and right input tuples.  The node's regular
	 * econtext won't do because it gets reset too often.
	 */
	mergestate->mj_OuterEContext = CreateExprContext(estate);
	mergestate->mj_InnerEContext = CreateExprContext(estate);

	/*
	 * initialize child expressions
	 */
	mergestate->js.ps.targetlist = (List *)
																																														ExecInitExpr((Expr *) node->join.plan.targetlist,
																																																(PlanState *) mergestate);
	mergestate->js.ps.qual = (List *)
																																														ExecInitExpr((Expr *) node->join.plan.qual,
																																																(PlanState *) mergestate);
	mergestate->js.jointype = node->join.jointype;
	mergestate->js.joinqual = (List *)
																																														ExecInitExpr((Expr *) node->join.joinqual,
																																																(PlanState *) mergestate);
	/* mergeclauses are handled below */

	/*
	 * initialize child nodes
	 *
	 * inner child must support MARK/RESTORE.
	 */

	outerPlanState(mergestate) = ExecInitNode(outerPlan(node), estate, eflags);
	innerPlanState(mergestate) = ExecInitNode(innerPlan(node), estate,
			eflags | EXEC_FLAG_MARK);
	/*  Yasin important notes.
    To test Join around we need to disable conmmutativity.
    For testing Join-Around we found that the optimizer conmmuted the relations.
    Consequently we had to replace the previous code by this commented code to reasign
    inner and outer to the right inputs.

	outerPlanState(mergestate) = ExecInitNode(innerPlan(node), estate, eflags);
	innerPlanState(mergestate) = ExecInitNode(outerPlan(node), estate, eflags | EXEC_FLAG_MARK);
	 */


#define MERGEJOIN_NSLOTS 12   /*CHANGED BY YASIN SJ*/
	/*This value was originally 4 */
	/*we added slots for previousInner, copyofCurrentInner, FirstMarkedSlot*/
	/*CHANGED BY S PEARSON*/
	/*Value was 7 before */
	/*Added extraOuter, extraInner, returnOuter, returnInnertempSlot*/

	/*
	 * tuple table initialization
	 */
	ExecInitResultTupleSlot(estate, &mergestate->js.ps);

	mergestate->mj_MarkedTupleSlot = ExecInitExtraTupleSlot(estate);
	ExecSetSlotDescriptor(mergestate->mj_MarkedTupleSlot,
			ExecGetResultType(innerPlanState(mergestate)));

	switch (node->join.jointype)
	{
	case JOIN_INNER:
	case JOIN_IN:
		mergestate->mj_FillOuter = false;
		mergestate->mj_FillInner = false;
		break;
	case JOIN_LEFT:
		mergestate->mj_FillOuter = true;
		mergestate->mj_FillInner = false;
		mergestate->mj_NullInnerTupleSlot =
				ExecInitNullTupleSlot(estate,
						ExecGetResultType(innerPlanState(mergestate)));
		break;
	case JOIN_RIGHT:
		mergestate->mj_FillOuter = false;
		mergestate->mj_FillInner = true;
		mergestate->mj_NullOuterTupleSlot =
				ExecInitNullTupleSlot(estate,
						ExecGetResultType(outerPlanState(mergestate)));

		/*
		 * Can't handle right or full join with non-nil extra joinclauses.
		 * This should have been caught by planner.
		 */
		if (node->join.joinqual != NIL)
			ereport(ERROR,
					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
							errmsg("RIGHT JOIN is only supported with merge-joinable join conditions")));
		break;
	case JOIN_FULL:
		mergestate->mj_FillOuter = true;
		mergestate->mj_FillInner = true;
		mergestate->mj_NullOuterTupleSlot =
				ExecInitNullTupleSlot(estate,
						ExecGetResultType(outerPlanState(mergestate)));
		mergestate->mj_NullInnerTupleSlot =
				ExecInitNullTupleSlot(estate,
						ExecGetResultType(innerPlanState(mergestate)));

		/*
		 * Can't handle right or full join with non-nil extra joinclauses.
		 */
		if (node->join.joinqual != NIL)
			ereport(ERROR,
					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
							errmsg("FULL JOIN is only supported with merge-joinable join conditions")));
		break;
	default:
		elog(ERROR, "unrecognized join type: %d",
				(int) node->join.jointype);
		break;
	}

	/*Changed by Yasin SJ*/
	if (mergestate->mj_SimjoinType == 1){ /*Epsilon-Join*/
		mergestate->mj_PreviousInnerTupleSlot = ExecInitExtraTupleSlot(estate);
		ExecSetSlotDescriptor(mergestate->mj_PreviousInnerTupleSlot,
				ExecGetResultType(innerPlanState(mergestate)));
	}
	else if (mergestate->mj_SimjoinType == 2){ /*Join-Around*/
		mergestate->mj_PreviousInnerTupleSlot = ExecInitExtraTupleSlot(estate);
		ExecSetSlotDescriptor(mergestate->mj_PreviousInnerTupleSlot,
				ExecGetResultType(innerPlanState(mergestate)));
		mergestate->mj_InnerCopyTupleSlot = ExecInitExtraTupleSlot(estate);
		ExecSetSlotDescriptor(mergestate->mj_InnerCopyTupleSlot,
				ExecGetResultType(innerPlanState(mergestate)));
		mergestate->mj_FirstMarkedSlot = ExecInitExtraTupleSlot(estate);
		ExecSetSlotDescriptor(mergestate->mj_FirstMarkedSlot,
				ExecGetResultType(innerPlanState(mergestate)));
	}


	/*
	 * initialize tuple type and projection info
	 */
	ExecAssignResultTypeFromTL(&mergestate->js.ps);
	ExecAssignProjectionInfo(&mergestate->js.ps, NULL);

	/*
	 * preprocess the merge clauses
	 */
	mergestate->mj_NumClauses = list_length(node->mergeclauses);

	/*Changed by Yasin*/
	if (mergestate->mj_SimjoinType == 0){ /*No SJ*/
		mergestate->mj_Clauses = MJExamineQuals(node->mergeclauses,   /*Old code*/
				(PlanState *) mergestate);
	}
	else if((mergestate->mj_SimjoinType == 1) || (mergestate->mj_SimjoinType == 2)){ /* EpsJoin, JoinAround*/
		mergestate->mj_Clauses = MJExamineQuals_AddsMinus(node->mergeclauses,
				(PlanState *) mergestate);
	}
	/*XXX
	 * Changed by S Pearson*/
	else if(mergestate->mj_SimjoinType == 3) { /* DBSimJoin*/
		mergestate->mj_Clauses =
				//MJExamineQuals_Strings			//Uncomment this line for String data
				MJExamineQuals_AddsProductMinus		//Uncomment this line for vector data
				(node->mergeclauses,(PlanState *) mergestate);
	}


	/*
	 * initialize join state
	 */
	/*Changed by S Pearson*/
	if(mergestate->mj_SimjoinType == 3)/* DBSimJoin*/
		mergestate->mj_JoinState = EXEC_DBSIMJOIN_INIT;
	else
		mergestate->mj_JoinState = EXEC_MJ_INITIALIZE_OUTER;
	mergestate->js.ps.ps_TupFromTlist = false;
	mergestate->mj_MatchedOuter = false;
	mergestate->mj_MatchedInner = false;
	mergestate->mj_OuterTupleSlot = NULL;
	mergestate->mj_InnerTupleSlot = NULL;

	/*
	 * get number of partitions and windows
	 * Added by S Pearson
	 */
	if(mergestate->mj_SimjoinType == 3)
	{
		int partSize;

		//Make sure to account for the memTuples array inside a Tuplestore
		void **tempArray = palloc(1024 * sizeof(void *));
		int arraySize = GetMemoryChunkSpace(tempArray);
		pfree(tempArray);
		partSize = sizeOfBufFile() + sizeOfTuplestore() + arraySize;

		mergestate->maxTuples = 15;
		mergestate->availMem = 32 * work_mem * 1024 - partSize; //work_mem in Bytes
		/*************************
		 * WINDOW INITIALIZATION *
		 *************************/
		//Calculate partitionSize and numPartitions
		//mergestate->partitionSize = floor((partSize - arraySize)/2);
		//mergestate->partitionSize = ceil((mergestate->partitionSize + arraySize) / 1024.0); //partitionSize stored in KBytes
		//mergestate->numPartitions = floor(sqrt(mergestate->availMem / partSize)/2);

		//Set # partitions
		mergestate->numPartitions = 50;
		mergestate->partitionSize = mergestate->availMem;
		/***********************
		 * INITIALIZATION END *
		 **********************/
		//Set up ints to keep track of number of partitions and windows
		mergestate->oldNumPartitions = mergestate->numPartitions;
		mergestate->numPartitionsLeft = mergestate->numPartitions;
		mergestate->currentPartition = 0;

		mergestate->numWindows = mergestate->numPartitions * (mergestate->numPartitions - 1);
		mergestate->oldNumWindows = mergestate->numWindows;
		mergestate->numWindowsLeft = mergestate->numWindows;
		mergestate->currentWindow = 0;

		//Set up partition array
		mergestate->partitions = (Tuplestorestate **) palloc(sizeof(Tuplestorestate *)
				* mergestate->numPartitions);
		USEMEM(mergestate, GetMemoryChunkSpace(mergestate->partitions));
		//Set up base window array
		mergestate->windows = (Tuplestorestate **) palloc(sizeof(Tuplestorestate *)
				* mergestate->numWindows);
		USEMEM(mergestate, GetMemoryChunkSpace(mergestate->windows));

		//Set up pivot array
		mergestate->pivots = (MinimalTuple *) palloc(sizeof(MinimalTuple *)
				* mergestate->numPartitions);
		USEMEM(mergestate, GetMemoryChunkSpace(mergestate->pivots));
		mergestate->currentPivot = 0;

		/*Initialize new TupleDescs*/
		mergestate->newOuterDesc = initNewTupleDescTableID(
				ExecGetResultType(outerPlanState(mergestate)));
		mergestate->newInnerDesc = initNewTupleDescTableID(
				ExecGetResultType(innerPlanState(mergestate)));

		/*
		 * Initialize extra TupleSlots
		 */
		mergestate->extraOuter = ExecInitExtraTupleSlot(estate);
		ExecSetSlotDescriptor(mergestate->extraOuter, mergestate->newOuterDesc);
		mergestate->extraInner = ExecInitExtraTupleSlot(estate);
		ExecSetSlotDescriptor(mergestate->extraInner, mergestate->newInnerDesc);
		//Return slots use the original TupleDescs from the plans.
		mergestate->returnOuter = ExecInitExtraTupleSlot(estate);
		ExecSetSlotDescriptor(mergestate->returnOuter,
			ExecGetResultType(outerPlanState(mergestate)));
		mergestate->returnInner = ExecInitExtraTupleSlot(estate);
		ExecSetSlotDescriptor(mergestate->returnInner,
			ExecGetResultType(innerPlanState(mergestate)));
		/*The TupleDesc for tempSlot is irrelevant, but one is required to store
		*values obtained from a Tuplestore
		*/
		mergestate->tempSlot = ExecInitExtraTupleSlot(estate);
		ExecSetSlotDescriptor(mergestate->tempSlot, mergestate->newOuterDesc);

		//Initialize mj_OuterTupleSlot and mj_InnerTupleSlot with dummy values
		mergestate->mj_OuterTupleSlot = ExecInitExtraTupleSlot(estate);
		ExecSetSlotDescriptor(mergestate->mj_OuterTupleSlot, mergestate->newOuterDesc);
		mergestate->mj_InnerTupleSlot = ExecInitExtraTupleSlot(estate);
		ExecSetSlotDescriptor(mergestate->mj_InnerTupleSlot, mergestate->newInnerDesc);

		//Set random seed for picking pivots
		srand(time(0));
	}
	/*
	 * initialization successful
	 */
	MJ1_printf("ExecInitMergeJoin: %s\n",
			"node initialized");

	return mergestate;
}

int
ExecCountSlotsMergeJoin(MergeJoin *node)
{
	return ExecCountSlotsNode(outerPlan((Plan *) node)) +
			ExecCountSlotsNode(innerPlan((Plan *) node)) +
			MERGEJOIN_NSLOTS;
}

/* ----------------------------------------------------------------
 *		ExecEndMergeJoin
 *
 * old comments
 *		frees storage allocated through C routines.
 * ----------------------------------------------------------------
 */
void
ExecEndMergeJoin(MergeJoinState *node)
{
	MJ1_printf("ExecEndMergeJoin: %s\n",
			"ending node processing");

	/*
	 * Free the exprcontext
	 */
	ExecFreeExprContext(&node->js.ps);

	/*
	 * clean out the tuple table
	 */
	ExecClearTuple(node->js.ps.ps_ResultTupleSlot);
	ExecClearTuple(node->mj_MarkedTupleSlot);

	/*
	 * shut down the subplans
	 */
	ExecEndNode(innerPlanState(node));
	ExecEndNode(outerPlanState(node));

	MJ1_printf("ExecEndMergeJoin: %s\n",
			"node processing ended");
}

void
ExecReScanMergeJoin(MergeJoinState *node, ExprContext *exprCtxt)
{
	ExecClearTuple(node->mj_MarkedTupleSlot);

	node->mj_JoinState = EXEC_MJ_INITIALIZE_OUTER;
	node->js.ps.ps_TupFromTlist = false;
	node->mj_MatchedOuter = false;
	node->mj_MatchedInner = false;
	node->mj_OuterTupleSlot = NULL;
	node->mj_InnerTupleSlot = NULL;

	/*
	 * if chgParam of subnodes is not null then plans will be re-scanned by
	 * first ExecProcNode.
	 */
	if (((PlanState *) node)->lefttree->chgParam == NULL)
		ExecReScan(((PlanState *) node)->lefttree, exprCtxt);
	if (((PlanState *) node)->righttree->chgParam == NULL)
		ExecReScan(((PlanState *) node)->righttree, exprCtxt);

}
